#include "utils/rand/rand.h"
#include "debug/ppdebug.h"
/**
 * @file Contains Skiplist data structure. The datastructure is
 * multilevel-linked-list, have O(log n) insert/delete/find.
 * In addition to std::map, it supports two operations
 * <ul>
 *  <li> kth() finds k-th element in sorted list
 *  <li> xth() finds a position of an iterator in the list
 * </ul>
 */
#include "skiplist_node.h"
#include "skiplist_iterator.h"
#include "skiplist_trail.h"
#include <vector>
#include "utils/macros/unused.h"
#include "utils/macros/evil_constructors.h"

namespace balanced_structures {
namespace skiplist {

/**
 * Skiplist is sorted range container with following operations
 * <ul>
 *  <li> insert() in expected <i>O(log N)</i>
 *  <li> erase() in expected <i>O(log N)</i>
 *  <li> find() element by value in expected <i>O(log N)</i>
 *  <li> kth() element in expected <i>O(log N)</i>
 *  <li> xth() find position of current element in expected <i>O(log N)</i>
 *
 * Skiplist:
 * <pre>
 * Nodes:
 * level: 3    1    4    2    1    3    4
 *
 *   # -----(3)---> # -------(4)------> #
 *   # -> # --(2)-> # ----(3)----> # -> #
 *   # -> # --(2)-> # -> # --(2)-> # -> #
 *   # -> # -> # -> # -> # -> # -> # -> #
 * HEAD   1    2    7   14   19   26   TAIL
 *
 * </pre>
 */
template<typename T> 
class Skiplist {
  /** Type of the node */
  typedef Node<T> NodeType;

  /** Type of lengths */
  typedef size_t SizeType;

  typedef short LevelType;

  /** Current TrailType type */
  typedef trail::Trail<T> TrailType;

 public:
  /** Type of the iterator */
  typedef skiplist::ConstIterator<T> iterator;

  /**
   * Constructor, creates empty skiplist
   */
  Skiplist(Rand rand_):rand(rand_) {
    head = new NodeType(MAXLEVEL);
    tail = new NodeType(MAXLEVEL);
    size_ = 0;
    for (LevelType i = 0; i < MAXLEVEL; i++) {
      head->forward[i] = tail;
      head->forward_length[i] = 1;
    }
    tail->previous = head;
  }

  /**
   * Destructor
   */
  ~Skiplist() {
    NodeType* node = head;
    while (node != NULL) {
      NodeType* tmp = node->next();
      delete node;
      node = tmp;
    }
  }


  /**
   * Iterator to the first element
   */
  iterator begin() const {
    return iterator(head->next());
  }

  /**
   * Iterator to the element after the last element
   */
  iterator end() const {
    return iterator(tail);
  }


  /**
   * @returns trail to element determined by a TrailFunction<T>
   */
  TrailType generic_trail(trail::TrailFunction<T>* function) const {
    TrailType trail;
    SizeType position = 0;
    NodeType* node = head;
    for (LevelType i = MAXLEVEL - 1; i >= 0; i--) {
      // invariant: function(current_node, cnt) is true 
      while ((node->forward[i] != tail) &&
          (function->goFurther(node->forward[i],
                               position + node->forward_length[i]))) {
        position += node->forward_length[i];
        node = node->forward[i];
      }
      trail.node[i] = node;
      trail.position[i] = position;
    }
    return trail;
  }

  /**
   * returns iterator to last node which is < value
   */
  iterator lower_bound(const T& value) const {
    trail::LowerBoundTrailFunction<T> funct(value);
    TrailType trail = generic_trail(&funct);
    return iterator(trail.node[0]->next());
  }

  /**
   * returns iterator to first node which is > value
   */
  iterator upper_bound(const T& value) const {
    trail::UpperBoundTrailFunction<T> funct(value);
    TrailType trail = generic_trail(&funct);
    return iterator(trail.node[0]->next());
  }

  /**
   * @returns iterator to first element equals to \a value
   * or end() if not found
   */
  iterator find(const T& value) const {
    iterator it = lower_bound(value);
    if (*it != value) return end();
    return it;
  }

  /**
   * inserts a new element into a skiplist
   *
   * @returns iterator to newly inserted element
   */
  iterator insert(const T& value) {
    trail::LowerBoundTrailFunction<T> funct(value);
    TrailType trail = generic_trail(&funct);
    SizeType pos = trail.position[0]+1;

    // create a new node
    NodeType* node = new NodeType(node_utils::randomLevel(&rand, LEVELUP_PROB,
          MAXLEVEL));
    node->value = value;
    node->previous = trail.node[0];
    trail.node[0]->forward[0]->previous = node;

    // set forward links of the newly created node
    // and insert it into lists
    for (LevelType i = 0; i < node->level; i++) {
      node->forward[i] = trail.node[i]->forward[i];
      node->forward_length[i] = trail.node[i]->forward_length[i] +
        trail.position[i] - pos + 1;
      trail.node[i]->forward[i] = node;
      trail.node[i]->forward_length[i] = pos - trail.position[i];
    }
    // set update forward lengths of higher levels
    for (LevelType i = node->level; i < MAXLEVEL; i++) {
      trail.node[i]->forward_length[i]++;
    }
    size_++;
    return iterator(node);
  }

  /**
   * @returns distance of the \a node from the start
   */
  SizeType nodePosition(const NodeType* node) const {
    SizeType distance_end = 0;
    while (node != tail) {
      assert(node->level > 0);
      LevelType top = node->level - 1;
      distance_end += node->forward_length[top];
      node = node->forward[top];
    }
    return size_ - distance_end;
  }

  /**
   * @returns distance of the iterator from the start of the list
   */
  SizeType xth(iterator it) const {
    return nodePosition(it.getNode());
  }

  /**
   * Remove element from skiplist
   *
   * @precondition \a it should be a valid iterator
   * @post only iterators pointing to the deleted element will be invalidated
   *
   * @param it iterator
   *
   */
  void erase(iterator it) {
    Preconditions::check(it != end(), "can't delete end()");
    NodeType* node = it.getNode();

    // get the full trail for the node
    trail::KthTrailFunction<T> funct(nodePosition(node));
    TrailType trail = generic_trail(&funct);
    assert(trail.node[0]->next() == node);

    NodeType* prev = trail.node[0];
    NodeType* next = node->next();

    next->previous = prev;
    // remove node at levels
    for (int i=0; i < node->level; i++) {
      trail.node[i]->forward[i] = node->forward[i];
      trail.node[i]->forward_length[i] += node->forward_length[i] - 1;
    };
    // update lengths of higher levels
    for (int i = node->level; i < MAXLEVEL; i++) {
      trail.node[i]->forward_length[i]--;
    }

    delete node;
    size_--;
  }


  /**
   * Returns iterator to the k-th element of the skiplist
   *
   * @precondition <i> 0 <= k < size() </i>
   */
  iterator kth(SizeType k) const {
    Preconditions::checkRange(k, size());
    trail::KthTrailFunction<T> funct(k);
    TrailType trail = generic_trail(&funct);
    return iterator(trail.node[0]->next());
  }

  /**
   * @returns number of elements in the skiplist
   */
  SizeType size() const {
    return size_;
  }

 private:

  /**
   * Head of the skiplist, does not contain data
   */
  NodeType* head;

  /**
   * Tail of the skiplist, does not contain data
   */
  NodeType* tail;

  /**
   * size of the skiplist
   */
  SizeType size_;

  /**
   * Random generator used for node level generation
   */
  Rand rand;

  DISALLOW_EVIL_CONSTRUCTORS(Skiplist);
};


} // namespace skiplist
} // namespace balanced_structures
