#include <algorithm>
#include "gtest/gtest.h"
#include <set>
#include "skiplist.h"
#include "template/template.h"
#include "utils/benchmark/benchmark.h"
#include "utils/macros/foreach.h"
#include "utils/memory_usage/memory_usage.h"
#include "utils/rand/rand.h"

namespace balanced_structures {
namespace skiplist {

class SkiplistBenchmark: public ::testing::Test {
 public:
  template <class SetLike> 
  void insertValues(const std::vector<int>& values, SetLike* set) {
    FOREACH(it, values) {
      set->insert(*it);
    }
  }

  template <class SetLike> 
  void deleteValues(const std::vector<int>& values, SetLike* set) {
    FOREACH(it, values) {
      set->erase(set->find(*it));
    }
  }

  template <class SetLike> 
  void findValues(const std::vector<int>& values, SetLike* set) {
    FOREACH(it, values) {
      set->find(*it);
    }
  }

  void kth(const Skiplist<int>& list) {
    int result = 0; // disallow optimizations
    for (size_t i = 0; i < list.size(); i++) {
      result += *list.kth(i);
    }
    printf("disallow_optimizations: %d\n", result);;
  }

  void xth(const Skiplist<int>& list) {
    int result = 0; // disallow optimizations
    FOREACH(it, list) {
      result += list.xth(it);
    }
    printf("disallow_optimizations: %d\n", result);;
  }



  std::vector<int> getData() {
    const int MAX = 1000000000;
    const int CNT = 2500000;
    Rand dataRand = Rand(111);
    for (int i = 0; i < CNT; i++) {
      data.push_back(dataRand.rand(0, MAX));
    }
    return data;
  }

  virtual void SetUp() {
    data = getData();
  }

  SkiplistBenchmark():list(Rand(42)) {
  };

  Skiplist<int> list;
  std::multiset<int> reference;
  std::vector<int> data;
};

TEST_F(SkiplistBenchmark, memoryUsage) {
  printf("Estimating memory usage:\n");
  int mem_start = utils::memory_usage::getUsedMemoryKb();
  BENCHMARK(1, insertValues(data, &list));
  int mem1 = utils::memory_usage::getUsedMemoryKb();
  BENCHMARK(1, insertValues(data, &reference));
  int mem2 = utils::memory_usage::getUsedMemoryKb();

  int size = list.size();
  printf("memory consumption per 1 item for %d items\n", size);
  printf("skiplist: %.1f\n", (mem1 - mem_start) * 1024.0 / size);
  printf("set: %.1f\n", (mem2 - mem1) * 1024.0 / size);
}

TEST_F(SkiplistBenchmark, randomInsertDelete) {
  printf("inserting random values\n");
  BENCHMARK(1, insertValues(data, &list));
  BENCHMARK(1, insertValues(data, &reference));
  
  BENCHMARK(1, deleteValues(data, &list));
  BENCHMARK(1, deleteValues(data, &reference));
}

TEST_F(SkiplistBenchmark, ascendingInsertDelete) {
  printf("inserting sorted values\n");
  std::sort(data.begin(), data.end());
  BENCHMARK(1, insertValues(data, &list));
  BENCHMARK(1, insertValues(data, &reference));

  BENCHMARK(1, deleteValues(data, &list));
  BENCHMARK(1, deleteValues(data, &reference));
}

TEST_F(SkiplistBenchmark, descendingInsertDelete) {
  printf("inserting reversed values\n");
  std::sort(data.begin(), data.end());
  std::reverse(data.begin(), data.end());
  BENCHMARK(1, insertValues(data, &list));
  BENCHMARK(1, insertValues(data, &reference));

  BENCHMARK(1, deleteValues(data, &list));
  BENCHMARK(1, deleteValues(data, &reference));
}

TEST_F(SkiplistBenchmark, ascendingInsertRandomDelete) {
  printf("inserting sorted values\n");
  std::sort(data.begin(), data.end());
  BENCHMARK(1, insertValues(data, &list));
  BENCHMARK(1, insertValues(data, &reference));
  std::random_shuffle(data.begin(), data.end());
  BENCHMARK(1, deleteValues(data, &list));
  BENCHMARK(1, deleteValues(data, &reference));
}

TEST_F(SkiplistBenchmark, findRandom) {
  printf("inserting values\n");
  BENCHMARK(1, insertValues(data, &list));
  BENCHMARK(1, insertValues(data, &reference));
  std::random_shuffle(data.begin(), data.end());
  BENCHMARK(1, findValues(data, &list));
  BENCHMARK(1, findValues(data, &reference));
}

TEST_F(SkiplistBenchmark, kth) {
  printf("inserting values\n");
  BENCHMARK(1, insertValues(data, &list));
  BENCHMARK(1, kth(list));
}

TEST_F(SkiplistBenchmark, xth) {
  printf("inserting values\n");
  BENCHMARK(1, insertValues(data, &list));
  BENCHMARK(1, xth(list));
}

} // namespace skiplist
} // namespace balanced_structures
