#ifndef H_BALANCED_STRUCTURES_SKIPLIST_SKIPLIST_ITERATOR
#define H_BALANCED_STRUCTURES_SKIPLIST_SKIPLIST_ITERATOR
/**
 * @file Contains Skiplist iterator class
 */
#include <iterator>

namespace balanced_structures {
namespace skiplist {

// Forward definition of Node
template<typename T>
class Node;

/**
 * (Bidirectional) iterator over skiplist nodes.
 */
template <typename T>
struct ConstIterator : public std::iterator<std::bidirectional_iterator_tag, T> {
  typedef const T value_type;
  typedef const T& reference;
  typedef const T* pointer;
  typedef ConstIterator<T> self;

  /**
   * Constructor
   */
  ConstIterator(): node(NULL) {
  }

  /**
   * Constructor, iterator pointing to specified node
   */
  explicit ConstIterator(Node<T>* x) :
    node(x) {
  }

  /**
   * dereferences an iterator, returns value
   */
  reference operator*() const {
    Preconditions::checkNotNull(node);
    return node->value;
  }

  /**
   * @returns corresponding skiplist node
   */
  Node<T>* getNode() {
    return node;
  }

  /**
   * pre-increment \a ++iterator
   */
  self& operator++() {
    Preconditions::checkNotNull(node);
    node = node->next();
    return *this;
  }

  /**
   * post-increment \a iterator++
   */
  self operator++(int) {
    Preconditions::checkNotNull(node);
    self tmp = *this;
    node = node->next();
    return tmp;
  }

  /**
   * pre-decrement \a --iterator
   */
  self& operator--() {
    Preconditions::checkNotNull(node);
    node = node->prev();
    return *this;
  }

  /**
   * post-decrement \a iterator--
   */
  self operator--(int) {
    Preconditions::checkNotNull(node);
    self tmp = *this;
    node = node->prev();
    return tmp;
  }

  /**
   * @returns true if the iterators are pointing to the same element
   */
  bool operator==(const self& x) const {
    return node == x.node;
  }

  /**
   * @returns true iff these two iterators are pointing to different elements
   */
  bool operator!=(const self& x) const {
    return !(*this == x);
  }

 private:
  Node<T>* node;
};

} // namespace skiplist
} // namespace balanced_structures
#endif
