#ifndef H_BALANCED_STRUCTURES_SKIPLIST_SKIPLIST_NODE
#define H_BALANCED_STRUCTURES_SKIPLIST_SKIPLIST_NODE

/**
 * @file Contains skiplist node implementation.
 */

#include "skiplist_utils.h"

namespace balanced_structures {
namespace skiplist {

typedef short LevelType;

/**
 * <i> LEVELUP_PROB </i> is probability of "increasing" level
 * of a node in the percent. Thus, node of level <i> L </i>
 * is created with probability
 * <i> LEVELUP_PROB^(L - 1) (1 - LEVELUP_PROB) </i>.
 * The average level of nodes in skiplist will be
 * <i> 1 / 0LEVELUP_PROB </i>
 */
static const int LEVELUP_PROB = 100 / 4; // 25%

/**
 * Maximum allowed level of a node.
 *
 * Note: current value of 15 is pretty good upper bound if
 * you are using \a LEVELUP_PROB = 4 and less than 10^8
 * nodes.
 */
static const LevelType MAXLEVEL = 15;

/**
 * Skiplist node
 */
template<typename T>
class Node {
  /** Type of position in skiplist */
  typedef int SizeType;

  /** Type of this object */
  typedef Node<T> Self;

 public:

  Node(SizeType level_):value(),level(level_),
      previous(NULL) {
    Preconditions::check(level >= 1);
    Preconditions::check(level <= MAXLEVEL);
    forward = new Self*[level];
    forward_length = new SizeType[level];
    for (LevelType i = 0; i < level; i++) {
      forward[i] = NULL;
      forward_length[i] = 1;
    }
  }

  /**
   * Return next node
   *
   * If this node is last node (skiplist 'tail'), it will
   * return NULL
   *
   * @returns pointer to the next node
   */
  Self* next() const {
    return forward[0];
  }

  /**
   * Return previous node
   *
   * If this node is first node (skiplist 'head'), it will
   * return NULL
   *
   * @returns pointer to the next node
   */
  Self* prev() const {
    return previous;
  }

  /**
   * Destructor
   */
  ~Node() {
    delete forward;
    delete forward_length;
  }

  /**
   * A value this node is holding.
   */
  T value;

  /**
   * Level of the node.
   *
   * Also size of the \a forward and \a forward_length arrays
   */
  LevelType level;

  /**
   * Forward links to next nodes.
   */
  Self** forward;

  /**
   * Lengths of forward links.
   */
  SizeType* forward_length;

  /**
   * Backward link.
   */
  Self* previous;
};

} // namespace skiplist
} // namespace balanced_structures
#endif
