#ifndef H_BALANCED_STRUCTURES_SKIPLIST_SKIPLIST_TRAIL
#define H_BALANCED_STRUCTURES_SKIPLIST_SKIPLIST_TRAIL

#include "skiplist_node.h"
#include "utils/macros/unused.h"

namespace balanced_structures {
namespace skiplist {
namespace trail {

typedef size_t SizeType;

/**
 * Trail is a collection of nodes directly preceding
 * the node we selected.
 *
 * The trail holds useful information for insertions and deletions.
 * <pre>
 * Example trail for node 19 (denoted @):
 *
 *   # -----------> @ ----------------> #
 *   # -> # ------> @ -----------> # -> #
 *   # -> # ------> # -> @ ------> # -> #
 *   # -> # -> # -> # -> # -> @ -> # -> #
 * HEAD   1    2    7   14   19   26   TAIL
 *
 * Trail nodes: 19, 14, 7, 7
 * Trail position: 5, 4, 3, 3
 * </pre>
 *
 */
template <typename T>
struct Trail {
  Node<T>* node[MAXLEVEL];
  SizeType position[MAXLEVEL];
};

/**
 * Trail function is a function which determines the node
 * which should be a result of the search.
 *
 * @precondition The function should be a binary predicate
 * returning True on all elements less/equal than the correct one,
 * and returning False on all elements greater.
 */
template <typename T>
class TrailFunction {
 public:
  /**
   * Determine if the search should continue
   *
   * @param node pointer to the current node
   * @param position distance of the current node from the start of the skiplist
   *
   * @returns true if the search should continue
   */
  virtual bool goFurther(const Node<T>* node, SizeType position)=0;

  /**
   * Destructor
   */
  virtual ~TrailFunction(){};
};

/**
 * Trail function which points to the first position
 * where element can be inserted without violating ordering.
 *
 * Note that this is equivalent to the first element >= \a value
 *
 */
template <typename T>
class LowerBoundTrailFunction : public TrailFunction<T> {
public:
  /**
   * Constructor
   */
  LowerBoundTrailFunction(const T& value_):value(value_) { 
  }

  /**
   * Goes further until we encounter a \a node with value greater or equal than
   * \a value
   */
  virtual bool goFurther(const Node<T>* node, SizeType UNUSED(position)) {
    return node->value < value;
  }

 private:
  /**
   * Value we are searching for
   */
  T value;
};

template <typename T>
class UpperBoundTrailFunction : public TrailFunction<T> {
public:
  UpperBoundTrailFunction(const T& value_):value(value_) { 
  }

  /**
   * Goes further until we encounter a \a node with value greater than \a value
   */
  virtual bool goFurther(const Node<T>* node, SizeType UNUSED(position)) {
    return node->value <= value;
  }

 private:
  /**
   * Value we are searching for
   */
  T value;
};

/**
 * Find k-th element
 */
template <typename T>
class KthTrailFunction : public TrailFunction<T> {
 public:
  /**
   * Constructor
   */
  KthTrailFunction(SizeType pos_):pos(pos_) { 
  }

  /**
   * goes further until we encounter node with specified position
   */
  virtual bool goFurther(const Node<T>* UNUSED(node), SizeType position) {
    return position <= pos;
  }

 private:
  /**
   * index of an element we are searching for
   */
  SizeType pos;
};

} // trail
} // skiplist
} // balanced_structures
#endif
