#include "gtest/gtest.h"
#include "skiplist.h"
#include "template/template.h"
#include "debug/ppdebug.h"
#include "utils/rand/rand.h"

namespace balanced_structures {
namespace skiplist {

class SkiplistTest : public testing::Test {
 public:
  static const int RANDOM_LEVEL_ITERATIONS = 1000000;
};

TEST_F(SkiplistTest, printRandomLevelDistribution) {
  const int MAXLEVEL = 15;
  Rand rand(47);
  int counts[MAXLEVEL+1];
  FOR(q, MAXLEVEL+1) {
    counts[q] = 0;
  }

  FOR(i, RANDOM_LEVEL_ITERATIONS) {
    int level = node_utils::randomLevel(&rand, LEVELUP_PROB, MAXLEVEL);
    ASSERT_TRUE(level > 0);
    ASSERT_TRUE(level <= MAXLEVEL);
    counts[level]++;
  }

  D(mp(counts, counts + MAXLEVEL));
}

TEST_F(SkiplistTest, iteratorsEmpty) {
  Skiplist<int> list(Rand(47));
  ASSERT_EQ(list.begin(), list.end());
}

TEST_F(SkiplistTest, iteratorsNonEmpty) {
  Skiplist<int> list(Rand(47));
  list.insert(10);
  ASSERT_NE(list.begin(), list.end());
}

TEST_F(SkiplistTest, iteratorsOnInsert) {
  Skiplist<int> list(Rand(47));
  list.insert(10);
  list.insert(5);
  list.insert(20);
  list.insert(7);
  Skiplist<int>::iterator it = list.begin();
  ASSERT_EQ(5, *it++);
  ASSERT_EQ(7, *it++);
  ASSERT_EQ(10, *it++);
  ASSERT_EQ(20, *it++);
  ASSERT_EQ(it, list.end());
}

TEST_F(SkiplistTest, kthOnInsert) {
  Skiplist<int> list(Rand(42));
  list.insert(10);
  list.insert(5);
  list.insert(20);
  list.insert(7);
  ASSERT_EQ(5, *list.kth(0));
  ASSERT_EQ(7, *list.kth(1));
  ASSERT_EQ(10, *list.kth(2));
  ASSERT_EQ(20, *list.kth(3));
  ASSERT_EQ((size_t) 4, list.size());
}

TEST_F(SkiplistTest, remove) {
  Skiplist<int> list(Rand(42));
  list.insert(10);
  list.insert(5);
  list.insert(20);
  list.insert(7);
  list.erase(list.kth(2));
  ASSERT_EQ((size_t) 3, list.size());
  ASSERT_EQ(5, *list.kth(0));
  ASSERT_EQ(7, *list.kth(1));
  ASSERT_EQ(20, *list.kth(2));
}

} // namespace skiplist
} // namespace balanced_structures
