#ifndef H_GEOMETRY_TWO_D_ANGLE
#define H_GEOMETRY_TWO_D_ANGLE
/**
 * @file Contains useful function for angle comparison
 */

#include "signum.h"
#include "point.h"

namespace geometry {
namespace two_d {

/**
 * Quadrant of an point
 *
 * Note that quadrants are ordered with increasing angle starting
 * with top right quadrant = 1
 */
enum Quadrant {
  /** Point is in the center */
  CENTER = 0,
  /** Up right quadrant, including x-axis, excluding y-axis */
  TOP_RIGHT = 1,
  /** Up left quadrant, including y-axis, excluding x-axis */
  TOP_LEFT = 2,
  /** Down left quadrant, including x-axis, excluding y-axis */
  BOTTOM_LEFT = 3,
  /** Down right quadrant, including y-axis, excluding x-axis */
  BOTTOM_RIGHT = 4,
}; 

template <typename T>
Quadrant getQuadrant(const Point<T> point) {
  int a = 1 + signum(point.x());
  int b = 1 + signum(point.y());

  // check if signum is correct
  assert(0 <= a && a <= 2);
  assert(0 <= b && b <= 2);

  Quadrant data[3][3] = {
    {BOTTOM_LEFT,  BOTTOM_LEFT, TOP_LEFT},
    {BOTTOM_RIGHT,   CENTER,    TOP_LEFT},
    {BOTTOM_RIGHT,  TOP_RIGHT, TOP_RIGHT}
  };

  return data[a][b];
}

/**
 * Compare two vectors by angle with x-axis
 *
 * @returns true if angle of the first vector is less than angle of the second
 */
template <typename T>
bool angleLess(const Point<T> point1, const Point<T> point2) {
  Quadrant q1 = getQuadrant(point1);
  Quadrant q2 = getQuadrant(point2);
  if (q1 != q2) {
    return q1 < q2;
  }
  // same quadrant, we can decide by cross product.
  return point1.cross(point2) > 0;
}

} // namespace two_d
} // namespace geometry

#endif
