#ifndef H_GEOMETRY_TWO_D_CONVEX_HULL
#define H_GEOMETRY_TWO_D_CONVEX_HULL
/**
 * @file Contains class capable of computing the convex hull of
 * a set of points.
 */

#include "point.h"
#include <vector>
#include <algorithm>
#include "utils/preconditions/preconditions.h"
#include <cassert>

namespace geometry {
namespace two_d {

/**
 * Class computing the convex hull of a set of points
 */
template <typename T>
class ConvexHull {
 public:
  typedef Point<T> PointType;

  /**
   * Removes all points.
   */
  void clear() {
    data.clear();
  }

  /**
   * Add a new point.
   */
  void addPoint(const PointType point) {
    data.push_back(point);
  }

  /**
   * Compute convex hull
   *
   * Note: Convex hull of an empty set is an empty set.
   */
  std::vector<PointType> convexHull() {
    if (data.size() < 2) {
      // one point is always in the convex hull
      return data;
    }

    std::sort(data.begin(), data.end(), PointCompare());
    std::vector<PointType> lower_chain = computeChain(data);

    std::reverse(data.begin(), data.end());
    std::vector<PointType> upper_chain =
      rotate180(computeChain(rotate180(data)));

    assert(upper_chain.size() > 0);
    if (upper_chain.size() == 1) {
      assert(lower_chain.size() == 1);
      return lower_chain;
    }
    lower_chain.insert(lower_chain.end(), upper_chain.begin() + 1, upper_chain.end() - 1);
    return lower_chain;
  }

 private:

  /**
   * Returns monotone chain
   *
   * Chain starts at leftmost point (in case of more leftmost points the
   *  one with smallest y coordinate), it is upper half of the convex
   *  hull and ends in rightmost (in case of tie the topmost one of them) point
   *
   * @precondition data sorted ascending by \a x coordinate, then \a y coordinate
   *
   * @returns upper monotone chain
   */
  std::vector<PointType> computeChain(const std::vector<PointType> data) {
    Preconditions::check(data.size() > 1);
    std::vector<PointType> chain;
    chain.push_back(data[0]);
    typedef typename std::vector<PointType>::size_type SizeType;
    SizeType pos = 1;
    while ((pos < data.size()) && 
           (data[0] == data[pos])) {
      pos++;
    }
    if (pos == data.size()) {
      return chain; // only 1 unique point and we were strict
    }
    chain.push_back(data[pos]);
    for (SizeType i = pos + 1; i < data.size(); i++) {
      PointType current = data[i];
      while (chain.size() >= 2) {
        PointType last1 = chain[chain.size() - 1];
        PointType last2 = chain[chain.size() - 2];

        if ((current - last1).cross(last1 - last2) > 0) {
          break;
        }
        chain.pop_back();
      }
      chain.push_back(current);
    }

    return chain;
  }

  /**
   * Rotates all points by 180 degrees
   */
  std::vector<PointType> rotate180(const std::vector<PointType> data) {
    std::vector<PointType> result;
    typedef typename std::vector<PointType>::size_type SizeType;
    for (SizeType i = 0; i < data.size(); i++) {
      PointType current = data[i];
      result.push_back(PointType(-current.x(), -current.y()));
    }
    return result;
  }

  /**
   * Comparison function.
   *
   * Points are sorted according \a x coordinate, than \a y.
   */
  class PointCompare {
   public:
    bool operator()(const PointType& p1, const PointType& p2) {
      if (p1.x() != p2.x()) {
        return p1.x() < p2.x();
      }
      return (p1.y() < p2.y());
    }
  };

  /**
   * inserted points.
   */
  std::vector<PointType> data;
};

} // namespace two_d
} // namespace geometry
#endif
