#include "gtest/gtest.h"
#include "convex_hull.h"
#include "point.h"
#include <vector>

namespace geometry {
namespace two_d {

typedef Point<int> Pt;
typedef ConvexHull<int> Ch;
typedef std::vector<Pt> vp;

TEST(ConvexHull, empty)
{
  Ch hull;
  EXPECT_EQ(0u, hull.convexHull().size());

}

TEST(ConvexHull, clear) {
  Ch hull;
  hull.addPoint(Pt(0, 0));
  EXPECT_EQ(1u, hull.convexHull().size());
  hull.clear();
  EXPECT_EQ(0u, hull.convexHull().size());
}

TEST(ConvexHull, duplicatesOnePoint) {
  Ch hull;
  hull.addPoint(Pt(0, 0));
  hull.addPoint(Pt(0, 0));
  hull.addPoint(Pt(0, 0));
  hull.addPoint(Pt(0, 0));
  EXPECT_EQ(1u, hull.convexHull().size());
}

TEST(ConvexHull, duplicatesTwoPoints) {
  Ch hull;
  hull.addPoint(Pt(0, 0));
  hull.addPoint(Pt(0, 0));
  hull.addPoint(Pt(0, 0));
  hull.addPoint(Pt(0, 0));
  hull.addPoint(Pt(0, 1));
  EXPECT_EQ(2u, hull.convexHull().size());
  hull.clear();
  hull.addPoint(Pt(0, 0));
  hull.addPoint(Pt(0, 1));
  hull.addPoint(Pt(0, 0));
  hull.addPoint(Pt(0, 1));
  EXPECT_EQ(2u, hull.convexHull().size());
}

TEST(ConvexHull, line) {
  Ch hull;
  hull.addPoint(Pt(0, 0));
  hull.addPoint(Pt(0, 1));
  hull.addPoint(Pt(0, 2));
  hull.addPoint(Pt(0, 3));
  hull.addPoint(Pt(0, -3));
  hull.addPoint(Pt(0, -2));
  vp res = hull.convexHull();
  EXPECT_EQ(2u, res.size());
  EXPECT_EQ(Pt(0, -3), res[0]);
  EXPECT_EQ(Pt(0, 3), res[1]);
}

TEST(ConvexHull, square) {
  Ch hull;
  hull.addPoint(Pt(0, -1));
  hull.addPoint(Pt(0, 0));
  hull.addPoint(Pt(0, 1));
  hull.addPoint(Pt(1, -1));
  hull.addPoint(Pt(1, 0));
  hull.addPoint(Pt(1, 1));
  hull.addPoint(Pt(-1, -1));
  hull.addPoint(Pt(-1, 0));
  hull.addPoint(Pt(-1, 1));
  vp expected;
  expected.push_back(Pt(-1, -1));
  expected.push_back(Pt(-1, 1));
  expected.push_back(Pt(1, 1));
  expected.push_back(Pt(1, -1));
  EXPECT_EQ(expected, hull.convexHull());
}

TEST(ConvexHull, sortCircle) {
  vp data;
  const int RADIUS = 40000;
  const int POINTS = 500;

  for (int i = 0; i < POINTS; i++) {
    data.push_back(Pt((int) (-RADIUS * cos(2 * M_PI / POINTS * i)),
                      (int) (RADIUS * sin(2 * M_PI / POINTS * i))));
  }

  vp tmp = data;
  std::random_shuffle(tmp.begin(), tmp.end());
  Ch hull;
  for (int i = 0; i < POINTS; i++) {
    hull.addPoint(tmp[i]);
    hull.addPoint(tmp[i]);
  }

  EXPECT_EQ(data, hull.convexHull());

}

} // namespace two_d
} // namespace geometry

