#ifndef H_GEOMETRY_TWO_D_DISTANCE
#define H_GEOMETRY_TWO_D_DISTANCE

#include "math/rational/rational.h"
#include "linesegment.h"

namespace geometry {
namespace two_d {

/**
 * Calculate square of the distance between two points.
 *
 * @returns square of the distance.
 */
template<typename T>
  T sqrDistancePointPoint(Point<T> p1, Point<T> p2) {
    Point<T> p = p1 - p2;
    return p.dot(p);
  }

/**
 * Calculate distance between two points
 *
 * @returns distance
 */
template<typename T>
  long double distancePointPoint(Point<T> p1, Point<T> p2) {
    Point<long double> p = p1 - p2;
    return sqrt(p.dot(p));
  }

/**
 * Calculate sqaure of distance between point and line.
 *
 * @returns square of the distance
 */
template<typename T>
  math::rational::Rational<T> sqrDistancePointLine(Point<T> p, LineSegment<T> line) {
    Point<T> lineVector = line.end - line.begin;
    Point<T> relativeP = p - line.begin;

    T pp = relativeP.dot(relativeP);
    T lp = relativeP.dot(lineVector);
    T ll = lineVector.dot(lineVector);

    math::rational::Rational<T> result(lp);
    return math::rational::Rational<T>(pp) - math::rational::Rational<T>(lp, ll) *
      math::rational::Rational<T>(lp);
  }

/**
 * Calculate distance between point and line.
 *
 * Note: We reuse code from sqrDistancePointLine() because
 * we do not need to compute fraction value and the whole Rational
 * class (and we won't have problems with overflows).
 *
 * @returns distance
 */
template<typename T>
  long double distancePointLine(Point<T> p, LineSegment<T> line) {
    Point<long double> lineVector(line.end - line.begin);
    Point<long double> relativeP(p - line.begin);

    long double pp = relativeP.dot(relativeP);
    long double lp = relativeP.dot(lineVector);
    long double ll = lineVector.dot(lineVector);

    return sqrt(pp - lp * lp / ll);
  }

/**
 * Calculate square of the distance between point and line segment.
 *
 * @returns square of the distance
 */
template<typename T>
  math::rational::Rational<T> sqrDistancePointLineSegment(Point<T> p, LineSegment<T> line) {
    Point<T> lineVector = line.end - line.begin;
    Point<T> relativeP = p - line.begin;

    T lp = relativeP.dot(lineVector);
    T ll = lineVector.dot(lineVector);
    if (lp <= 0) {
      return sqrDistancePointPoint(p, line.begin);
    }
    if (lp >= ll) {
      return sqrDistancePointPoint(p, line.end);
    }

    return sqrDistancePointLine(p, line);
  }

/**
 * Calculate distance between point and line segment.
 *
 * Note: We don't reuse sqrDistancePointPoint because this
 * whole function can be calculated in long doubles and won't
 * have problems with Rational overflows.
 *
 * @returns distance
 */
template<typename T>
  long double distancePointLineSegment(Point<T> p, LineSegment<T> line) {
    Point<long double> lineVector = line.end - line.begin;
    Point<long double> relativeP = p - line.begin;

    long double lp = relativeP.dot(lineVector);
    long double ll = lineVector.dot(lineVector);
    if (lp <= 0) {
      return distancePointPoint(p, line.begin);
    }
    if (lp >= ll) {
      return distancePointPoint(p, line.end);
    }

    return distancePointLine(p, line);
  }

} // namespace two_w
} // namespace geometry
#endif
