#ifndef H_GEOMETRY_TWO_D_INTERSECT
#define H_GEOMETRY_TWO_D_INTERSECT
#include "utils/preconditions/preconditions.h"
#include "signum.h"
#include <algorithm>
#include "linesegment.h"

namespace geometry {
namespace two_d {

/**
 * Determine whether point lies on line.
 * Warning: On float numbers use distanceToLineSegment
 * instead and check for zero with epsilon error.
 * Warning: On integer types, be sure to compute this in
 * double-sized type.
 */
template<typename T>
bool pointOnLine(Point<T> p, LineSegment<T> s) {
  Preconditions::check(s.begin != s.end,
      "Line ends can't be same!");
  return (s.end - s.begin).cross(p - s.begin) == 0; 
}

/**
 * Determine whether point lies on line.
 * Warning: On float numbers use distanceToLineSegment
 * instead and check for zero with epsilon error.
 * Warning: On integer types, be sure to compute this in
 * double-sized type.
 */
template<typename T>
bool pointOnLineSegment(Point<T> p, LineSegment<T> s, bool
    acceptCorners) {
  if (s.begin == s.end) {
    return acceptCorners && (s.begin == p);
  }
  if (!pointOnLine(p, s)) {
    return false;
  }
  // and now check we are really in segment
  Point<T> lineVector = s.end - s.begin;
  T t = lineVector.dot(p - s.begin);
  T l = lineVector.dot(lineVector);
  if (acceptCorners) {
    return (t>=0 && t<= l);
  } else {
    return (t>0 && t<l);
  }
}

enum IntersectType {
  NO_INTERSECT, // no intersection
  INTERSECT, // single intersection inside objects
  TANGENCY, // objects are touching
  OVERLAY, // line/area intersection of objects
};

/**
 * Determine whether two closed intervals intersects.
 * Note that we allow any ordering on interval ranges;
 */
template<typename T>
IntersectType intervalIntersect(T a1, T a2, T b1, T b2) {
  if (a1 > a2) std::swap(a1,a2);
  if (b1 > b2) std::swap(b1,b2);

  if (a2 < b1 || b2 < a1) return NO_INTERSECT;

  // we have interval and point or two points and they are in
  // overlay
  if (a1 == a2 || b1 == b2) return INTERSECT;
  if (a2 == b1 || b2 == a1) return TANGENCY;
  return OVERLAY;
}

/**
 * Determine whether line intersects with line segment.
 * Warning: On integer types, be sure to compute this in
 * double-sized type.
 */
template<typename T>
IntersectType intersectLineLineSegment(const LineSegment<T>& line,
    const LineSegment<T>& segment) {
  Preconditions::check(line.begin != line.end);
  Preconditions::check(segment.begin != segment.end);

  Point<T> lineVector = line.end - line.begin;
  int t1 = signum(lineVector.cross(segment.begin - line.begin));
  int t2 = signum(lineVector.cross(segment.end - line.begin));
  if (t1 == 0 && t2 == 0) {
    return OVERLAY;
  }
  if (t1 == 0 || t2 == 0) {
    return TANGENCY;
  }
  if (t1 * t2 < 0) {
    return INTERSECT;
  }
  return NO_INTERSECT;
}

/**
 * Determine whether two line segments intersects.
 * Warning: On integer types, be sure to compute this in
 * double-sized type.
 */
template<typename T>
IntersectType intersectLineSegmentLineSegment(const LineSegment<T>& segment1,
    const LineSegment<T>& segment2) {
  Preconditions::check(segment1.begin != segment1.end);
  Preconditions::check(segment2.begin != segment2.end);

  IntersectType i1 = intersectLineLineSegment(segment1, segment2);
  IntersectType i2 = intersectLineLineSegment(segment2, segment1);

  if (i1 == NO_INTERSECT || i2 == NO_INTERSECT) {
    return NO_INTERSECT;
  }
  // i1,i2 can be INTERSECT, OVERLAY, TANGENCY
  if ((i1 == INTERSECT && i2 == TANGENCY) ||
      (i2 == INTERSECT && i1 == TANGENCY)) {
    return TANGENCY;
  }
  // now we expect only same values
  assert(i1 == i2);

  if (i1 == INTERSECT) {
    return INTERSECT;
  }
  if (i1 == TANGENCY) {
    return TANGENCY;
  }

  // The hard case, we have two times overlay, which means
  // that both segments are part of the same line
  if ((segment1.end - segment1.begin).x() == 0) {
    return intervalIntersect(
        segment1.begin.y(), segment1.end.y(),
        segment2.begin.y(), segment2.end.y());
  } else {
    return intervalIntersect(
        segment1.begin.y(), segment1.end.y(),
        segment2.begin.y(), segment2.end.y());
  }
}

} // namespace two_d
} // namespace geometry
#endif
