#include "gtest/gtest.h"
#include "point.h"
#include "intersect.h"

namespace geometry {
namespace two_d {

TEST(Intersect, LinePoint)
{
  Point<int> a(2,5);
  Point<int> b(3,7);
  LineSegment<int> l(a,b);
  EXPECT_EQ(true, pointOnLine(a, l));
  EXPECT_EQ(true, pointOnLine(b, l));
  EXPECT_EQ(true, pointOnLine(a + (b-a)*2, l));
  EXPECT_EQ(true, pointOnLine(a - (b-a)*5, l));
  EXPECT_FALSE(pointOnLine(Point<int>(0,0), l));
  EXPECT_FALSE(pointOnLine(Point<int>(1,0), l));
  EXPECT_EQ(true, pointOnLine(Point<int>(0,1), l));
  EXPECT_FALSE(pointOnLine(Point<int>(1,1), l));
  EXPECT_FALSE(pointOnLine(Point<int>(2000,40001), l));
}

// we cannot determine line from 2 same points
TEST(Intersect, LinePointBadLine)
{
  Point<int> a(2,5);
  Point<int> b(3,7);
  LineSegment<int> l(a,a);
  ASSERT_ANY_THROW(pointOnLine(b, l));
  ASSERT_ANY_THROW(pointOnLine(a, l));
}


TEST(Intersect, LineSegmentPoint)
{
  Point<int> a(20,50);
  Point<int> b(30,70);
  LineSegment<int> l(a,b);
  // line ends
  EXPECT_EQ(true, pointOnLineSegment(a, l, true));
  EXPECT_EQ(true, pointOnLineSegment(b, l, true));
  EXPECT_FALSE(pointOnLineSegment(a, l, false));
  EXPECT_FALSE(pointOnLineSegment(b, l, false));
  // on line but not on the line segment
  EXPECT_FALSE(pointOnLineSegment(a + (b-a)*2, l, true));
  EXPECT_FALSE(pointOnLineSegment(a - (b-a)*5, l, true));
  EXPECT_FALSE(pointOnLineSegment(a + (b-a)*2, l, false));
  EXPECT_FALSE(pointOnLineSegment(a - (b-a)*5, l, false));
  // inside line segment
  EXPECT_EQ(true, pointOnLineSegment(a + (b-a)/2, l, true));
  EXPECT_EQ(true, pointOnLineSegment(a + (b-a)/5, l, true));
  EXPECT_EQ(true, pointOnLineSegment(a + (b-a)/2, l, false));
  EXPECT_EQ(true, pointOnLineSegment(a + (b-a)/5, l, false));

  // other points
  EXPECT_FALSE(pointOnLineSegment(Point<int>(0,0), l, true));
  EXPECT_FALSE(pointOnLineSegment(Point<int>(0,0), l, false));
  EXPECT_FALSE(pointOnLineSegment(Point<int>(1,0), l, true));
  EXPECT_FALSE(pointOnLineSegment(Point<int>(1,0), l, false));
  EXPECT_FALSE(pointOnLineSegment(Point<int>(0,10), l, true));
  EXPECT_FALSE(pointOnLineSegment(Point<int>(0,10), l, false));
  EXPECT_FALSE(pointOnLineSegment(Point<int>(1,1), l, true));
  EXPECT_FALSE(pointOnLineSegment(Point<int>(1,1), l, false));
  EXPECT_FALSE(pointOnLineSegment(Point<int>(2000,40001), l, true));
  EXPECT_FALSE(pointOnLineSegment(Point<int>(2000,40001), l, false));
}

TEST(Intersect, LineSegmentPointDegenerated)
{
  Point<int> a(20,50);
  LineSegment<int> l(a,a);
  // line segment is only one point
  EXPECT_EQ(true, pointOnLineSegment(a, l, true));
  EXPECT_FALSE(pointOnLineSegment(a, l, false));
}

TEST(Intersect, IntersectsLineLineSegment)
{
  Point<int> a(20,50);
  Point<int> b(30,70);
  LineSegment<int> line(a, b);

  EXPECT_EQ(NO_INTERSECT, intersectLineLineSegment(line,
        LineSegment<int>(Point<int>(0,0), Point<int>(1,1))
        ));

  // overlay of line segments
  EXPECT_EQ(OVERLAY, intersectLineLineSegment(line, line));
  EXPECT_EQ(OVERLAY, intersectLineLineSegment(line,
        LineSegment<int>(Point<int>(21,52), Point<int>(22,54))
        ));
  EXPECT_EQ(OVERLAY, intersectLineLineSegment(line,
        LineSegment<int>(Point<int>(19,48), Point<int>(22,54))
        ));
  EXPECT_EQ(OVERLAY, intersectLineLineSegment(line,
        LineSegment<int>(Point<int>(29,68), Point<int>(32,74))
        ));

  // tangency overlay
  EXPECT_EQ(OVERLAY, intersectLineLineSegment(line,
        LineSegment<int>(Point<int>(19,48), Point<int>(20,50))
        ));
  EXPECT_EQ(OVERLAY, intersectLineLineSegment(line,
        LineSegment<int>(Point<int>(30,70), Point<int>(32,74))
        ));

  // non-tangency overlay
  EXPECT_EQ(OVERLAY, intersectLineLineSegment(line,
        LineSegment<int>(Point<int>(19,48), Point<int>(18,46))
        ));
  EXPECT_EQ(OVERLAY, intersectLineLineSegment(line,
        LineSegment<int>(Point<int>(31,72), Point<int>(32,74))
        ));

  // edge
  EXPECT_EQ(TANGENCY, intersectLineLineSegment(line,
        LineSegment<int>(Point<int>(23,44), Point<int>(20,50))
        ));
  EXPECT_EQ(TANGENCY, intersectLineLineSegment(line,
        LineSegment<int>(Point<int>(30,70), Point<int>(22,174))
        ));
  // middle
  EXPECT_EQ(TANGENCY, intersectLineLineSegment(line,
        LineSegment<int>(Point<int>(25,60), Point<int>(22,174))
        ));

  // normal intersect
  EXPECT_EQ(INTERSECT, intersectLineLineSegment(line,
        LineSegment<int>(Point<int>(24,60), Point<int>(26,60))
        ));

  // no intersect
  EXPECT_EQ(NO_INTERSECT, intersectLineLineSegment(line,
        LineSegment<int>(Point<int>(26,60), Point<int>(28,60))
        ));

}

/**
 * Test intersection of line segments
 */
TEST(Intersect, IntersectsLineSegmentLineSegment)
{
  Point<int> a(20,50);
  Point<int> b(30,70);
  LineSegment<int> line(a, b);

  EXPECT_EQ(NO_INTERSECT, intersectLineSegmentLineSegment(line,
        LineSegment<int>(Point<int>(0,0), Point<int>(1,1))
        ));

  // overlay of line segments
  EXPECT_EQ(OVERLAY, intersectLineSegmentLineSegment(line, line));
  EXPECT_EQ(OVERLAY, intersectLineSegmentLineSegment(line,
        LineSegment<int>(Point<int>(21,52), Point<int>(22,54))
        ));
  EXPECT_EQ(OVERLAY, intersectLineSegmentLineSegment(line,
        LineSegment<int>(Point<int>(19,48), Point<int>(22,54))
        ));
  EXPECT_EQ(OVERLAY, intersectLineSegmentLineSegment(line,
        LineSegment<int>(Point<int>(29,68), Point<int>(32,74))
        ));

  // tangency overlay
  EXPECT_EQ(TANGENCY, intersectLineSegmentLineSegment(line,
        LineSegment<int>(Point<int>(19,48), Point<int>(20,50))
        ));
  EXPECT_EQ(TANGENCY, intersectLineSegmentLineSegment(line,
        LineSegment<int>(Point<int>(30,70), Point<int>(32,74))
        ));

  // non-tangency overlay
  EXPECT_EQ(NO_INTERSECT, intersectLineSegmentLineSegment(line,
        LineSegment<int>(Point<int>(19,48), Point<int>(18,46))
        ));
  EXPECT_EQ(NO_INTERSECT, intersectLineSegmentLineSegment(line,
        LineSegment<int>(Point<int>(31,72), Point<int>(32,74))
        ));

  // edge
  EXPECT_EQ(TANGENCY, intersectLineSegmentLineSegment(line,
        LineSegment<int>(Point<int>(23,44), Point<int>(20,50))
        ));
  EXPECT_EQ(TANGENCY, intersectLineSegmentLineSegment(line,
        LineSegment<int>(Point<int>(30,70), Point<int>(22,174))
        ));
  // middle
  EXPECT_EQ(TANGENCY, intersectLineSegmentLineSegment(line,
        LineSegment<int>(Point<int>(25,60), Point<int>(22,174))
        ));

  // normal intersect
  EXPECT_EQ(INTERSECT, intersectLineSegmentLineSegment(line,
        LineSegment<int>(Point<int>(24,60), Point<int>(26,60))
        ));

  // no intersect
  EXPECT_EQ(NO_INTERSECT, intersectLineSegmentLineSegment(line,
        LineSegment<int>(Point<int>(26,60), Point<int>(28,60))
        ));
}

/**
 * Test intersection of intervals
 */
TEST(Intersect, IntersectIntervals)
{
  EXPECT_EQ(NO_INTERSECT,
      intervalIntersect(0,1,2,3));
  EXPECT_EQ(NO_INTERSECT,
      intervalIntersect(0,1,3,2));
  EXPECT_EQ(NO_INTERSECT,
      intervalIntersect(-5,10, 11, 21));
  EXPECT_EQ(NO_INTERSECT,
      intervalIntersect(-5,10, 21, 11));
  EXPECT_EQ(NO_INTERSECT,
      intervalIntersect(21,11, -5, 10));

  EXPECT_EQ(OVERLAY,
      intervalIntersect(0,1,-1,3));

  EXPECT_EQ(TANGENCY,
      intervalIntersect(0,1,1,3));
  EXPECT_EQ(TANGENCY,
      intervalIntersect(0,1,-1,0));
  EXPECT_EQ(TANGENCY,
      intervalIntersect(-2,1,3,1));

  EXPECT_EQ(OVERLAY,
      intervalIntersect(0,10,-1,1));
  EXPECT_EQ(OVERLAY,
      intervalIntersect(0,10,-1,10));
  EXPECT_EQ(OVERLAY,
      intervalIntersect(0,10,-1,100));
  EXPECT_EQ(OVERLAY,
      intervalIntersect(10,100,-1,100));
}

} // namespace two_d
} // namespace geometry
