#ifndef H_INTERVAL_TREES_ARRAY_INTERVAL_ARRAY
#define H_INTERVAL_TREES_ARRAY_INTERVAL_ARRAY

/**
 * @file Contains "interval tree" implementation done in a simple
 * array.
 *
 * @warning This class is inefficient, use only for testing!
 */
#include <vector>
#include <algorithm>

#pragma GCC diagnostic ignored "-Wtype-limits"

/**
 * Interval-tree-like implementation done in a simple
 * array.
 *
 * @warning This class is inefficient, use only for testing!
 */
template<typename ValueType>
class IntervalSumArray {
  typedef typename std::vector<ValueType>::size_type SizeType;
 public:
   /**
    * Initialize whole array
    */
    void initialize(SizeType size) {
      Preconditions::check(size > 0);
      data.clear();
      data.resize(size);
    }

    /**
     * increment interval [start, end)
     */
    void increment(SizeType start, SizeType end, ValueType value) {
      Preconditions::checkRange(start, end);
      Preconditions::check(end <= data.size());
      for (SizeType i = start; i < end; i++) {
        data[i] += value;
      }
    }

    /**
     * get sum of interval [start, end)
     */
    ValueType get_sum(SizeType start, SizeType end) {
      Preconditions::check(start >= 0);
      Preconditions::check(start <= end);
      Preconditions::check(end <= data.size());

      ValueType result = ValueType();
      for (SizeType i = start; i < end; i++) {
          result += data[i];
      }
      return result;
    }

 private:
    std::vector<ValueType> data;
};

/**
 * Interval-tree-like implementation done in a simple
 * array.
 *
 * @warning This class is inefficient, use only for testing!
 */
template<typename ValueType>
class IntervalMaxArray {
  typedef typename std::vector<ValueType>::size_type SizeType;
 public:
    void initialize(SizeType size) {
        data.clear();
        data.resize(size);
    }

    /**
     * sets value on whole interval [start, end)
     */
    void set(SizeType start, SizeType end, ValueType value) {
      Preconditions::checkRange(start, end);
      Preconditions::check(end <= data.size());

      for (SizeType i = start; i < end; i++) {
        data[i] = value;
      }
    }

    /**
     * update interval with new max [start, end)
     */
    void update(SizeType start, SizeType end, ValueType value) {
      Preconditions::checkRange(start, end);
      Preconditions::check(end <= data.size());

      for (SizeType i = start; i < end; i++) {
        data[i] = std::max(data[i], value);
      }
    }


    /**
     * get maximum of interval [start, end)
     */
    ValueType get_max(SizeType start, SizeType end) {
      Preconditions::check(start >= 0);
      Preconditions::check(start <= end);
      Preconditions::check(end <= data.size());

      ValueType result = ValueType();
      for (SizeType i = start; i < end; i++) {
        result = std::max(result, data[i]);
      }
      return result;
    }

 private:
    std::vector<ValueType> data;
};

#endif
