#ifndef H_INTERVAL_TREES_FENWICK_FENWICK
#define H_INTERVAL_TREES_FENWICK_FENWICK
/**
 * @file Contains Fenwick interval tree which can
 *  update a single value and query an interval
 *  [0, pos] or [pos, size) depeding on the implementation
 */

#include <unistd.h>
#include <vector>
#include <algorithm>
#include "utils/preconditions/preconditions.h"
#include <functional>
#include "debug/ppdebug.h"

namespace interval_trees {
namespace fenwick {

/**
 * Type of a fenwick tree, determines which type of range
 * cna Fenwick tree query
 */
enum FenwickDirection {
  TO_ZERO, // can query interval [0..pos]
  TO_INFTY, // can query interval [pos..size)
};

/**
 * General Fenwick interval tree.
 *
 * <li> can update position pos by applying "Operation"
 * <li> can report an "Operation" over values at interval [0..pos] or
 * [pos..size) (not both at once)
 *
 * Note: sensible operations are only plus, min, max
 * @precondition If you need a custom operation, you will need following concept:
 *  operation on [a..b) ==
 *    operation( operation on [a..c), operation on [c..b) )
 */
template<typename ValueType, class Operation>
class FenwickTree {
 public:
  typedef size_t SizeType;

  /**
   * Initialize fenwick tree
   */
  void initialize(FenwickDirection type_, size_t size,
     const ValueType& initial) {
    Preconditions::check(size > 0, "size shoud be positive");
    data.clear();
    data.resize(size, initial);
    type = type_;
  }

  /**
   * Update value on position \pos by \a value using operator
   * Operation::operation(old, value);
   */
  void update(SizeType pos, ValueType value) {
    Preconditions::checkRange(pos, data.size());
    pos++;
    while (pos > 0 && pos <= data.size()) {
      data[pos-1] = Operation::operation(data[pos-1], value);
      pos = _advance(pos, TO_ZERO);
    }
  }

  /**
   * Return value of operation on interval
   * [0, pos] or [pos, size) depending on the type of a tree
   */
  ValueType get_on_interval(SizeType pos) {
    Preconditions::checkRange(pos, data.size());
    ValueType result = data[pos];
    pos = _advance(pos + 1, TO_INFTY);
    while (pos > 0 && pos <= data.size()) {
      result = Operation::operation(result, data[pos-1]);
      pos = _advance(pos, TO_INFTY);
    }
    return result;
  }

 private:
  /**
   * advance to next position in the structure
   */
  inline SizeType _advance(SizeType pos, FenwickDirection direction) {
    if (type == direction) {
      return pos + last_one(pos);
    } else {
      return pos - last_one(pos);
    }
  }

  /**
   * @returns last set bit of an integer
   */
  SizeType last_one(SizeType x) {
    return x ^ (x & (x - 1));
  }

  std::vector<ValueType> data;
  FenwickDirection type;
};


/**
 * Fenwick operation for sums
 */
template <typename T>
struct BinaryPlus {
    static T operation(const T& x, const T& y) {
      return x + y;
    }
};

template<typename T>
class FenwickSumTree {
  /**
   * Underlying Fenwick tree type
   */
  typedef class FenwickTree<T, BinaryPlus<T> > FenwickType;

  /**
   * Type of indexes
   */
  typedef typename FenwickType::SizeType SizeType;

 public:

  void initialize(SizeType size) {
    fenwick.initialize(TO_ZERO, size, 0);
  }

  void increment(SizeType pos, T value) {
    fenwick.update(pos, value);
  }

  T get_prefix_sum(SizeType pos) {
    return fenwick.get_on_interval(pos);
  }

 private:
  /**
   * Underlying fenwick tree implementation
   */
  FenwickType fenwick;
};

template <typename T>
struct BinaryMax  {
    static T operation(const T& x, const T& y) {
      return std::max(x, y);
    }
};

template<typename T>
class FenwickMaxTree {
 public:
  void initialize(FenwickDirection type, size_t size) {
    T initial = T(); // zero

    fenwick.initialize(type, size, initial);
  }

  void update(int pos, T value) {
    fenwick.update(pos, value);
  }

  T get_max(int pos) {
    return fenwick.get_on_interval(pos);
  }

 private:
  typedef FenwickTree<T, BinaryMax<T> > FenwickType;
  FenwickType fenwick;
};

} // namespace fenwick
} // namespace interval_trees
#endif
