#ifndef H_INTERVAL_TREES_FULL_BINARY_TREE_FULL_BINARY_TREE
#define H_INTERVAL_TREES_FULL_BINARY_TREE_FULL_BINARY_TREE

/**
 * This is implementation of full binary tree
 */

#include <stdint.h>
#include <vector>
#include "utils/preconditions/preconditions.h"
#include "interval_trees/utils/heap.h"

namespace interval_trees {

/**
 * Simple interval tree with
 * get/set value and get_max over interval
 */
template<typename NodeType>
class FullBinaryTree {
  private:
    /**
     * number of leaves in heap structure.
     * Also, pos+base is the index of leaf in data
     */
    std::vector<NodeType> data;

    typedef size_t Tpos;
  public:
    FullBinaryTree() {
      _clear();
    }

    /**
     * Clear tree and set it's size to zero.
     */
    void _clear() {
      data.clear();
    }

    /**
     * Shorthand for initialization
     * @see initialize(size, default_value
     */
    void initialize(Tpos size)
    {
      initialize(size, NodeType());
    }

    /**
     * Initialize tree
     */
    void initialize(Tpos size_, const Tpos& default_value) {
      _clear();
      Preconditions::check(size_ > 0, "size should be positive");
      Tpos s = heap::nextPowerOfTwo(size_);
      Preconditions::check(size_ == s,
          "Wrong size - should be power of two");
      if (s >= std::numeric_limits<Tpos>::max() / 2) {
        throw std::overflow_error("Too big size");
      }

      data.resize(s * 2, default_value);
    }



class Traverser { 
  private:
    std::vector<NodeType>* data_ptr;
    Tpos pos;
    Tpos r_left;
    Tpos r_right;
  public:
  explicit Traverser(std::vector<NodeType>* data_,
      Tpos pos_, Tpos left_, Tpos right_) {
    data_ptr = data_;
    pos = pos_;
    r_left = left_;
    r_right = right_;
  }

  NodeType& operator *() {
    return (*data_ptr)[pos];
  }

  const NodeType& operator *() const {
    return (*data_ptr)[pos];
  }

  Traverser left() {
    Preconditions::check(pos < data_ptr->size() / 2, "already leaf");
    // Note: left + right is max 
    return Traverser(data_ptr, heap::left(pos), r_left, (r_left + r_right) / 2);
  }

  Traverser right() {
    Preconditions::check(pos < data_ptr->size() /2, "already leaf");
    return Traverser(data_ptr, heap::right(pos), (r_left + r_right) / 2, r_right);
  }

  Traverser parent() {
    Preconditions::check(pos > 1, "already parent");
    Tpos l,r;
    if (heap::isLeftChild(pos)) {
      l = r_left;
      r = 2 * r_right - r_left;
    } else {
      l = r_right - 2 * r_left;
      r = r_right;
    }
    return Traverser(data_ptr, heap::parent(pos), l, r);
  }

  Tpos range_left() const {
    return r_left;
  }

  Tpos range_right() const {
    return r_right;
  }

};

Traverser root(){
  return Traverser(&data, 1, 0, data.size() / 2);
}

Traverser leaf(Tpos pos) {
  Preconditions::checkRange(pos, data.size() / 2);
  return Traverser(&data, data.size() / 2 + pos, pos, pos + 1);
}

};
} // namespace interval_trees
#endif
