#ifndef H_INTERVAL_TREES_SIMPLE_SIMPLE_MAX
#define H_INTERVAL_TREES_SIMPLE_SIMPLE_MAX
/**
 * @file Contains simple interval tree with
 * get/set of a single value and get_max over interval
 */

#include <vector>
#include "utils/preconditions/preconditions.h"
#include "interval_trees/utils/heap.h"

namespace interval_trees {
namespace simple {

/**
 * This implementation of simple interval tree with following
 * operations:
 * - can change element at position i
 * - can compute maximum over some range [left, right)
 *
 * Implementation - specific details:
 * This implementation is solely non-recursive and uses bottom-up
 * approach instead of standard top-bottom recursive calls.
 * Because of lack of recursion, you can't (easily) adapt this
 * implementation to change whole intervals.
 */
template<typename T>
class SimpleMaxTree {

  /**
   * SizeType is enough to store index to vector.
   */
  typedef typename std::vector<T>::size_type SizeType;
 private:
    /**
     * Number of leaves in heap structure.
     *
     * Also, pos+base is the index of leaf in data.
     * Note that 2*base is the size of the whole tree.
     */
    SizeType base;

    /**
     * The requested size of the structure,
     * we won't allow access outside this range
     */
    SizeType original_size;

    /**
     * The data of the tree
     */
    std::vector<T> data;


  public:
    SimpleMaxTree() {
      base = 0;
      _clear();
    }

    /**
     * Clear tree and set it's size to zero.
     */
    void _clear() {
      base = 0;
      original_size = 0;
      data.clear();
    }

    /**
     * Shorthand for initialization
     * @see initialize(size, default_value
     */
    void initialize(SizeType size)
    {
      initialize(size, T());
    }

    /**
     * Initialize tree
     */
    void initialize(SizeType size_, const T& default_value) {
      _clear();
      Preconditions::check(size_ > 0, "size should be positive");
      SizeType s = heap::nextPowerOfTwo(size_);
      // We need twice the size of s to allocate,
      // check first it will fit into SizeType
      if (s >= std::numeric_limits<SizeType>::max() / 2) {
        throw std::overflow_error("Too big size");
      }
      data.resize(s * 2, default_value);
      // Note: these assigments should be *after* resize as
      // resize can throw an bad_alloc exception and in that case we
      // want to have size zero!
      base = s;
      original_size = size_;
    }

    /**
     * get value at position pos
     */
    T get(SizeType pos) {
      Preconditions::checkRange(pos, original_size);
      return data[pos + base];
    }

    /**
     * set value at position pos
     */
    void set(SizeType pos, T value) {
      Preconditions::checkRange(pos, original_size);
      pos += base; // leafs are indexed starting at "base"
      data[pos] = value;

      while (pos > 1) {
        pos = heap::parent(pos);
        // update this node
        data[pos] = std::max(data[heap::left(pos)], data[heap::right(pos)]);
      }
    }

    /**
     * get maximum over interval [left, right)
     *
     * Note: right should be greater than left!
     *
     * Implementation details:
     * we walk simultaneously from the left and right
     * up the heap-structure until we meet at the same point
     */
    T get_max(SizeType left, SizeType right) {
      Preconditions::checkRange(left, original_size);
      // original_size <= SizeType max/2 and thus it won't overflow 
      Preconditions::checkRange(right, (SizeType)1, original_size + 1);
      Preconditions::check(left < right);

      // safe, right + base - 1 will fit into SizeType
      left += base;
      right += base - 1;

      // Note: if (at this point) left == right
      // we will maximize from two same values.
      // This is not a problem, but beware of this
      // corner case if you use other function (sum for example)
      T result = std::max(data[left], data[right]);
      while (heap::parent(left) != heap::parent(right)) {
        if (heap::isLeftChild(left)) {
          // include right sibling
          result = std::max(result, data[heap::sibling(left)]);
        }
        if (heap::isRightChild(right)) {
          // include left sibling
          result = std::max(result, data[heap::sibling(right)]);
        }
        left = heap::parent(left);
        right = heap::parent(right);
      }
      return result;
    }
};

} // namespace simple
} // namespace interval_trees

#endif
