#ifndef H_INTERVAL_TREES_UTILS_HEAP
#define H_INTERVAL_TREES_UTILS_HEAP
/**
 * @file Contains useful functions which can operate on heap stored in array.
 * from index 1, the organization of heap is this:
 * <pre> 
 *           1
 *         /    \
 *       /        \
 *      2          3
 *    /   \      /   \
 *   4     5    6     7
 *  / \   / \  / \   / \
 *  8 9 10 11 12 13 14 15
 * </pre>
 */
#include <stdint.h>
#include <limits>
#include "utils/preconditions/preconditions.h"

namespace heap {
/** 
 * @param x node_position
 *
 * @returns position of node corresponding to left child
 */
template <typename T>
inline T left(T x) {
  Preconditions::check(x > 0);
  if (x > std::numeric_limits<T>::max() / 2) {
    throw std::overflow_error("");
  }
  return x * 2;
};

/** 
 * @param x node_position
 *
 * @returns position of node corresponding to right child
 */
template <typename T>
inline T right(T x) {
  Preconditions::check(x > 0);
  if (x > std::numeric_limits<T>::max() / 2) {
    throw std::overflow_error("");
  }
  return x * 2 + 1;
};

/**
 * @param x node_position
 *
 * @returns position of the parent node
 */
template <typename T>
inline T parent(T x) {
  Preconditions::check(x > 1, "Node does not have a parent");
  return x / 2;
};

/**
 * @returns true if this is left child
 */
template <typename T>
inline bool isLeftChild(T x) {
  Preconditions::check(x > 1);
  return (x % 2) == 0;
}
/**
 * @returns true if this is right child
 */
template <typename T>
inline bool isRightChild(T x) {
  Preconditions::check(x > 1);
  return (x % 2) == 1;
}

/**
 * @returns position of sibling of this node
 */
template <typename T>
inline T sibling(T x) {
  Preconditions::check(x > 1);
  return x ^ 1;
}

/**
 * Find smallest power of two that is at least x.
 * Example: 2->2, 3->4, 4->4, 5->8
 *
 * @returns next power of two
 * @throws overflow_error
 */
template <typename T>
T nextPowerOfTwo(T x) {
  Preconditions::check(x > 0, "Argument should be positive!");

  if (x > 1 + std::numeric_limits<T>::max() / 2) {
    throw std::overflow_error("Next power of two overflows.");
  }

  size_t tmp = x - 1;
  int n = 1;
  // LogLog N implementation.
  while (n < std::numeric_limits<T>::digits) {
    tmp |= (tmp >> n);
    n *= 2;
  }
  return tmp + 1;
}

}

#endif
