#include "heap.h"
#include "gtest/gtest.h"

namespace heap {
TEST(HeapOperations, left) {
  EXPECT_THROW(left(0), std::invalid_argument);
  EXPECT_EQ(2, left(1));
  EXPECT_EQ(4, left(2));
  EXPECT_EQ(6, left(3));
  EXPECT_EQ(8, left(4));
  EXPECT_EQ(10, left(5));
  EXPECT_EQ(12, left(6));
  EXPECT_EQ(14, left(7));
}

TEST(HeapOperations, right) {
  EXPECT_THROW(right(0), std::invalid_argument);
  EXPECT_EQ(3, right(1));
  EXPECT_EQ(5, right(2));
  EXPECT_EQ(7, right(3));
  EXPECT_EQ(9, right(4));
  EXPECT_EQ(11, right(5));
  EXPECT_EQ(13, right(6));
  EXPECT_EQ(15, right(7));
}

TEST(HeapOperations, parent) {
  EXPECT_THROW(parent(0), std::invalid_argument);
  EXPECT_THROW(parent(1), std::invalid_argument);

  EXPECT_EQ(1, parent(2));
  EXPECT_EQ(1, parent(3));

  EXPECT_EQ(2, parent(4));
  EXPECT_EQ(2, parent(5));

  EXPECT_EQ(3, parent(6));
  EXPECT_EQ(3, parent(7));

  EXPECT_EQ(4, parent(8));
  EXPECT_EQ(4, parent(9));

  EXPECT_EQ(5, parent(10));
  EXPECT_EQ(5, parent(11));

  EXPECT_EQ(6, parent(12));
  EXPECT_EQ(6, parent(13));

  EXPECT_EQ(7, parent(14));
  EXPECT_EQ(7, parent(15));
}

TEST(HeapOperations, sibling) {
  EXPECT_THROW(sibling(0), std::invalid_argument);
  EXPECT_THROW(sibling(1), std::invalid_argument);

  EXPECT_EQ(3, sibling(2));
  EXPECT_EQ(2, sibling(3));

  EXPECT_EQ(5, sibling(4));
  EXPECT_EQ(4, sibling(5));

  EXPECT_EQ(7, sibling(6));
  EXPECT_EQ(6, sibling(7));

  EXPECT_EQ(9, sibling(8));
  EXPECT_EQ(8, sibling(9));

  EXPECT_EQ(11, sibling(10));
  EXPECT_EQ(10, sibling(11));

  EXPECT_EQ(13, sibling(12));
  EXPECT_EQ(12, sibling(13));

  EXPECT_EQ(15, sibling(14));
  EXPECT_EQ(14, sibling(15));
}

TEST(NextPowerOfTwo, badInput)
{
  EXPECT_THROW(nextPowerOfTwo(0), std::invalid_argument);
}

TEST(NextPowerOfTwo, overflow)
{
  EXPECT_THROW(nextPowerOfTwo(
                  std::numeric_limits<size_t>::max()),
               std::overflow_error);

  EXPECT_THROW(nextPowerOfTwo(
                 std::numeric_limits<size_t>::max() / 2 + 2),
               std::overflow_error);

  EXPECT_NO_THROW(nextPowerOfTwo(
                    std::numeric_limits<size_t>::max() / 2 + 1));

  EXPECT_NO_THROW(nextPowerOfTwo(
                    std::numeric_limits<size_t>::max() / 2));
}

TEST(NextPowerOfTwo, powersOfTwo)
{
  EXPECT_EQ(1, nextPowerOfTwo(1));
  EXPECT_EQ(2, nextPowerOfTwo(2));
  EXPECT_EQ(4, nextPowerOfTwo(4));
  EXPECT_EQ(8, nextPowerOfTwo(8));
  EXPECT_EQ(16, nextPowerOfTwo(16));
  EXPECT_EQ(32, nextPowerOfTwo(32));

  EXPECT_EQ(1u<<10, nextPowerOfTwo(1u<<10));
  EXPECT_EQ(1u<<20, nextPowerOfTwo(1u<<20));
  EXPECT_EQ(1u<<30, nextPowerOfTwo(1u<<30));
}

TEST(NextPowerOfTwo, powersMinusOne)
{
  EXPECT_EQ(4, nextPowerOfTwo(3));
  EXPECT_EQ(8, nextPowerOfTwo(7));
  EXPECT_EQ(16, nextPowerOfTwo(15));
  EXPECT_EQ(32, nextPowerOfTwo(31));

  EXPECT_EQ(1u<<10, nextPowerOfTwo((1u<<10) - 1));
  EXPECT_EQ(1u<<20, nextPowerOfTwo((1u<<20) - 1));
  EXPECT_EQ(1u<<30, nextPowerOfTwo((1u<<30) - 1));
}

TEST(NextPowerOfTwo, powersPlusOne)
{
  EXPECT_EQ(4, nextPowerOfTwo(3));
  EXPECT_EQ(8, nextPowerOfTwo(5));
  EXPECT_EQ(16, nextPowerOfTwo(9));
  EXPECT_EQ(32, nextPowerOfTwo(17));
  EXPECT_EQ(64, nextPowerOfTwo(33));

  EXPECT_EQ(1u<<11, nextPowerOfTwo((1u<<10) + 1));
  EXPECT_EQ(1u<<21, nextPowerOfTwo((1u<<20) + 1));
  // this must be unsigned!
  EXPECT_EQ(1u<<31, nextPowerOfTwo((1u<<30) + 1));
}

TEST(NextPowerOfTwo, random)
{
  unsigned int testdata[][2] = {
    {804334813, 1073741824},
    {7888, 8192},
    {327463, 524288},
    {436, 512},
    {4706125, 8388608},
    {13721806, 16777216},
    {0, 0}
  };
  for (int i = 0; testdata[i][0] != 0; i++) {
    EXPECT_EQ(testdata[i][1], nextPowerOfTwo(testdata[i][0]));
  }
}

}

