#ifndef H_MATH_BINSEARCH_FUNCTION_BINSEARCH
#define H_MATH_BINSEARCH_FUNCTION_BINSEARCH

/**
 * This file contains implementation of function root/minimum finding
 * algorithms based on binary search
 */

#include "utils/preconditions/preconditions.h"
#include <cmath>

namespace math {
namespace binsearch {

/**
 * Base class of all unary functions
 */
template<typename T> class Function
{
 public:
  virtual T operator()(T x) = 0;
};

/**
 * Base class of all convex functions
 * 
 * Function is convex, iff 
 *  <i> f(t * x1 + (1 - t) * x2) <= t * f(x1) + (1 - t) * f(x2) </i>
 *  for every <i> x1 != x2 </i> and <i> 0 < t < 1 <i>
 */
template<typename T>
class ConvexFunction: public Function<T> {
};

template<typename T>
class FunctionBinsearch {
 private:
   /**
    * Calculate the number of iterations needed to obtain specified
    * precision if original range is <i> [left, right) </i> and
    * each iteration the range is reduced to <i> 1/division </i>
    * of its size.
    *
    * @warning This is exact-math result, in floating-point numbers
    * it may not be possible to obtain required precision. In that case
    * the iteration count should be hovewer sufficient as after the required
    * number of iterations, the computation will be stalled
    */
   int number_of_iterations(T left, T right, T precision, double
       division) {
    Preconditions::check(right > left, "invalid range");
    Preconditions::check(precision > 0, "precision must be positive");

    double tmp = log(right - left) - log(precision);
    return 2 + (int) (tmp / log(division));
   }

 public:
   /**
    * Finds any root of function on interval <i> [left, right) </i>
    * with specified precision
    *
    * @precondition You must provide such an interval, that
    * function values are of opposite sign at the interval ends.
    *
    * @warning Real error can be greater than precision
    * in case that precision is smaller than distance between two floating-point
    * values at required range
    *
    * @returns any root on the interval.
    */
  T root(Function<T>* f, T left, T right, T precision)
  {
    Preconditions::check(right >= left, "invalid range");
    Preconditions::check(precision > 0, "precision must be positive");
    Preconditions::check((*f)(left) * (*f)(right) <= 0,
        "endpoint should have different sign");

    if (right - left <= precision) {
      return left;
    }

    assert(left < right);

    int sgn = (*f)(left) > 0 ? 1 : -1;
  
    int iterations = number_of_iterations(left, right, precision, 2.0);
    while (iterations--) {
      T middle = (left + right) / 2;
      if (sgn * (*f)(middle) > 0) {
        left = middle;
      } else {
        right = middle;
      }
    }
    return left;
  }

  /**
   * Finds a minimum value of convex function on given interval
   * <i> [left, right) </i>
   * 
   * @warning Real error can be greated than precision if precision is
   * small compared to magnitude of the result, @see root()
   *
   * @precondition function should be convex
   *
   * @param left left end of the interval
   * @param right right end of the interval
   * @param precision required precision
   *
   * @returns argmin
   */
  T convex_min(ConvexFunction<T>* f,
                  T left, T right, T precision)
  {
    Preconditions::check(right >= left);
    Preconditions::check(precision > 0);
    
    if (right - left <= precision) {
      return left;
    }

    assert(left < right);

    int iterations = number_of_iterations(left, right, precision, 1.5);
    while (iterations--) {
      T mid1 = (left * 2 + right    ) / 3;
      T mid2 = (left     + right * 2) / 3;

      if ((*f)(mid1) < (*f)(mid2)) {
        right = mid2;
      } else {
        left = mid1;
      }
    }

    return left;
  }
};

} // namespace binsearch
} // namespace math
#endif
