#include <gtest/gtest.h>
#include "function_binsearch.h"
#include "template/template.h"
#include "utils/rand/rand.h"

namespace math {
namespace binsearch {

template<typename T>
class Quadr: public Function<T> {
 public:
  Quadr(T a_, T b_, T c_) {
      a=a_;
      b=b_;
      c=c_;
  }

  T operator()(T x) {
      return a * x * x + b * x + c;
  }

 private:
  T a,b,c;
};

TEST(Binsearch, LinearSimple) {
  Quadr<double> q = Quadr<double>(0,1,0);
  FunctionBinsearch<double> bs;
  EXPECT_NEAR(0, bs.root(&q, -1, 1, 1e-10), 1e-10);
  EXPECT_NEAR(0, bs.root(&q, -10, 10, 1e-10), 1e-10);
  // try to not hang
  EXPECT_NEAR(0, bs.root(&q, -1e20, 1e20, 1e-10), 1e-10);
}

TEST(Binsearch, LinearNegative) {
  Quadr<double> q = Quadr<double>(0,-1,0);
  FunctionBinsearch<double> bs;
  EXPECT_NEAR(0, bs.root(&q, -1, 1, 1e-10), 1e-10);
  EXPECT_NEAR(0, bs.root(&q, -10, 10, 1e-10), 1e-10);
}

TEST(Binsearch, Zero) {
  Quadr<double> q = Quadr<double>(0,0,0);
  FunctionBinsearch<double> bs;
  EXPECT_NEAR(0, bs.root(&q, -1, 1, 1e-10), 1);
}

TEST(Binsearch, Linear) {
  Quadr<double> q = Quadr<double>(0,3,5);
  FunctionBinsearch<double> bs;
  EXPECT_NEAR(-5.0 / 3.0, bs.root(&q, -2, 0, 1e-10), 1e-10);
}

#define LIN_TESTS 1000
TEST(Binsearch, RandomLinear) {
  Rand rand(47);
  FunctionBinsearch<double> bs;
  FOR(q, LIN_TESTS) {
    double a = rand.expranddouble(1e-5, 1e10);
    double b = rand.randdouble(-1e5, 1e5);
    Quadr<double> q = Quadr<double>(0, a, b);
    double left = -b/a - rand.expranddouble(1e-5, 1e10);
    double right = -b/a + rand.expranddouble(1e-5, 1e10);
    // use reasonable precision for that range.
    double precision = abs(-b/a * 1e-14) + 1e-14;
    EXPECT_NEAR(-b/a, bs.root(&q, left, right, precision), precision);
  }
}

#define QUADR_TESTS 1000
TEST(Binsearch, RandomQuadratic) {
  Rand rand(42);
  FunctionBinsearch<double> bs;
  FOR(q, QUADR_TESTS) {
    double r = rand.randdouble(-5000, 10000);
    double l = rand.randdouble(-10000, r);
    double t = rand.expranddouble(0.1, 100);

    Quadr<double> q = Quadr<double>(t, -t*(l+r), t*l*r);
    double left = l - rand.expranddouble(1e-5, 1e5);
    double right = r + rand.expranddouble(1e-5, 1e5);
    double middle = (l+r)/2;

    EXPECT_NEAR(l, bs.root(&q, left, middle, 1e-8), 1e-8);
    EXPECT_NEAR(r, bs.root(&q, middle, right, 1e-8), 1e-8);
  }
}

} // namespace binseach
} // namespace math
