#include <gtest/gtest.h>
#include "function_binsearch.h"
#include "template/template.h"
#include "utils/rand/rand.h"

namespace math {
namespace binsearch {

template<typename T>
class ConvexQuadr: public ConvexFunction<T> {
 public:
  ConvexQuadr(T a_, T b_, T c_) {
    Preconditions::check(a_ >= 0);
    a=a_;
    b=b_;
    c=c_;
  }

  T operator()(T x) {
      return a * x * x + b * x + c;
  }

 private:
  T a,b,c;
};

TEST(Binsearch, LinearSimple) {
  ConvexQuadr<double> q = ConvexQuadr<double>(0,1,0);
  FunctionBinsearch<double> bs;
  EXPECT_NEAR(-1, bs.convex_min(&q, -1, 1, 1e-10), 1e-10);
  EXPECT_NEAR(-10, bs.convex_min(&q, -10, 10, 1e-10), 1e-10);
  EXPECT_NEAR(-1e10, bs.convex_min(&q, -1e10, 1e20, 1e-2), 1e-2);
}

TEST(Binsearch, LinearNegative) {
  ConvexQuadr<double> q = ConvexQuadr<double>(0,-1,0);
  FunctionBinsearch<double> bs;
  EXPECT_NEAR(1, bs.convex_min(&q, -1, 1, 1e-10), 1e-10);
  EXPECT_NEAR(10, bs.convex_min(&q, -10, 10, 1e-10), 1e-10);
}

TEST(Binsearch, Zero) {
  ConvexQuadr<double> q = ConvexQuadr<double>(0,0,0);
  FunctionBinsearch<double> bs;
  EXPECT_NEAR(0, bs.convex_min(&q, -1, 1, 1e-10), 1);
}

#define QUADR_TESTS 1000
TEST(Binsearch, RandomQuadratic) {
  Rand rand(111);
  FunctionBinsearch<double> bs;
  FOR(q, QUADR_TESTS) {
    double r = rand.randdouble(-10, 20);
    double l = rand.randdouble(-20, r);
    double t = rand.expranddouble(0.1, 2);

    ConvexQuadr<double> q = ConvexQuadr<double>(t, -t*(l+r), t*l*r);
    double left = l - rand.expranddouble(1e-5, 1e5);
    double right = r + rand.expranddouble(1e-5, 1e5);
    double middle = (l+r)/2;

    EXPECT_NEAR(middle, bs.convex_min(&q, left, right, 1e-6), 1e-6);
  }
}

} // namespace binsearch
} // namespace math
