#ifndef H_MATH_BINSEARCH_INT_BINSEARCH
#define H_MATH_BINSEARCH_INT_BINSEARCH
/**
 * @file Contains lower_bound and upper_bound
 * functions similar to std::lower_bound and std::upper_bound
 */
#include "utils/preconditions/preconditions.h"
#include "utils/static_assert/static_assert.h"
#include <stdexcept>

namespace math {
namespace binsearch {

/**
 * Finds the middle of the range <i> [left, right) </i>
 *
 * Middle is defined as <i> floor((left + right) / 2) </i>
 *
 * @precondition \a T is integral type
 * @precondition <i> (right-left) </i> is representable in type \a T
 *
 * @param left start of the interval
 * @param right first index after the end of the interval
 *
 * @returns midde of the interval
 */
template <typename T>
T range_middle(T left, T right) {
  STATIC_ASSERT(std::numeric_limits<T>::is_integer,
      "T should be of integral type");
  Preconditions::check(left < right, "Invalid range");
  T len = right - left;
  if (len <= 0) {
    throw std::overflow_error("Too big range!");
  }
  return left + len / 2;
}

/**
 * Find first index in array range <i> [left, right) </i>
 * where the value may be inserted without violating the ordering
 *
 * Note that the definition is same as
 * ``index of first element which is  >= value''
 * except that the result is <i> right </i> if no such value exists
 *
 * Example:
 * <pre>
 *     a = 1 1 2 2 2 3 5
 * lb(1) = ^
 * lb(2) =     ^
 * lb(4) =             ^
 * lb(6) =               ^ (==right)
 * </pre>
 *
 * @precondition sorted array
 * @precondition \a SizeType is integral type
 * @precondition <i>(right-left)</i> will fit into type \a SizeType
 *
 * @param left start of the interval
 * @param right index after the end of the interval
 *
 * @returns index of the binsearched value
 *
 */
template <typename ValueType, typename SizeType>
  SizeType lower_bound(ValueType pole[], SizeType left, SizeType right, ValueType value) 
{
  STATIC_ASSERT(std::numeric_limits<SizeType>::is_integer,
      "SizeType should be of integral type");
  Preconditions::check(left <= right, "Invalid range");
  if (left == right) {
    return left;
  }
  if (std::numeric_limits<SizeType>::is_signed) {
    // Checks for signed values overflow.
    // Note that right - left == 0 could be also caused by overflow!
    if (right - left < 0) {
      throw std::overflow_error("Too big range!");
    }
  }  
  while (right - left > 0) {
    SizeType middle = range_middle(left, right);
    if (pole[middle] < value) {
      left = middle + 1;
    } else {
      right = middle;
    }
  }
  return left;
}

/**
 * Finds last position from range <i>[left, right) </i>
 * where the value may be inserted without violating ordering
 *
 * Note that the definitions is the same as
 * ``index of first element that is greater than value''
 * except that the result is <i> right </i> if no such value exists
 *
 * Example:
 * <pre>
 *       1 1 2 2 3 5
 * ub(6)             ^ (== right)
 * ub(2)         ^
 * ub(1)     ^
 * ub(0) ^
 * </pre>
 *
 * @precondition <i> (right-left) </i> should fit into ValueType
 * @precondition \a SizeType should be integral type
 *
 * @param left start of the interval
 * @param right index after the end of the interval
 *
 * @returns index of the binsearched value
 */
template <typename ValueType, typename SizeType>
SizeType upper_bound(ValueType pole[], SizeType left, SizeType right, ValueType value) {
  Preconditions::check(left <= right, "Invalid range");
  if (left == right) {
    return left;
  }
  if (std::numeric_limits<SizeType>::is_signed) {
    // Checks for signed values overflow.
    // Note that right - left == 0 could be also caused by overflow!
    if (right - left <= 0) {
      throw std::overflow_error("Too big range");
    }
  }
  while (right - left > 0) {
    SizeType middle = range_middle(left, right);
    if (value < pole[middle]) {
      right = middle;
    } else {
      left = middle + 1;
    }
  }
  return left;
}

} // namespace binsearch
} // namespace math
#endif
