#include <gtest/gtest.h>
#include "int_binsearch.h"
#include "template/template.h"
#include "utils/rand/rand.h"

namespace math {
namespace binsearch {

TEST(RangeMiddle, Simple) {
  EXPECT_EQ(5, range_middle(1, 9));
  EXPECT_EQ(4, range_middle(0, 9));
  EXPECT_EQ(4, range_middle(1, 8));
  EXPECT_EQ(4, range_middle(0, 8));
}

TEST(RangeMiddle, NegativeValues) {
  EXPECT_EQ(2, range_middle(-5, 9));
  EXPECT_EQ(1, range_middle(-6, 9));
  EXPECT_EQ(1, range_middle(-5, 8));
  EXPECT_EQ(1, range_middle(-6, 8));

  EXPECT_EQ(-15, range_middle(-19, -11));
  EXPECT_EQ(-15, range_middle(-19, -10));
  EXPECT_EQ(-15, range_middle(-18, -11));
  EXPECT_EQ(-14, range_middle(-18, -10));
}

TEST(RangeMiddle, BadInput) {
  EXPECT_THROW(range_middle(5, 5), std::invalid_argument);
  EXPECT_THROW(range_middle(-5, -5), std::invalid_argument);
  EXPECT_THROW(range_middle(-5, -6), std::invalid_argument);
  EXPECT_THROW(range_middle(5, -1), std::invalid_argument);
}

TEST(RangeMiddle, overflow) {
  unsigned char m = std::numeric_limits<unsigned char>::max();
  EXPECT_EQ(m / 2, range_middle((unsigned char) 0, m));

  char n = std::numeric_limits<char>::max();
  EXPECT_EQ(n / 2, range_middle((char) 0, n));
  EXPECT_THROW(range_middle((char) -1, n), std::overflow_error);
}

TEST(Binsearch, Empty) {
  int data[0];
  EXPECT_EQ(0, lower_bound(data, 0,0,0));
  EXPECT_EQ(0, upper_bound(data, 0,0,0));

  EXPECT_EQ(10, lower_bound(data, 10, 10, 47));
  EXPECT_EQ(10, upper_bound(data, 10, 10, 47));
}

TEST(Binsearch, Simple) {
  int data[3]={1,2,3};
  EXPECT_EQ(0, lower_bound(data, 0, 3, 0));
  EXPECT_EQ(0, upper_bound(data, 0, 3, 0));

  EXPECT_EQ(1, lower_bound(data, 1, 2, 2));
  EXPECT_EQ(2, upper_bound(data, 1, 2, 2));

  EXPECT_EQ(2, lower_bound(data, 1, 2, 47));
  EXPECT_EQ(2, upper_bound(data, 1, 2, 47));
}


TEST(Binsearch, SameValues) {
  int data[5]={1,2,2,2,3};
  EXPECT_EQ(0, lower_bound(data, 0, 5, 0));
  EXPECT_EQ(0, upper_bound(data, 0, 5, 0));

  EXPECT_EQ(1, lower_bound(data, 0, 5, 2));
  EXPECT_EQ(4, upper_bound(data, 0, 5, 2));

  EXPECT_EQ(5, lower_bound(data, 0, 5, 47));
  EXPECT_EQ(5, upper_bound(data, 0, 5, 47));
}

TEST(Binsearch, SameValuesSpecialRange) {
  int data[5]={1,2,2,2,3};
  EXPECT_EQ(2, lower_bound(data, 2, 3, 2));
  EXPECT_EQ(3, upper_bound(data, 2, 3, 2));
}

#define MAX 10000000
int data[MAX];

#define CNT_TESTS 100
#define CNT_SEARCHES 10000

TEST(Binsearch, CompareToStd) {
  Rand rand(99);
    printf("Testing %d scenarios, each %d searches\n",
        CNT_TESTS, CNT_SEARCHES);

    int cnt = CNT_TESTS;
    while (cnt--) {
        int n = rand.exprand(1, MAX+1)-1;
        int mod = rand.exprand(1, RAND_MAX);
        int start = rand.rand(0, n+1);
        FOR(q, n) data[q] = rand.rand(0, mod);
        std::sort(data+start, data+n);
        FOR(q, CNT_SEARCHES) {
            int prvok = rand.rand(-5, mod + 5);
            int l = std::lower_bound(data+start, data+n, prvok) - data;
            int u = std::upper_bound(data+start, data+n, prvok) - data;
            assert(lower_bound(data,start,n, prvok) == l);
            assert(upper_bound(data,start,n, prvok) == u);
        }
    }
}

} // namespace binsearch
} // namespace math
