#include <gtest/gtest.h>
#include "factorize_with_oracle.h"
#include "math/primes/primes_basic.h"

namespace math {
namespace factorize {

TEST(FactorizeBrent, negativeInput) {
  EXPECT_THROW(FactorizeBrent::factorize(-123456),
               std::invalid_argument);
  EXPECT_THROW(FactorizeBrent::factorize(-2),
               std::invalid_argument);
  EXPECT_THROW(FactorizeBrent::factorize(-1),
               std::invalid_argument);
}

TEST(FactorizeBrent, smallBadInput) {
  EXPECT_THROW(FactorizeBrent::factorize(0),
               std::invalid_argument);
  EXPECT_THROW(FactorizeBrent::factorize(1),
               std::invalid_argument);
}

TEST(FactorizeBrent, smallOkInput) {
  std::vector<std::pair<int, int> > expected;
  expected.push_back(std::make_pair(2, 1));
  EXPECT_EQ(expected, FactorizeBrent::factorize(2));

  expected.clear();
  expected.push_back(std::make_pair(3, 1));
  EXPECT_EQ(expected, FactorizeBrent::factorize(3));

  expected.clear();
  expected.push_back(std::make_pair(2, 2));
  EXPECT_EQ(expected, FactorizeBrent::factorize(4));

  expected.clear();
  expected.push_back(std::make_pair(2, 2));
  expected.push_back(std::make_pair(3, 1));
  EXPECT_EQ(expected, FactorizeBrent::factorize(12));
}

const int CNT = 50000;

TEST(FactorizeBrent, increasing) {
  for (int i = 2; i < CNT; i++) {
    std::vector<std::pair<int, int> > result = 
        FactorizeBrent::factorize(i);
    for (size_t j = 1; j < result.size(); j++) {
      EXPECT_TRUE(result[j-1].first < result[j].first);
    }
  }
}

TEST(FactorizeBrent, positiveValues) {
  for (int i = 2; i < CNT; i++) {
    std::vector<std::pair<int, int> > result = 
        FactorizeBrent::factorize(i);
    for (size_t j = 0; j < result.size(); j++) {
      EXPECT_LT(1, result[j].first);
      EXPECT_LT(0, result[j].second);
    }
  }
}


TEST(FactorizeBrent, primes) {
  for (int i = 2; i < CNT; i++) {
    std::vector<std::pair<int, int> > result = 
        FactorizeBrent::factorize(i);
    for (size_t j = 0; j < result.size(); j++) {
      EXPECT_TRUE(math::primes::PrimesBasic::isPrime(result[j].first));
    }
  }
}

TEST(FactorizeBrent, consistency) {
  for (int i = 2; i < CNT; i++) {
    std::vector<std::pair<int, int> > result = 
        FactorizeBrent::factorize(i);
    int tmp = 1;
    for (size_t j = 0; j < result.size(); j++) {
      for (int k = 0; k < result[j].second; k++) {
        tmp *= result[j].first;
      }
    }
    EXPECT_EQ(i, tmp);
  }
}

} // namespace factorize
} // namespace math
