#ifndef H_MATH_FACTORIZE_FACTORIZE_NAIVE
#define H_MATH_FACTORIZE_FACTORIZE_NAIVE
/**
 * @file Contains naive factorization method
 */
#include "utils/preconditions/preconditions.h"
#include <vector>
#include <utility>
#include <limits>
#include <stdexcept>
#include "utils/static_assert/static_assert.h"

namespace math {
namespace factorize {

/**
 * Naive factorization method.
 * Running time is <i> O(m) </i>, where <i> m </i> is 
 * greatest prime factor, or <i> O(sqrt(n)) </i> if the
 * number is prime.
 */
template <typename CountType>
class FactorizeNaive_ {
  public:
  /**
   * Factorize number.
   *
   * @preconditions \a T should be integral type
   * 
   * @param number integer to factorize, should be greater than 1
   *
   * @returns factorization, factors in increasing order
   */
  template <typename T>
  static  std::vector<std::pair<T, CountType> >
      factorize(T number)
  {
    STATIC_ASSERT(std::numeric_limits<T>::is_integer, 
      "T should be integral");
    Preconditions::check(number > 1, "Too small or negative to factorize");
    std::vector<std::pair<T, CountType> > result;

    // Note: following check is a bit relaxed, exact value is
    // sqr(sqrt(max)-1), but it is a way complicated to check
    // This simple check is sufficient
    if (number > std::numeric_limits<T>::max() >> 1) {
      throw std::overflow_error("Sorry, argument is too big to process");
    }
    for (T i = 2; i * i <= number; i++) {
      if (number % i == 0) {
        CountType cnt = 0;
        while (number % i == 0) {
          cnt++;
          number /= i;
        }
        result.push_back(std::make_pair(i, cnt));
      }
    }

    if (number > 1) {
      result.push_back(std::make_pair(number, 1));
    }
    return result;
  }
};

/** Default naive factorization with \a CountType=int */
typedef FactorizeNaive_<int> FactorizeNaive;

} // namespace factorize
} // namespace math
#endif
