#ifndef H_MATH_FACTORIZE_FACTORIZE_WITH_ORACLE
#define H_MATH_FACTORIZE_FACTORIZE_WITH_ORACLE
/**
 * This file holds an implementation of factorizing method with
 * supplied "oracle" which can give one factor for each composite.
 */
#include "utils/preconditions/preconditions.h"
#include <vector>
#include <utility>
#include <limits>
#include <map>
#include <stdexcept>
#include "math/primes/primes_fast.h"
#include "oracle_pollard.h"
#include "oracle_brent.h"

namespace math {
namespace factorize {

// Note: We can't use math::primes::PrimesBasic because it will ruin our
// complexity (usually Oracles are O(n^(1/4)) versus PrimesBasic O(n^(1/2))).
template <typename CountType, class Oracle, class Primes = math::primes::PrimesFast>
class FactorizeWithOracle_ {
  public:
  template <typename T>
  static  std::vector<std::pair<T, CountType> >
      factorize(T number)
  {
    typedef typename std::vector<std::pair<T, CountType> > ResultType;

    Preconditions::check(number > 1, "Too small or negative to factorize");

    std::map<T, CountType> factors;
    while (!Primes::isPrime(number)) {
      T factor = Oracle::findFactor(number);
      // safety checks on factor
      assert(factor > 1);
      assert(factor != number);
      assert(number % factor == 0);
      number /= factor;
      ResultType tmp = factorize(factor);
      for (typename ResultType::size_type i = 0; i < tmp.size(); i++)
        factors[tmp[i].first] += tmp[i].second;
    }
    factors[number]++;
    return ResultType(factors.begin(), factors.end());
  }

};

typedef FactorizeWithOracle_<int, OraclePollard> FactorizePollard;
typedef FactorizeWithOracle_<int, OracleBrent> FactorizeBrent;


} // namespace factorize
} // namespace math
#endif
