#ifndef H_MATH_FACTORIZE_ORACLE_POLLARD
#define H_MATH_FACTORIZE_ORACLE_POLLARD

#include "utils/preconditions/preconditions.h"
#include <vector>
#include <utility>
#include <limits>
#include <stdexcept>
#include "math/powermod/powermod.h"
#include "math/gcd/gcd.h"
#include "utils/rand/rand.h"

namespace math {
namespace factorize {

template <class Powermod>
class OraclePollard_ {
  public:
  template <typename T>
  static T  findFactor(T number)
  {
    Preconditions::check(number > 1, "Too small or negative to factorize");
    Preconditions::check(!math::primes::PrimesFast::isPrime(number),
        "nothing to factorize!");
    if (number == 4) {
      // This is special case, as pollard will fail for any combination of
      // parameters
      return 2;
    }

    const int MAX_ITERATIONS = 1000;
    Rand r(47);
    for (int i = 0; i < MAX_ITERATIONS; i++) {
      T start = r.rand(0, std::min(number, (T) 10000));
      T a = r.rand(0, std::min(number - 1, (T) 10000));
      T res = pollard(number, start, a);
      if (res != 0) {
        assert(res > 1);
        assert(number % res == 0);
        return res;
      }
    }

    throw std::runtime_error("Can't factorize number");
  }
  private:
  template <typename T>
  static T advance(T x, T n, T a) {
    Preconditions::checkRange(x, n);
    x = (Powermod::multmod(x, x, n) + a) % n;
    if (x < 0) x+= n; // fix on negative value overflow
    return x;
  }

  template <typename T>
  static T pollard(T number, T start, T a) {
    T x = start;
    T y = start;
    T d = 1;
    while (d == 1) {
      x = advance(x, number, a);
      y = advance(y, number, a);
      y = advance(y, number, a);
      T diff = (x > y) ? x - y : y - x;
      d = math::gcd::gcd(diff, number);
    }
    if (d == number) {
      return 0; // can't factorize
    }
    return d;
  }

};

typedef OraclePollard_<math::powermod::PowermodExtended> OraclePollard;

} // namespace factorize
} // namespace math
#endif
