#ifndef H_MATH_GCD_EXTENDED_GCD
#define H_MATH_GCD_EXTENDED_GCD
/**
 * @file Contains implementation of extended Euclidian algorithm
 * for computing greatest common divisor
 */

#include <utility>
#include <limits>
#include "utils/static_assert/static_assert.h"
#include "utils/preconditions/preconditions.h"


namespace math {
namespace gcd {

/**
 * Class computing extended GCD using recursion
 */
class ExtendedGCD {
public:
  /**
   * Returns extended gcd, i.e. <i> pair(x, y) <i> such that
   * <i> x * a + y * b == gcd(a, b) </i>
   *
   * Note: that returned values <b> can </b> be negative
   *
   * Note: for both zero arguments function will return (0,0)
   *
   * @warning type \a T can be only a signed integer
   * @warning Note that computation of <i> a*x+b*y </i> may overflow
   *   for partial results but the result of whole computation is correct!
   *
   * @param a non-negative integer 
   * @param b non-negative integer
   *
   * @returns <i>(x,y)</i> such that <i>a*x+b*y==gcd(a,b)</i>
   */
  template <typename T>
  static std::pair<T, T> extended_gcd_positive(T a, T b) {
    STATIC_ASSERT(std::numeric_limits<T>::is_integer, "");
    STATIC_ASSERT(std::numeric_limits<T>::is_signed, "");
    Preconditions::check(a >= 0);
    Preconditions::check(b >= 0);
    if (a == 0 && b == 0) {
      return std::make_pair(0, 0); // special case
    }
    if (b == 0) {
      return std::make_pair(1, 0);
    }
    std::pair<T, T> tmp = extended_gcd_positive(b, a % b);
    //FIXME(nemoze to tu pretiect?)
    return std::make_pair(tmp.second, tmp.first - tmp.second * (a / b));
  }

  /**
   * Returns extended gcd, i.e. <i> pair(x, y) <i> such that
   * <i> x * a + y * b == gcd(a, b) </i>
   *
   * Note: that returned values <b> can </b> be negative
   *
   * Note: for both zero arguments function will return (0,0)
   *
   * @warning type \a T can be only a signed integer
   * @warning Note that computation of <i> a*x+b*y </i> may overflow
   *   for partial results but the result of whole computation is correct!
   *
   * @param a 
   * @param b
   *
   * @returns <i>(x,y)</i> such that <i>a*x+b*y==gcd(a,b)</i>
   */
  template <typename T>
  static std::pair<T, T> extended_gcd(T a, T b) {
    STATIC_ASSERT(std::numeric_limits<T>::is_integer, "");
    STATIC_ASSERT(std::numeric_limits<T>::is_signed, "");
    if (a < 0) {
      if (a == std::numeric_limits<T>::min()) {
        throw std::overflow_error("Overflow in GCD computation.");
      }
      std::pair<T, T> tmp = extended_gcd(-a, b);
      return std::make_pair(-tmp.first, tmp.second);
    }
    if (b < 0) {
      if (b == std::numeric_limits<T>::min()) {
        throw std::overflow_error("Overflow in GCD computation.");
      }
      std::pair<T, T> tmp = extended_gcd(a, -b);
      return std::make_pair(tmp.first, -tmp.second);
    }
    return extended_gcd_positive(a, b);
  }
};

} // namespace gcd
} // namespace math
#endif
