#ifndef H_MATH_GCD_EXTENDED_GCD_LOOP
#define H_MATH_GCD_EXTENDED_GCD_LOOP
/**
 * @file Contains extended GCD computation using
 * loop method
 */

#include "utils/static_assert/static_assert.h"
#include "utils/preconditions/preconditions.h"
#include <utility>

namespace math {
namespace gcd {

/**
 * Class computing extended GCD using loop
 */
class ExtendedGCDLoop {
public:
  /**
   * Returns extended gcd, i.e. <i> pair(x, y) <i> such that
   * <i> x * a + y * b == gcd(a, b) </i>
   *
   * Note: that returned values <b> can </b> be negative
   *
   * Note: for both zero arguments function will return (0,0)
   *
   * @warning \a T can be only to signed integer
   *
   * @param a non-negative integer 
   * @param b non-negative integer
   *
   * @returns <i>(x,y)</i> such that <i>a*x+b*y==gcd(a,b)</i>
   */
  template<typename T>
  static std::pair<T, T> extended_gcd_positive(T a, T b) {
    STATIC_ASSERT(std::numeric_limits<T>::is_integer, "");
    STATIC_ASSERT(std::numeric_limits<T>::is_signed, "");
    Preconditions::check(a >= 0, "a should be non-negative");
    Preconditions::check(b >= 0, "b should be non-negative");

    if ((a == 0) && (b == 0)) {
      return std::make_pair(0, 0); // special case
    }

    T x = 0;
    T lastx = 1;
    T y = 1;
    T lasty = 0;

    while (b) {
        T q = a / b;
        T tmp;

        tmp = b; b = a % b; a = tmp;

        tmp = x; x = lastx - q * x; lastx = tmp;
        tmp = y; y = lasty - q * y; lasty = tmp;
    }
    return std::make_pair(lastx, lasty);
  }
};

} // namespace gcd
} // namespace math
#endif
