#ifndef H_MATH_PRIMES_MODULAR_INVERSE_FERMAT
#define H_MATH_PRIMES_MODULAR_INVERSE_FERMAT
/**
 * This file holds an implementation of the
 * modular inverse computation modulo prime p
 * by using Fermat's little theorem.
 */

#include "math/primes/primes_basic.h"
#include "math/powermod/powermod.h"

namespace math {
namespace modular_inverse {

/**
 * Compute modular inverses using Fermat's little theorem.
 *
 * @precondition PowerModImpl contains static function
 *  \a powermod(base,exponent,modulo)
 *
 * @param PowerModImpl implementation of powermod
 * @param checkPrimality whether to check that the modulo is a prime number
 */
template <typename PowerModImpl, bool checkPrimality = true>
class ModularInverseFermat_ {
  public:
  /**
   * Compute modular inverse of number i over field Z_p.
   *
   * The result is based on the fact, that a^(p-2) = a^(-1) mod p
   * based on Fermat's little theorem.
   *
   * Note: More general case is Euler's theorem:
   * a^(fi(m)-1) = a^(-1) mod m
   * but in this case, calculating fi(m) needs to factorize m and it is
   * therefore inefficient. If you have good implementation of euler's fi,
   * you may tweak this implementation.
   * Anyway, consider using ModularInverseGcd for inverses
   * for non-primes.
   *
   * @param i number for which the inverse should be calculated
   * @param p prime modulus
   *
   * @returns inverse of the number, i.e. number such that
   * (inverse * i) mod p == 1, or zero if such a number does not exists (case i==0)
   */
  template <typename T>
  static T getInverse(T i, T p) {
    Preconditions::check(p > 1);
    Preconditions::checkRange(i, p);

    // * This check is slowing down whole function! *
    // * If you are sure about primality of p, consider    *
    // * using checkPrimality = false                      *
    if (checkPrimality) {
      Preconditions::check(math::primes::PrimesBasic::isPrime(p),
          "ModularInverseFermat needs prime to work!");
    }

    // Note: we can't ignore this check and be happy that powermod(0, p-2, p) == 0
    // as powemod(0, 0, 2) == 1 and that is wrong output in our case!
    if (i == 0) {
      return 0;
    }
    return PowerModImpl::powermod(i, p-2, p);
  }
};

typedef ModularInverseFermat_<math::powermod::PowermodSimple> ModularInverseFermat;

} // namespace modular_inverse
} // namespace math

#endif
