#ifndef H_MATH_PRIMES_MODULAR_INVERSE_GCD
#define H_MATH_PRIMES_MODULAR_INVERSE_GCD
/**
 * This file holds an implementation of the
 * modular inverse computation modulo prime p
 * by using Extended Euclid's algorithm computing
 * greatest common divisor.
 */

#include "math/primes/primes_basic.h"
#include "math/powermod/powermod.h"
#include "math/gcd/extended_gcd.h"

namespace math {
namespace modular_inverse {

/**
 * Computes modular inverse of number over any (even non-prime) modulus.
 * 
 * The implementation is O(1) for initialization (no initialization)
 * and O(log n) for query, but the queries are generally fast.
 *
 * If you need to precompute modular inverses over all elements of field
 * Z_p where p is prime, then you may use ModularInverse
 */
class ModularInverseGcd {
  public:
  /**
   * Compute modular inverse of number i modulo p.
   *
   * @return inverse of the number i or zero if such number does not exists
   */
  template <typename T>
  static T getInverse(T i, T p) {
    Preconditions::check(p > 1);
    Preconditions::checkRange(i, p);
    std::pair<T, T> tmp = math::gcd::ExtendedGCD::extended_gcd(i, p);
    // Note: the following statement may overflow during computation. But this is OK!
    // The overflows in subexpressions will cancel each other!
    T g = tmp.first * i + tmp.second * p;
    if (g == 1) { // the inverse exists
      assert(tmp.first < p);
      assert(tmp.second > -p);
      assert(tmp.first != 0);
      if (tmp.first >= 0) {
        return tmp.first;
      } else {
        return tmp.first + p;
      }
    } else {
      return 0;
    }
  }
};

} // namespace modular_inverse
} // namespace math

#endif
