#ifndef H_MATH_PRIMES_MODULAR_INVERSE_PRECOMPUTED
#define H_MATH_PRIMES_MODULAR_INVERSE_PRECOMPUTED

/**
 * This file holds an implementation of the
 * computation of inverse numbers modulo prime p
 */
#include <vector> 
#include <stack>
#include <utils/preconditions/preconditions.h>
#include <utils/assert/integer_overflow.h>
#include <math/powermod/powermod.h>
#include <math/primes/primes_basic.h>

namespace math {
namespace modular_inverse {

/**
 * Calculate the modular inverses of all numbers modulo <STRONG>prime</STRONG> p
 * in O(p) time and 2*p space.
 *
 * The initialization is in O(p), query is in O(1).
 * The implementation is thus useful for medium-sized p (limited memory), for which
 * you will ask for many different inverses.
 * If you will ask only limited number of questions, but the p is big,
 * it is probably better to use ModularInverseGcd class
 */
template <class PowerModImpl>
class ModularInversePrecomputed_ {
  typedef std::vector<bool>::size_type SizeType;
 public:
  template <typename T>
    void initialize(T t_size) {
      Preconditions::check(math::primes::PrimesBasic::isPrime(t_size),
          "ModularInverse needs prime to work!");
      STATIC_ASSERT_CHECK_INTEGER_OVERFLOW<T, SizeType>();
      SizeType p = (SizeType) t_size;
      // FIXME
      //Preconditions::check(p <= PowerModImpl::max_argument((SizeType) 0));
      inverses.resize(p);

    for (SizeType q = 0; q < p; q++) inverses[q] = 0;
    inverses[1] = 1;
    std::stack<SizeType> subgroup;

    for (SizeType i = 2; i < p; i++) {
        if (inverses[i] != 0) continue; // we have already computed this
        SizeType current = i;
        // go over multiplicatevi subgroup of `current'
        while (inverses[current] == 0) {
            subgroup.push(current);
            current = PowerModImpl::multmod(current, i, p);
        }

        // now go back
        SizeType t = PowerModImpl::multmod(i, inverses[current], p);
        while (!subgroup.empty()) {
            inverses[subgroup.top()] = t;
            subgroup.pop();
            t = PowerModImpl::multmod(i, t, p);
        }
    }
    }

    /**
     * Return inverse of number pos
     */
    template <typename T>
    SizeType getInverse(T t_pos) {
      STATIC_ASSERT_CHECK_INTEGER_OVERFLOW<T, SizeType>();
      SizeType pos = (SizeType) t_pos;
      Preconditions::checkRange(pos, inverses.size());
      return inverses[pos];
    }
  private:
    std::vector<SizeType> inverses;
};

typedef ModularInversePrecomputed_<math::powermod::PowermodSimple> ModularInversePrecomputed;

} // namespace modular_inverse
} // namespace math
#endif
