#ifndef H_MATH_POWERMOD_MULTMOD_EXTENDED
#define H_MATH_POWERMOD_MULTMOD_EXTENDED
/**
 * This file contains fast computation of power of a to b modulo m.
 */

#include "utils/preconditions/preconditions.h"
#include "utils/static_assert/static_assert.h"
#include <limits.h>
#include "utils/macros/unused.h"

namespace math {
namespace powermod {

/**
 * Extended modular multiplication. Does not require bigger
 * types for temporaries.
 *
 * The computation is done by using some type, the drawback is that
 * we need to reserve several bits for our computations.
 *
 * @param shift number of bits reserved for the computation
 */
template <int shift>
class MultmodExtended {
  STATIC_ASSERT(shift > 0, "");

  public:
  template <typename T>
  static T max_argument(T UNUSED(x)) {
    return std::numeric_limits<T>::max() >> (shift);
  }
  /**
   * @precondition T integer type
   */
  template <typename T>
  static T multmod(T a, T b, T modulo) {
    STATIC_ASSERT(std::numeric_limits<T>::is_integer, "")
    // TODO(check this)
    STATIC_ASSERT(shift * 2 < std::numeric_limits<T>::digits,
      "");
    const T max = max_argument((T) 0);
    Preconditions::checkRange(modulo, (T) 1, max);
    Preconditions::checkRange(a, modulo);
    Preconditions::checkRange(b, modulo);
    const T step = ((T)1) << shift;
    const T mask = step - 1;

    T result = 0;
    while (b != 0) {
      result = (result + a * (b & mask)) % modulo;
      a = (a << shift) % modulo;
      b >>= shift;
    }

    return result;
  }
};

/**
 * Optimized version of MultmodExtended
 *
 * This is previous version with fixed shift=1, but
 * we optimized modular arithmetic using only 
 * additions/subtractions
 */
class MultmodExtendedOpt {
  public:
  template <typename T>
  static T max_argument(T UNUSED(x)) {
    return std::numeric_limits<T>::max() >> (1);
  }

  template<typename T>
  static T multmod(T a, T b, T modulo) {
    // We will be using shift-by-one strategy
    const T max = max_argument((T) 0);
    Preconditions::checkRange(modulo, (T) 1, max);
    Preconditions::checkRange(a, modulo);
    Preconditions::checkRange(b, modulo);
    T result = 0;
    while (b != 0) {
      if ((b & 1) == 1) {
        result += a;
      }
      b >>= 1;
      a <<= 1;
      // Note that both result and a may be at most
      // 2*modulo-1 and we may optimize the slow remainder calculation
      if (result > modulo) result -= modulo;
      if (a > modulo) a -= modulo;
    }

    return result;
  }
};

} // namespace powermod
} // namespace math
#endif
