#include "gtest/gtest.h"
#include "powermod.h"
#include "debug/ppdebug.h"

namespace math {
namespace powermod {

/**
 * Tests checking fishy inputs.
 */
TEST(Powermod, checkBadInput) {
  EXPECT_THROW(PowermodSimple::powermod(2,10, 0), std::invalid_argument);
  EXPECT_THROW(PowermodSimple::powermod(2,10, -10), std::invalid_argument);
  EXPECT_THROW(PowermodSimple::powermod(2,-10, 10), std::invalid_argument);
  EXPECT_THROW(PowermodSimple::powermod(10, 10, 10), std::invalid_argument);
  EXPECT_THROW(PowermodSimple::powermod(-1, 10, 10), std::invalid_argument);
  EXPECT_NO_THROW(PowermodSimple::powermod(2,100, 10));
}

/**
 * Is 0^x working correctly?
 */
TEST(Powermod, checkZero) {
  EXPECT_EQ(0, PowermodSimple::powermod(0, 1, 10));
  EXPECT_EQ(0, PowermodSimple::powermod(0, 2, 10));
  EXPECT_EQ(0, PowermodSimple::powermod(0, 7, 10));
  EXPECT_EQ(0, PowermodSimple::powermod(0, 7, 11));
}

TEST(Powermod, checkZeroToZero) {
  EXPECT_EQ(1, PowermodSimple::powermod(0, 0, 10));
  EXPECT_EQ(1, PowermodSimple::powermod(0, 0, 7));
  EXPECT_EQ(1, PowermodSimple::powermod(0, 0, 4));
  EXPECT_EQ(1, PowermodSimple::powermod(0, 0, 3));
  EXPECT_EQ(1, PowermodSimple::powermod(0, 0, 2));
  // very special corner case
  EXPECT_EQ(0, PowermodSimple::powermod(0, 0, 1));
}

/**
 * Is 1^x working correctly?
 */
TEST(Powermod, checkOne) {
  EXPECT_EQ(1, PowermodSimple::powermod(1, 0, 10));
  EXPECT_EQ(1, PowermodSimple::powermod(1, 1, 10));
  EXPECT_EQ(1, PowermodSimple::powermod(1, 2, 10));
  EXPECT_EQ(1, PowermodSimple::powermod(1, 7, 10));
  EXPECT_EQ(1, PowermodSimple::powermod(1, 7, 11));
}

/**
 * Is (-1)^x working correctly?
 */
TEST(Powermod, checkMinusOne) {
  EXPECT_EQ(1, PowermodSimple::powermod(10-1, 0, 10));
  EXPECT_EQ(10-1, PowermodSimple::powermod(10-1, 1, 10));
  EXPECT_EQ(1, PowermodSimple::powermod(10-1, 2, 10));
  EXPECT_EQ(10-1, PowermodSimple::powermod(10-1, 7, 10));
  EXPECT_EQ(11-1, PowermodSimple::powermod(11-1, 7, 11));
}

/**
 * Is 2^x working correctly?
 */
TEST(Powermod, checkTwo) {
  EXPECT_EQ(1, PowermodSimple::powermod(2, 0, 10));
  EXPECT_EQ(2, PowermodSimple::powermod(2, 1, 10));
  EXPECT_EQ(4, PowermodSimple::powermod(2, 2, 10));
  EXPECT_EQ(8, PowermodSimple::powermod(2, 3, 10));
  EXPECT_EQ(6, PowermodSimple::powermod(2, 4, 10));
  EXPECT_EQ(5, PowermodSimple::powermod(2, 4, 11));
}

/**
 * Check little Fermat theorem, i.e. for every prime p
 * and coprime a, a^(p-1) == 1 (mod p)
 */
TEST(Powermod, checkLittleFermat) {
  int primes[] = {2, 3, 5, 7, 11, 13, 17, 19,
    229, 3581, 7927, 17393, 27457, 37831, 0};

  for (int i = 0; primes[i]; i++) {
    int p = primes[i];
    for (int j = 1; j < p; j++) {
      EXPECT_EQ(1, PowermodSimple::powermod(j, p-1, p));
      EXPECT_EQ(j, PowermodSimple::powermod(j, p, p));
    }
  }
}

/**
 * Check some random values
 */
TEST(Powermod, checkRandomValues) {
  int test[][4] = {
      {27, 1729102721, 104, 27},
      {22, 1071853691, 110, 88},
      {434523027, 1126853948, 755342232, 524558121},
      {14415178, 339778747, 100978827, 82574350},
      {700689054, 1654000343, 1079545798, 641613110},
      {370377108, 440497685, 961852825, 401753793},
      {0,0,0,0},
  };

  for (int i = 0; test[i][0] > 0; i++) {
      EXPECT_EQ(test[i][3], 
          PowermodSimple::powermod(test[i][0], test[i][1], test[i][2]));
  }
}

TEST(Powermod, overflow) {
  // test very big values
  EXPECT_EQ(715827882, 
      PowermodSimple::powermod(2147483644, 2147483645, 2147483647));

}

int sqrmod(int x, int modulo) {
  return PowermodSimple::multmod(x, x, modulo);
}

TEST(Sqrmod, checkBadInput) {
  EXPECT_THROW(sqrmod(-2, 10), std::invalid_argument);
  EXPECT_THROW(sqrmod(10, 10), std::invalid_argument);
  EXPECT_THROW(sqrmod(0, 0), std::invalid_argument);
  sqrmod(2, 100);
}

TEST(Sqrmod, checkZero) {
  EXPECT_EQ(0, sqrmod(0, 7));
  EXPECT_EQ(0, sqrmod(0, 6));
  EXPECT_EQ(0, sqrmod(0, 5));
  EXPECT_EQ(0, sqrmod(0, 4));
  EXPECT_EQ(0, sqrmod(0, 3));
  EXPECT_EQ(0, sqrmod(0, 2));
  EXPECT_EQ(0, sqrmod(0, 1));
}

TEST(Sqrmod, checkOne) {
  EXPECT_EQ(1, sqrmod(1, 7));
  EXPECT_EQ(1, sqrmod(1, 6));
  EXPECT_EQ(1, sqrmod(1, 5));
  EXPECT_EQ(1, sqrmod(1, 4));
  EXPECT_EQ(1, sqrmod(1, 3));
  EXPECT_EQ(1, sqrmod(1, 2));
  // corner case
  EXPECT_EQ(0, sqrmod(0, 1));
}

TEST(Sqrmod, checkMinusOne) {
  EXPECT_EQ(1, sqrmod(7-1, 7));
  EXPECT_EQ(1, sqrmod(6-1, 6));
  EXPECT_EQ(1, sqrmod(5-1, 5));
  EXPECT_EQ(1, sqrmod(4-1, 4));
  EXPECT_EQ(1, sqrmod(3-1, 3));
  EXPECT_EQ(1, sqrmod(2-1, 2));
}

TEST(Sqrmod, checkRandomValues) {
  int test[][3] = {
      {547619155, 1283522449, 904973},
      {726537656, 1131612350, 313258686},
      {1366196582, 1759473779, 1179791302},
      {585507447, 649689020, 343504769},
      {616329716, 1678029956, 803725732},
      {0, 0, 0},
  };

  for (int i = 0; test[i][0] > 0; i++) {
      EXPECT_EQ(test[i][2], 
          sqrmod(test[i][0], test[i][1]));
  } 
}

TEST(Sqrmod, overwlow) {
  // check for big values
  EXPECT_EQ(1, sqrmod(2147483646, 2147483647));
  EXPECT_EQ(4, sqrmod(2147483645, 2147483647));
}



} // namespace powermod
} // namespace math
