/**
 * This file implements basic version of eratosthenes sieve.
 */
#include "utils/preconditions/preconditions.h"
#include "utils/assert/integer_overflow.h"
#include <limits>

namespace math {
namespace prime_sieve {

/**
 * The basic implementation of eratosthenes sieve.
 *
 * Usage:
 * EratosthenesBasic e;
 * e.initialize(100);
 * cout << "7:"<< e.isPrime(7) << " 10:" << e.isPrime(10);
 */
class EratosthenesBasic {
  /** Internal type of the size index */
  // Note: using upper-half of the range of SizeType is ok
  // as in this class we won't use difference type.
  typedef std::vector<bool>::size_type SizeType;

  public:
    /**
     * Initialize eratosthenes sieve to use size up to 
     * specified size.
     *
     * @param T size size of the input
     */
    template <typename T>
    void initialize(T t_size) {
      STATIC_ASSERT_CHECK_INTEGER_OVERFLOW<T, SizeType>();
      SizeType size = (SizeType) t_size;
      Preconditions::check(size > 2);
      // because of j = 2 * i
      Preconditions::check(size < std::numeric_limits<SizeType>::max() / 2,
          "Out of range for computation, there will be overflow");
      // and now we are ready to go
      data.resize(size);
      data[0] = false;
      data[1] = false;
      for (SizeType i = 2; i < size; i++) {
        data[i] = true;
      }

      for (SizeType i = 2; i < size; i++) if (data[i]) {
        for (SizeType j = 2 * i; j < size; j += i)
          data[j] = false;
      }
    }

    /**
     * @param p - number to test
     * @returns true if p is prime
     */
    template <typename T>
    bool isPrime(T p) {
      STATIC_ASSERT_CHECK_INTEGER_OVERFLOW<T, SizeType>();
      Preconditions::checkRange((SizeType) p, data.size());
      return data[(SizeType) p];
    }

  private:
    std::vector<bool> data;
};

} // namespace prime_sieve
} // namespace math
