/**
 * This file implements basic version of eratosthenes sieve.
 */
#include "utils/preconditions/preconditions.h"
#include <vector>
#include <algorithm>
#include <cmath>

namespace math {
namespace prime_sieve {

// Optimized implementation of Eratosthenes sieve.
// Running time is O(n log log n).
class EratosthenesOptimized {
  public:
    void initialize(int size_) {
      Preconditions::check(size_ > 2);
      // Note: this won't set the elements to true on downsizing
      // only sets new elements when upsizing.
      // But in this algorithm, it is ok!
      data.resize(size_ / 2 + 1, true);
      size = size_;

      data[0] = false;

      // Note: size >= 3 implies sqrt(size) + 1 < size
      // and we are safe when indexing data[i]
      int sqrtsize = (int) sqrt(size) + 1;

      for (int i = 1; i < sqrtsize; i++)
        if (data[i]) { // eliminate only from primes
          // we can start eliminating from p^2 = (2 * i + 1)^2
          // which corresponds to data[2 * i^2 + 2 * i]
          for (int j = 2 * i * (i + 1); j < size / 2; j += 2 * i + 1) {
            data[j] = false;
          }
        }
    }

    // Note: you may query for range [0, size)
    bool isPrime(int p) {
      Preconditions::checkRange(p, size);
      if (p % 2 == 0) {
        return p == 2 ? true : false;
      }
      return data[p/2];
    }

  private:
    // Note: if you know maximum range in advance, use bitset<MAX/2>
    // instead of vector<bool>;
    // Internal representation is that data contains only info
    // about primality of odd numbers so the correspondence is
    // data[0]<->1, data[1]<->3, data[2]<->5, ...
    // @see isPrime implementation
    std::vector<bool> data;
    int size;
};

} // namespace prime_sieve
} // namespace math
