#include "gtest/gtest.h"
#include "eratosthenes_basic.h"
#include "eratosthenes_optimized.h"
#include <set>

namespace math {
namespace prime_sieve {

TEST(EratosthenesBasic, smallInitializaton) {
  // just check, that we don't allow small initializers
  EXPECT_THROW(EratosthenesBasic e; e.initialize(0u),
               std::invalid_argument);

  EXPECT_THROW(EratosthenesBasic e; e.initialize(1u),
               std::invalid_argument);

  EXPECT_THROW(EratosthenesBasic e; e.initialize(2u),
               std::invalid_argument);

  EXPECT_NO_THROW(EratosthenesBasic e; e.initialize(3u));

  EXPECT_NO_THROW(EratosthenesBasic e; e.initialize(4u));
}


TEST(EratosthenesBasic, invalidRange) {
  EXPECT_THROW(EratosthenesBasic e; e.initialize(0u),
               std::invalid_argument);
  //Compile time:
  //EXPECT_ANY_THROW(e.initialize(-1));
  //EXPECT_ANY_THROW(e.initialize(-42));
  EXPECT_NO_THROW(EratosthenesBasic e; e.initialize(47u));
}

TEST(EratosthenesBasic, smallNumbers) {
  EratosthenesBasic e;
  e.initialize(10u);
  EXPECT_FALSE(e.isPrime(1u));
  EXPECT_TRUE(e.isPrime(2u));
  EXPECT_TRUE(e.isPrime(3u));
  EXPECT_FALSE(e.isPrime(4u));
  EXPECT_TRUE(e.isPrime(5u));
}

TEST(EratosthenesBasic, bigNumbers) {
  EratosthenesBasic e;
  e.initialize(105000u);

  unsigned int low = 104729;
  unsigned int high = 104774;
  std::set<int> primes;
  primes.insert(104729u);
  primes.insert(104743u);
  primes.insert(104759u);
  primes.insert(104761u);
  primes.insert(104773u);

  for (unsigned int i = low; i<high; i++) {
    EXPECT_EQ(primes.find(i) != primes.end(),
        e.isPrime(i));
  }
}

TEST(EratosthenesOptimized, smallInitializaton) {
  // just check, that we don't allow small initializers
  EXPECT_THROW(EratosthenesOptimized e; e.initialize(0),
               std::invalid_argument);

  EXPECT_THROW(EratosthenesOptimized e; e.initialize(1),
               std::invalid_argument);

  EXPECT_THROW(EratosthenesOptimized e; e.initialize(2),
               std::invalid_argument);

  EXPECT_NO_THROW(EratosthenesOptimized e; e.initialize(3));
  EXPECT_NO_THROW(EratosthenesOptimized e; e.initialize(4));
}

TEST(EratosthenesOptimized, invalidRange) {
  EXPECT_THROW(EratosthenesOptimized e; e.initialize(0),
               std::invalid_argument);

  EXPECT_THROW(EratosthenesOptimized e; e.initialize(-1),
               std::invalid_argument);

  EXPECT_THROW(EratosthenesOptimized e; e.initialize(-42),
               std::invalid_argument);

  EXPECT_NO_THROW(EratosthenesOptimized e; e.initialize(47));
}

TEST(EratosthenesOptimized, smallNumbers) {
  EratosthenesOptimized e;
  e.initialize(10);
  EXPECT_FALSE(e.isPrime(1));
  EXPECT_EQ(true, e.isPrime(2));
  EXPECT_EQ(true, e.isPrime(3));
  EXPECT_FALSE(e.isPrime(4));
  EXPECT_EQ(true, e.isPrime(5));
}

// implementation-specific test
TEST(EratosthenesOptimized, oddInitializerRangeCheck) {
  EratosthenesOptimized e;
  e.initialize(7);
  EXPECT_EQ(true, e.isPrime(5));
  EXPECT_FALSE(e.isPrime(6));
  EXPECT_THROW(e.isPrime(7), std::invalid_argument); // this is out of range
}

// implementation-specific test
TEST(EratosthenesOptimized, evenInitializerRangeCheck) {
  EratosthenesOptimized e;
  e.initialize(8);
  EXPECT_EQ(true, e.isPrime(5));
  EXPECT_FALSE(e.isPrime(6));
  EXPECT_EQ(true, e.isPrime(7));
  EXPECT_THROW(e.isPrime(8), std::invalid_argument); // this is out of range
}

TEST(EratosthenesOptimized, bigNumbers) {
  EratosthenesOptimized e;
  e.initialize(105000);

  int low = 104729;
  int high = 104774;
  std::set<int> primes;
  primes.insert(104729);
  primes.insert(104743);
  primes.insert(104759);
  primes.insert(104761);
  primes.insert(104773);

  for (int i = low; i<high; i++) {
    EXPECT_EQ(primes.find(i) != primes.end(),
        e.isPrime(i));
  }
}



TEST(EratosthenesBasic, badQuery) {
  EratosthenesBasic e;
  e.initialize(100u);

  //Following is compile-time
  //EXPECT_ANY_THROW(e.isPrime(-1));

  EXPECT_THROW(e.isPrime(101u), std::invalid_argument);
  EXPECT_THROW(e.isPrime(1000u), std::invalid_argument);
}

} // namespace prime_sieve
} // namespace math
