/**
 * This file implements basic segmented eratosthenes sieve.
 * It's purpose is to find all primes up to n with
 * space complexity O(sqrt(n)).
 *
 * Implementation is based on paper
 *   The Segmented Sieve of Eratosthenes
 *   and Primes in Arithmetic Progressions to 10^18
 * by Carter Bays and Richard H. Hudsom
 *
 * We used mainly algorithm B from this article.
 */
#include "utils/preconditions/preconditions.h"
#include "debug/ppdebug.h"
#include "template/template.h"
#include "math/prime_sieve/eratosthenes_basic.h"

namespace math {
namespace prime_sieve {

/**
 * Because we can't store all primes found in segmented sieve
 * due to space complexity of result, we need way to
 * "stream" primes.
 */
class SieveCallback {
 public:
  virtual void foundPrime(long long int p) = 0;
};

/**
 * Implementation of segmented sieve
 *
 * The basic idea is to
 * a) find all primes up to sqrt(n) (inclusive)
 * b) sieve sucessive intervals of length sqrt(n)
 *    - each interval can be sieved in the same way
 *      as in normal sieve, we just need to maintain
 *      for each prime[i] position L[i] of last crossed-out number
 */
class SegmentedSieve {
  private:
    /**
     * Interval sieving
     */
    static void sieve(
        const vector<int> &primes,
        long long int segment_start,
        long long int delta,
        long long int n,
        SieveCallback* callback) {
      Preconditions::check(segment_start % 2 == 1); 
      Preconditions::check(segment_start > 0.1 + sqrt(delta));
      // segment size is 2 * delta as we do not bother to sieve even values

      vector<long long int> L(primes.size());
      // Using vector<char> over vector<bool> gains 20% performance
      // and we do not need so tight memory optimization
      vector<char> sieve(delta, true);

      // prepare values of L
      FOR(i, primes.size()) {
        long long int p = primes[i];
        L[i] = (p - ((segment_start - p)/2) % p) % p;
      }


      while (segment_start < n) {
        // sieve all values in range [segment_start, segment_start + 2 * delta)
        FOR(i, primes.size()) {
          while (L[i] < delta) {
            sieve[L[i]] = false;
            L[i] += primes[i];
          }
          L[i] -= delta; // prepare L for next segment
        }

        for (int i = 0; i < delta; i++) {
          if (sieve[i]) {
            long long int p = segment_start + 2 * i;
            if (p >= n) break; // we should return only primes < n
            callback->foundPrime(segment_start + 2 * i);
          } else {
            sieve[i] = true; // reset sieve array to true
          }
        }
        segment_start += 2 * delta;
      }

    }

  public:
    /**
     * Find all primes less than n and report them to the supplied
     * callback
     *
     * @param n 
     * @param callback use this callback on each found prime
     */
    static void findPrimes(long long int n, SieveCallback* callback) {
      Preconditions::check(n > 2);
      callback->foundPrime(2);

      // find delta >= sqrt(n)

      long long int start = 1 + (int) sqrt((double) n);
      if (start % 2 == 0) start++; // we need odd number

      // find all odd primes < start
      // (i.e. all primes which may factor numbers up to n)
      EratosthenesBasic e;
      // FIXME(ppershing): nepretecie?
      e.initialize((unsigned int) start);

      vector<int> primes;
      for (int i = 3; i < start; i += 2 ) {
        if (e.isPrime((unsigned) i)) {
          callback->foundPrime(i);
          primes.push_back(i);
        }
      }

      long long int delta = start;
      sieve(primes, start, delta, n, callback);
    };
};

} // namespace prime_sieve
} // namespace math
