#include "gtest/gtest.h"
#include "utils/benchmark/benchmark.h"
#include "math/prime_sieve/eratosthenes_basic.h"
#include "math/prime_sieve/eratosthenes_optimized.h"
#include "primes_slow.h"
#include "primes_basic.h"
#include "primes_fast.h"
#include "utils/rand/rand.h"
#include "utils/si_units/si_units.h"

using namespace math::prime_sieve;

namespace math {
namespace primes {

template <class T>
void interval(int low, int high) {
  for (int i = low; i < high; i++) {
    T::isPrime(i);
  }
}

TEST(BenchmarkPrimes, slow) {
  BENCHMARK(100000, interval<PrimesSlow>(10, 100));
  BENCHMARK(2000, interval<PrimesSlow>(100, 1000));
  BENCHMARK(100, interval<PrimesSlow>(1000, 10000));
}

TEST(BenchmarkPrimes, basic) {
  BENCHMARK(100000, interval<PrimesBasic>(10, 100));
  BENCHMARK(10000, interval<PrimesBasic>(100, 1000));
  BENCHMARK(500, interval<PrimesBasic>(1000, 10000));
  BENCHMARK(200, interval<PrimesBasic>(10000, 20000));
  BENCHMARK(100, interval<PrimesBasic>(100*Ki, 100*Ki +10000));
  BENCHMARK(50, interval<PrimesBasic>(1*Mi, 1*Mi + 10000));
  BENCHMARK(20, interval<PrimesBasic>(10*Mi, 10*Mi + 10000));
  BENCHMARK(5, interval<PrimesBasic>(100*Mi, 100*Mi + 10000));
  BENCHMARK(2, interval<PrimesBasic>(1000*Mi, 1000*Mi + 10000));
}


TEST(BenchmarkPrimes, fast) {
  BENCHMARK(100000, interval<PrimesFast>(10, 100));
  BENCHMARK(10000, interval<PrimesFast>(100, 1000));
  BENCHMARK(1000, interval<PrimesFast>(1000, 10000));
  BENCHMARK(500, interval<PrimesFast>(10000, 20000));
  BENCHMARK(500, interval<PrimesFast>(100*Ki, 100*Ki + 10000));
  BENCHMARK(200, interval<PrimesFast>(1*Mi, 1*Mi + 10000));
  BENCHMARK(200, interval<PrimesFast>(10*Mi, 10*Mi +10000));
  BENCHMARK(100, interval<PrimesFast>(100*Mi, 100*Mi +10000));
  BENCHMARK(100, interval<PrimesFast>(1000*Mi, 1000*Mi +10000));
}

template <class T>
void randomValues(int max, int cnt) {
  Rand rand(47);
  for (int i= 0; i < cnt; i++) {
    T::isPrime((unsigned int) rand.rand(2,max));
  }
}


TEST(Compare, primes) {
  BENCHMARK(2, randomValues<PrimesFast>(1 * Ki, Mi));
  BENCHMARK(2, randomValues<PrimesFast>(10 * Ki, Mi));
  BENCHMARK(2, randomValues<PrimesFast>(100 * Ki, Mi));
  BENCHMARK(2, randomValues<PrimesFast>(1 * Mi, Mi));
  BENCHMARK(1, randomValues<PrimesFast>(10 * Mi, Mi));
  BENCHMARK(1, randomValues<PrimesFast>(100 * Mi, Mi));
  BENCHMARK(1, randomValues<PrimesFast>(1000 * Mi, Mi));

  BENCHMARK(10, randomValues<PrimesBasic>(1 * Ki, Mi));
  BENCHMARK(1, randomValues<PrimesBasic>(10 * Ki, Mi));
  BENCHMARK(1, randomValues<PrimesBasic>(100 * Ki, Mi));
  BENCHMARK(1000, randomValues<PrimesBasic>(1 * Mi, Ki));
  BENCHMARK(100, randomValues<PrimesBasic>(10 * Mi, Ki));
  BENCHMARK(100, randomValues<PrimesBasic>(100 * Mi, Ki));

  BENCHMARK(5000, randomValues<PrimesSlow>(1 * Ki, Ki));
  BENCHMARK(500, randomValues<PrimesSlow>(10 * Ki, Ki));
  BENCHMARK(50, randomValues<PrimesSlow>(100 * Ki, Ki));
  BENCHMARK(5, randomValues<PrimesSlow>(1 * Mi, Ki));
  BENCHMARK(1, randomValues<PrimesSlow>(10 * Mi, Ki));

}

template <class T>
void sieveInit(unsigned int size) {
  EratosthenesBasic e;
  e.initialize(size);
}


TEST(BenchmarkErastothenes, basicInit) {
  BENCHMARK(200000, sieveInit<EratosthenesBasic>(100));
  BENCHMARK(20000, sieveInit<EratosthenesBasic>(1000));
  BENCHMARK(2000, sieveInit<EratosthenesBasic>(10*Ki));
  BENCHMARK(200, sieveInit<EratosthenesBasic>(100*Ki));
  BENCHMARK(10, sieveInit<EratosthenesBasic>(1*Mi));
  BENCHMARK(10, sieveInit<EratosthenesBasic>(10*Mi));
  BENCHMARK(1, sieveInit<EratosthenesBasic>(100*Mi));
}

TEST(BenchmarkErastothenes, optimizedInit) {
  BENCHMARK(1000000, sieveInit<EratosthenesOptimized>(100));
  BENCHMARK(100000, sieveInit<EratosthenesOptimized>(1000));
  BENCHMARK(10000, sieveInit<EratosthenesOptimized>(10*Ki));
  BENCHMARK(1000, sieveInit<EratosthenesOptimized>(100*Ki));
  BENCHMARK(100, sieveInit<EratosthenesOptimized>(1*Mi));
  BENCHMARK(10, sieveInit<EratosthenesOptimized>(10*Mi));
  BENCHMARK(1, sieveInit<EratosthenesOptimized>(100*Mi));
}


template <class T>
void sieveRandomQuery(unsigned int size, int queries) {
  Rand rand(47);
  EratosthenesBasic e;
  e.initialize(size);
  for (int q=0; q < queries; q++) {
    e.isPrime((unsigned int) rand.rand(0, size));
  }
}

TEST(BenchmarkErastothenes, queries) {
  BENCHMARK(1, sieveRandomQuery<EratosthenesBasic>(20*Mi, 100*Mi));
  BENCHMARK(1, sieveRandomQuery<EratosthenesOptimized>(20*Mi, 100*Mi));
}

}
}
