#ifndef H_MATH_PRIMES_PRIMES_BASIC
#define H_MATH_PRIMES_PRIMES_BASIC
/**
 * @file Contains a trivial implementation of primality test.
 * It's running time is O(sqrt(n))
 */
#include "utils/preconditions/preconditions.h"
namespace math {
namespace primes {

/**
 * Primality testing using trial division up to square root
 */
class PrimesBasic {
public:
  /** Basic type for all computations */
  typedef long long int BaseType;

  /**
   * Checks whether a number is a prime number.
   *
   * @param p number to test
   *
   * @returns true iff p is a prime
   */
  static bool isPrime(BaseType p) {
    Preconditions::check(p > 0, "tested integer should be positive!");
    // Note: following check is to ensure that i*i in the
    // rest of the code do not overflow.
    // Hovewer, this check is very rough and can be relaxed a bit.
    if (p >= std::numeric_limits<BaseType>::max() / 2) {
      throw std::overflow_error("tested integer is too big!");
    }
    if (p == 1) {
      return false;
    }
    BaseType i = 2;
    // This is reformulation of i <= sqrt(p)
    // We do this transformation for two reasons
    // a) (int) sqrt(n*n) where n is prime may give wrong result if badly
    // rounded
    // b) 64bit number doesn't fit into double and we may loose some digits!
    while (i * i <= p) {
      if (p % i == 0) {
        return false;
      }
      i++;
    }
    // now i > sqrt(p)
    return true;
  }
};

} // namespace primes
} // namespace math
#endif
