#ifndef H_MATH_PRIMES_PRIMES_FAST
#define H_MATH_PRIMES_PRIMES_FAST
/**
 * @file Contains deterministic version of Miller-Rabbin primality test.
 */
#include "utils/preconditions/preconditions.h"
#include "math/powermod/powermod.h"
#include "utils/static_assert/static_assert.h"
#include <limits>

namespace math {
namespace primes {

/**
 * Deterministic version of Miller-Rabbin primality test.
 *
 * The implementation is based on article
 * ON STRONG PSEUDOPRIMES TO SEVERAL BASES
 * by GERHARD JAESCHKE
 * It's running time is O(log(n)^2) (multiplication is supposed to be O(log n))
 *
 * @param PowerModImpl implementation of modular exponentiation
 *
 * @precondition PowerModImpl have functions
 * <ul>
 *   <li> powermod(base, exponent, modulus)
 *   <li> multmod(a, b, modulus)
 * </ul>
 * @precondition long long int is 64bit
 */
template <class PowerModImpl>
class PrimesFast_ {
  /** Base type for our computation, at least 50 bits */
  typedef long long int BaseType;
  // check base type
  STATIC_ASSERT(std::numeric_limits<BaseType>::digits > 50,
      "We need at least 50-bit integer as a base tyep");

  public:
  /**
   * Miller-Rabin test with fixed withesses which is correct
   * up to 3*10^14. (but the implementation is only for integers.)
   *
   * @param p number to test, should be positive
   *
   * @returns true iff \a p is prime
   */
  static bool isPrime(BaseType p) {
    // theoretical limit of the test based on original article
    const BaseType TEST_MAX = 341550071728321LL;
    const BaseType witnesses[] = {2, 3, 5, 7, 11, 13, 17, 0};

    Preconditions::check(p > 0, "tested integer should be positive!");
    Preconditions::check(p < TEST_MAX,
        "Sorry, but we support only calculations up to 3*10^14");

    // small trivial cases
    if (p == 2) {
      return true;
    } else if (p == 1 || p % 2 == 0) {
      return false;
    }

    // p is odd, now we want to write p-1 = 2^s * d
    BaseType s = 0;
    BaseType d = p - 1;
    while (d % 2 == 0) {
      s++;
      d /= 2;
    }
    // Now apply Miller-Rabin algorithm
    // witnesses up to 4*10^9
    for (int i = 0; witnesses[i] != 0 && witnesses[i] < p; i++) {
      BaseType a = witnesses[i];
      bool ok;
      /* According to the article
         When n = 1 + 2^h * d with d odd and h > 0
         and when n is a composite number,
         then n is called a "strong pseudoprime to base b"
         if either
          (2a) b^d = 1 mod n                or
          (2b) b^{2^k*d} = -l mod n   for 0 <= k < h
      */
      BaseType t = PowerModImpl::powermod(a, d, p);
      ok = (t == 1);
      for (int r = 0; (r < s) && !ok; r++) {
        ok |= (t == p-1);
        // Optimization of computation (2b) by iterative squaring
        t = PowerModImpl::multmod(t, t, p);
      }

      if (!ok) {
        return false;
      }
    }
    return true;
  }
};

// Can't use simple powermod as we are using 64bit integers.
typedef PrimesFast_<math::powermod::PowermodExtended> PrimesFast;

} // namespace primes
} // namespace math
#endif
