#ifndef H_MATH_RATIONAL_RATIONAL
#define H_MATH_RATIONAL_RATIONAL
/**
 * @file Contains implementation of rational numbers
 */
#include <ostream>
#include <stdexcept>
#include <stdint.h>
#include "math/gcd/gcd.h"
#include "utils/preconditions/preconditions.h"

#define NEEDS_INT_DEFINED
#include "third_party/safeint3.hpp"


/**
 * This pragma is for removal of compile warnings
 * for "denominator < 0" when denominator is unsigned!
 */
#pragma GCC diagnostic ignored "-Wtype-limits"
namespace math {
namespace rational {

template <typename T>
class Rational {
 public:
  // compile assert
  C_ASSERT(NumericType<T>::isInt);

  /**
   * Constructor, creates zero
   */
  Rational():num(0),den(1) {
  }

  /**
   * Constructor,
   * @param value integer value of the rational number
   */
  Rational(T value):num(value),den(1) {
    normalize();
  }

  /**
   * Constructor
   *
   * @param n numerator
   * @param d denominator
   */
  Rational(T n, T d):num(n),den(d) {
    Preconditions::check(d != 0,
        "Rational number shoud have non-zero denominator");

    normalize();
  }

  /**
   * Copy constructor
   */
  Rational(const Rational<T>& value):num(value.num),den(value.den) {
    Preconditions::check(value.den != 0,
        "Rational numer shoud have non-zero denominator");
    normalize();
  }

  /**
   * @returns inverse number 1/x
   */
  Rational<T> inverted() const {
    if (num == 0) {
      throw std::overflow_error("Division by zero!");
    }
    return Rational<T>(den, num);
  }

  /**
   * @returns numerator
   */
  T numerator() const {
    return num;
  }

  /**
   * @returns denominator
   */
  T denominator() const {
    return den;
  }

  /**
   * Normalize rational number. After this operation,
   * numerator and demoninator will be coprime and
   * denominator will be positive.
   */
  void normalize() {
    T h = math::gcd::gcd(num, den);
    assert(h != 0); // gcd shouldn't return zero (den != 0)
    num /= h;
    den /= h;

    if (IntTraits<T>::isSigned && (den < 0)) {
      den = -den;
      num = -num;
    }
  }


  private:
    /** numerator */
    T num;
    /** denominator */
    T den;
};


/**
 * Multiply two fractions
 *
 * Note: there exists slight optimization of multiplication - for normalized
 * fractions (a/b) and (c/d) multiplication (a/b) * (c/d)
 * may be written as normalize(c/b) * normalize(a/d)
 * hovewer we will not use it here.
 *
 * @related Rational
 *
 * @returns a*b
 */
template<typename T>
Rational<T> operator*(const Rational<T> &a, const Rational<T> &b)
{
  // here we use safe versions of computations
  // if you believe your code will won't need them, just make it as
  // normal multiplication (but it is not recommended)
  try {
    SafeInt<T> num(a.numerator()); num *= b.numerator();
    SafeInt<T> den(a.denominator()); den *= b.denominator();
    return Rational<T>((T)num, (T)den);
  } catch (SafeIntException err){
    throw std::overflow_error("Overflow while multiplying");
  }
}

/**
 * Divide two fractions
 *
 * @related Rational
 *
 * @returns a / b
 */
template<typename T>
Rational<T> operator/(const Rational<T> &a, const Rational<T> &b)
{
  if (b.numerator() == 0) {
    throw std::overflow_error("Division by zero!");
  }

  return a * b.inverted();
}

/**
 * Unary minus
 *
 * @returns -a
 */
template<typename T>
Rational<T> operator-(const Rational<T> &a)
{
  // can't do unary minus on unsigned types
  C_ASSERT(IntTraits<T>::isSigned);
  // we need SafeInt here because unary minus on
  // MIN_INT will overflow.
  try {
    return Rational<T>(-SafeInt<T>(a.numerator()), a.denominator());
  } catch (SafeIntException err){
    throw std::overflow_error("Overflow while add/subtract");
  }
}

/**
 * Add two fractions
 *
 * @related Rational
 *
 * @returns a + b
 */
template<typename T>
Rational<T> operator+(const Rational<T> &a, const Rational<T> &b)
{
  try {
    SafeInt<T> t1(a.numerator()); t1 *= b.denominator();
    SafeInt<T> t2(b.numerator()); t2 *= a.denominator();
    SafeInt<T> den(a.denominator()); den *= b.denominator();
    return Rational<T>( (T) (t1 + t2), den);
  } catch (SafeIntException err){
    throw std::overflow_error("Overflow while add/subtract");
  }
}

/**
 * Subtract two fractions
 *
 * @related Rational
 *
 * @returns a - b
 */
template<typename T>
Rational<T> operator-(const Rational<T> &a, const Rational<T> &b)
{
  try {
    SafeInt<T> t1(a.numerator()); t1 *= b.denominator();
    SafeInt<T> t2(b.numerator()); t2 *= a.denominator();
    SafeInt<T> den(a.denominator()); den *= b.denominator();
    return Rational<T>( (T) (t1 - t2), den);
  } catch (SafeIntException err){
    throw std::overflow_error("Overflow while add/subtract");
  }
}

/**
 * Test equality of fractions
 *
 * @precondition Fractions should be normalized first, this is done
 * automatically unless you modify fraction's member variables
 *
 * @returns true iff \a a is equal to \a b
 */
template<typename T>
bool operator==(const Rational<T> &a, const Rational<T> &b) {
  return (a.numerator() == b.numerator()) && (a.denominator() == b.denominator());
}

/**
 * Test inequality of fractions
 *
 * @returns true iff \a a is less than \a b
 */
template<typename T>
bool operator<(const Rational<T> &a, const Rational<T> &b) {
  try {
    SafeInt<T> t1(a.numerator()); t1 *= b.denominator();
    SafeInt<T> t2(b.numerator()); t2 *= a.denominator();
    return t1 < t2; 
  } catch (SafeIntException err){
    throw std::overflow_error("Overflow while comparison");
  }
}

/**
 * Test inequality of fractions
 *
 * @returns true iff \a a is greater than \a b
 */
template<typename T>
bool operator>(const Rational<T> &a, const Rational<T> &b) {
  return b < a;
}

/**
 * Test inequality of fractions
 *
 * @returns true iff \a a is less or equal than \a b
 */
template<typename T>
bool operator<=(const Rational<T> &a, const Rational<T> &b) {
  return !(b>a);
}

/**
 * Test inequality of fractions
 *
 * @returns true iff \a a is greated or equals to \a b
 */
template<typename T>
bool operator>=(const Rational<T> &a, const Rational<T> &b) {
  return !(a<b);
}

/**
 * Write fraction to stream in form "a/b" where a/b is normalized fraction
 *
 * @precondition Fraction should be normalized, this hold unless you
 * modify class members directly.
 *
 * @param out stream to be used
 * @param a number to be written
 *
 * @returns stream \a out
 */
template<typename T>
std::ostream& operator<<(std::ostream& out, const Rational<T> &a) {
  out << a.numerator() << "/" << a.denominator();
  return out;
}

} // rational
} // math
#endif
