#include "gtest/gtest.h"
#include "rational.h"
#include <stdint.h>
namespace math {
namespace rational {

TEST(Rational, construct)
{
  EXPECT_NO_THROW(Rational<int32_t> r1(2,5));
  EXPECT_NO_THROW(Rational<int32_t> r2(-2,-5));
  EXPECT_NO_THROW(Rational<int32_t> r3(0,-5));

  // check for division by zero
  EXPECT_THROW(Rational<int32_t>(-2,0), std::invalid_argument);
}


TEST(Rational, constructNormalizedPositiveDenominator)
{
  EXPECT_EQ(2, Rational<int32_t>(2, 5).numerator());
  EXPECT_EQ(5, Rational<int32_t>(2, 5).denominator());

  EXPECT_EQ(-2, Rational<int32_t>(-2, 5).numerator());
  EXPECT_EQ(5, Rational<int32_t>(-2, 5).denominator());


  EXPECT_EQ(0, Rational<int32_t>(0, 100).numerator());
  EXPECT_EQ(1, Rational<int32_t>(0, 100).denominator());


  EXPECT_EQ(1, Rational<int32_t>(30, 90).numerator());
  EXPECT_EQ(3, Rational<int32_t>(30, 90).denominator());

  EXPECT_EQ(-1, Rational<int32_t>(-30, 90).numerator());
  EXPECT_EQ(3, Rational<int32_t>(-30, 90).denominator());

}

TEST(Rational, constructNormalizedNegativeDenominator)
{
  EXPECT_EQ(-2, Rational<int32_t>(2, -5).numerator());
  EXPECT_EQ(5, Rational<int32_t>(2, -5).denominator());

  EXPECT_EQ(2, Rational<int32_t>(-2, -5).numerator());
  EXPECT_EQ(5, Rational<int32_t>(-2, -5).denominator());

  EXPECT_EQ(0, Rational<int32_t>(0, -100).numerator());
  EXPECT_EQ(1, Rational<int32_t>(0, -100).denominator());

  EXPECT_EQ(-3, Rational<int32_t>(60, -40).numerator());
  EXPECT_EQ(2, Rational<int32_t>(60, -40).denominator());

  EXPECT_EQ(3, Rational<int32_t>(-60, -40).numerator());
  EXPECT_EQ(2, Rational<int32_t>(-60, -40).denominator());
}

TEST(Rational, multiply)
{
  EXPECT_EQ(Rational<int32_t>(0,1),
            Rational<int32_t>(0,100)*Rational<int32_t>(5,-10));

  EXPECT_EQ(Rational<int32_t>(1,1),
            Rational<int32_t>(1,-100)*Rational<int32_t>(-100,1));

  EXPECT_EQ(Rational<int32_t>(1,6),
            Rational<int32_t>(1, 2)*Rational<int32_t>(1, 3));

  EXPECT_EQ(Rational<int32_t>(4, 7),
            Rational<int32_t>(3, 14)*Rational<int32_t>(8, 3));
}

TEST(Rational, multiplyOverflowAssert)
{
  // clearly off
  EXPECT_THROW(Rational<int32_t>(1,1000000) * Rational<int32_t>(1,10000000),
               std::overflow_error);

  // sign change
  EXPECT_THROW(Rational<int32_t>(1,7919) * Rational<int32_t>(1,271883),
               std::overflow_error);

  EXPECT_THROW(Rational<int32_t>(1000000,1) * Rational<int32_t>(100000000,1),
                   std::overflow_error);

  EXPECT_THROW(Rational<int32_t>(7919, 1) * Rational<int32_t>(271883,1),
                   std::overflow_error);

  // check that uint32_t works well
  EXPECT_EQ(2147483651u, (Rational<uint32_t>(20261, 1) *
                          Rational<uint32_t>(105991, 1)).numerator());

  EXPECT_THROW(Rational<int32_t>(20261, 1) *
                    Rational<int32_t>(105991, 1),
                   std::overflow_error);
}

TEST(Rational, inverse) {
  EXPECT_EQ(Rational<int32_t>(3,2),
            Rational<int32_t>(2,3).inverted());

  EXPECT_EQ(Rational<int32_t>(-4,7),
            Rational<int32_t>(-7,4).inverted());

  // division by zero is reported as overflow_error!
  EXPECT_THROW(
            Rational<int32_t>(0,1).inverted(),
            std::overflow_error);

  EXPECT_EQ(Rational<int32_t>(1,2),
            Rational<int32_t>(-2,-1).inverted());
}


TEST(Rational, divide)
{
  EXPECT_EQ(Rational<int32_t>(0, 1),
            Rational<int32_t>(0, 100) / Rational<int32_t>(5, -10));

  EXPECT_EQ(Rational<int32_t>(1, 10000),
            Rational<int32_t>(1, -100) / Rational<int32_t>(-100, 1));

  EXPECT_EQ(Rational<int32_t>(3, 2),
            Rational<int32_t>(1, 2) / Rational<int32_t>(1, 3));

  EXPECT_EQ(Rational<int32_t>(3, 14),
            Rational<int32_t>(4, 7) / Rational<int32_t>(8, 3));
}

TEST(Rational, divideOverflowAssert)
{
  // clearly off
  EXPECT_THROW(Rational<int32_t>(1000000, 1) / Rational<int32_t>(1,10000000),
               std::overflow_error);

  // sign change
  EXPECT_THROW(Rational<int32_t>(7919, 1) / Rational<int32_t>(1,271883),
               std::overflow_error);
}

TEST(Rational, unaryMinus) {
  EXPECT_EQ(Rational<int32_t>(0,1), -Rational<int32_t>(0,1));
  EXPECT_EQ(Rational<int32_t>(-1,1), -Rational<int32_t>(1,1));
  EXPECT_EQ(Rational<int32_t>(-4,7), -Rational<int32_t>(4,7));
  EXPECT_EQ(Rational<int32_t>(-4,7), -Rational<int32_t>(-4,-7));
  EXPECT_EQ(Rational<int32_t>(4,7), -Rational<int32_t>(4,-7));
  EXPECT_EQ(Rational<int32_t>(4,7), -Rational<int32_t>(-4,7));
}

TEST(Rational, unaryMinusOverflow) {
  //EXPECT_EQ(-128, Rational<int8_t>(-128, 1).numerator());
  // this should overflow
  EXPECT_THROW(Rational<int8_t>(-1, 1) * Rational<int8_t>(-128, 1),
               std::overflow_error);
  EXPECT_THROW(-Rational<int8_t>(-128, 1),
               std::overflow_error);
}

TEST(Rational, unaryMinusuint32_t) {
  // this should be compile-time test, but there is no framework for
  // this now.
}

TEST(Rational, plus)
{
  EXPECT_EQ(Rational<int32_t>(5,6),
      Rational<int32_t>(1,2) + Rational<int32_t>(1,3));

  EXPECT_EQ(Rational<int32_t>(1,6),
      Rational<int32_t>(1,2) + Rational<int32_t>(-1,3));

  EXPECT_EQ(Rational<int32_t>(-1,6),
      Rational<int32_t>(-1,2) + Rational<int32_t>(1,3));

  EXPECT_EQ(Rational<int32_t>(-5,6),
      Rational<int32_t>(-1,2) + Rational<int32_t>(-1,3));

  EXPECT_EQ(Rational<int32_t>(0,1),
      Rational<int32_t>(-4,7) + Rational<int32_t>(4,7));

  EXPECT_EQ(Rational<int32_t>(40238, 53535),
      Rational<int32_t>(428, 5160) + Rational<int32_t>(222, 332));
}

TEST(Rational, plusOverflowAssert)
{
  // clearly off
  EXPECT_THROW(Rational<int32_t>(1,1000000) + Rational<int32_t>(1,999999999),
               std::overflow_error);

  // sign change
  EXPECT_THROW(Rational<int32_t>(1,7919) + Rational<int32_t>(1,271883),
               std::overflow_error);
}

TEST(Rational, minus)
{
  EXPECT_EQ(Rational<int32_t>(5,6),
      Rational<int32_t>(1,2) - Rational<int32_t>(-1,3));

  EXPECT_EQ(Rational<int32_t>(1,6),
      Rational<int32_t>(1,2) - Rational<int32_t>(1,3));

  EXPECT_EQ(Rational<int32_t>(-1,6),
      Rational<int32_t>(-1,2) - Rational<int32_t>(-1,3));

  EXPECT_EQ(Rational<int32_t>(-5,6),
      Rational<int32_t>(-1,2) - Rational<int32_t>(1,3));

  EXPECT_EQ(Rational<int32_t>(0,1),
      Rational<int32_t>(4,7) - Rational<int32_t>(4,7));

  EXPECT_EQ(Rational<int32_t>(-31357, 53535),
      Rational<int32_t>(428, 5160) - Rational<int32_t>(222, 332));
}

TEST(Rational, uint32_tMinus) {
  EXPECT_EQ(Rational<uint32_t>(5,6),
            Rational<uint32_t>(1) - Rational<uint32_t>(1,6));

  EXPECT_THROW(Rational<uint32_t>(1) - Rational<uint32_t>(7,6),
               std::overflow_error);
}

TEST(Rational, minusOverflowAssert) {
  // clearly off
  EXPECT_THROW(Rational<int32_t>(1,1000000) - Rational<int32_t>(1,999999999),
               std::overflow_error);

  // sign change
  EXPECT_THROW(Rational<int32_t>(1,7919) - Rational<int32_t>(1,271883),
               std::overflow_error);

  EXPECT_THROW(Rational<int8_t>(0, 1) - Rational<int8_t>(-128, 1),
               std::overflow_error);
}

} // namespace rational
} // namespace math
