#ifndef H_DUVAL
#define H_DUVAL
/**
 * @file Contains implementation of algorithms related to
 * factorization of strings to Lydon words.
 *
 * Implementation is based on the article
 *   Factorizing words over an ordered alphabet
 *   Jean Pierre Duval
 *
 * We cover lexicographically minimal cyclic shift algorithm
 * and computation of lexicographically minimal suffix of each prefix
 */

#include "debug/ppdebug.h"
#include "utils/preconditions/preconditions.h"

namespace strings {
namespace cyclic {

template<typename T>
class Duval {
 public:
  typedef int SizeType;
   /**
     * Compute lexicographically minimal suffix for each prefix
     * of the input sequence.
     *
     * Based on pseudocode of algorithm 3.1 from the Duval's article.
     * Note that "end of the string" is smaller than any letter, i.e.
     * "x" < "xa"
     *
     * @param sequence
     * @param length length of the sequence
     * @param out start positions of suffixes of each prefix.
     *    For example, out[2] holds starting index of smallest
     *    suffix for sequence seq[0],seq[1],seq[2]
     */
   static void minimumSuffixes(T* sequence, SizeType length, 
       std::vector<int>* out) {
     Preconditions::check(length > 0);
     Preconditions::check(length < std::numeric_limits<SizeType>::max() - 1);

     int start = 0;
     int i;
     int j = 2;
     // Warning: original Duval's pseudocode works by indexing from 1
     // We will shift sequence and index original minus one in the out array
     sequence--; // simple trick how to shift sequence

     out->clear();
     out->resize(length);
     std::vector<int> f(length + 1);

     out->at(0) = 0;
     f[2] = 1;
     while (j <= length) {
       i = start + f[j - start];
       while ((j <= length) && sequence[i] <= sequence[j]) {
         if (sequence[i] == sequence[j]) {
           out->at(j - 1) = out->at(i - 1) + j - i;
           i++;
         } else {
           out->at(j - 1) = start;
           i = start + 1;
         }

         j++;
         assert(j >= start);
         if (j - start <= length) {
           f[j - start] = i - start;
         }
       };
       start = out->at(i - 1) + j - i;
       // !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!! 
       // ! BIG WARNING: original pseudocode omits "j <= length" check !
       // ! and the result may be out of bounds                        !
       // !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
       if (j <= length && start == j - 1) {
         out->at(j - 1) = start;
         j++;
       }
     }
   }

   /**
     * Find lexicographically minimal cyclic shift of the sequence.
     *
     * This is almost identical to "factor" algorithm 2.1 from article,
     * but we incorporated concatenation(sequence, sequence) into
     * the algorithm instead of explicitly concatening sequences and
     * running original algorithm.
     * The idea is based on the last proposition in the article
     *
     * @param sequence
     * @param length
     *
     * @returns index of the start of the minimal lexicographic
     *   cyclic shift of the sequence. For example result 2 means
     *   that smallest shift is sequence(s[2],s[3], ..., s[0],s[1])
     */
   static int leastCyclicShift(T* sequence, int length) {
     Preconditions::check(length > 0);
     // Duval's original article uses indexing from 1,
     // we need to adjust only "start" and "return" values
     int start = -1;
     while (start < 2 * length - 1) {
       int i = start + 1;
       int j = start + 2;
       // Invariant is such that
       // a_{start+1} ... a_{i-1} is Lyndon word
       // a_{start+1} ... a_{i-1} == a_{j-i+1} ... a_{j-1}
       while (j < 2 * length && 
              sequence[i % length] <= sequence[j % length]) {
         if ((j  == start + 1 + length) && 
             (length % (j - i) == 0)) {
           return start + 1; // we have found the cyclic shift we needed
         }
         if (sequence[i % length] == sequence[j % length]) {
           // the same, we can increment both i and j
           i++;
         } else {
           // we need to reset i
           i = start + 1;
         }
         ++j;
       }
       // now we have sequence[i] > sequence[j]
       // and we can start outputting lydon factors
       while (start < i) {
         start += j - i;
       };
     }
     // oops, we should never reach this point.
     assert(false);
   }

   /**
     * Find lexicographically minimal cyclic shift of the sequence.
     *
     * This implementation is based on http://e-maxx.ru/algo/duval_algorithm
     *
     * @param sequence
     * @param length
     *
     * @returns index of the start of the minimal lexicographic
     *   cyclic shift of the sequence. For example result 2 means
     *   that smallest shift is sequence(s[2],s[3], ..., s[0],s[1])
     */
   static int leastCyclicShiftEmaxx(T* sequence, int length) {
     Preconditions::check(length > 0);
     // Duval's original article uses indexing from 1,
     // we need to adjust only "start" and "return" values
     int start = 0;
     int answer = 0;
     while (start < length) {
       answer = start;
       int i = start;
       int j = start + 1;
       while ((j < 2 * length) &&
           (sequence[i % length] <= sequence[j % length])) {
         if (sequence[ i % length] < sequence[ j % length]) {
           i = start;
         } else {
           ++j;
         }
         ++start;
       }
       while (start <= i) {
         start += j - i;
       }
     }
     return answer;
   }
};

} // namespace cyclic
} // namespace strings
#endif
