#ifndef H_STRINGS_LCS_LCS
#define H_STRINGS_LCS_LCS
/**
 * This file holds classes caculating longest common subsequence
 * of two sequences.
 */
#include <vector>
#include <algorithm>
#include "utils/preconditions/preconditions.h"
#include "debug/ppdebug.h"

namespace strings {
namespace lcs {

/**
 * Trivial dynamic programming computation of LCS
 */
template<typename T>
class LCS {
 public:
  /**
   * Compute the length of the longest common subsequence
   * of two sequences.
   *
   * Running time is <i> O(nm) </i>, memory is <i> O(n+m) </i>
   *
   * @returns LCS length
   */
  static int length(const T* seq1, int len1, const T* seq2, int len2) {
    Preconditions::check(len1 >= 0);
    Preconditions::check(len2 >= 0);
    if (len1 == 0 || len2 == 0) {
      return 0;
    }
    std::vector<int> current(len1 + 1, 0);
    std::vector<int> next(len1 + 1);
    for (int j = 0; j < len2; j++) {
      // we are going to process best results for
      // LCS(seq1, seq2[0,j) into next[j] )
      next[0] = 0;
      for (int i = 0; i < len1; i++) {
        if (seq1[i] == seq2[j]) { // match
          next[i + 1] = current[i] + 1;
        } else { // take the best of previous values
          next[i + 1] = std::max(current[i + 1], next[i]);
        }
      }
      current.swap(next); // advance by 1 line
    }

    return current[len1];
  };
  
  /**
   * Computes the longest common subsequence of two sequences.
   *
   * Running time <i> O(nm) </i>, memory <i> O(nm) </i>
   * @see length() if you only need length of the LCS
   *
   * @param out LCS will be written here
   * 
   * @returns length of the LCS
   */
  static int subsequence(const T* seq1, int len1, const T* seq2, int len2,
      std::vector<T>* out) {
    Preconditions::check(len1 >= 0);
    Preconditions::check(len2 >= 0);
    out->clear();
    if (len1 == 0 || len2 == 0) {
      return 0;
    }

    // length[i][j] is length of the LCS of
    // seq1[0..i) and seq2[0..j)
    std::vector<std::vector<int> > length(len1 + 1);
    std::vector<std::vector<bool> > prev(len1 + 1);
    // Initialize first row
    length[0].resize(len2 + 1, 0);
    prev[0].resize(len2 + 1);

    for (int i = 0; i < len1; i++) {
      length[i].resize(len2 + 1, 0);
      prev[i].resize(len2 + 1);

      for (int j = 0; j < len2; i++) {
        if (seq1[i] == seq2[j]) { // match
          length[i + 1][j + 1] == length[i][j] + 1;
        } else { // take the best of previous values
          if (length[i][j + 1] > length[i+1][j]) {
            length[i+1][j+1] = length[i][j+1];
            prev[i+1][j+1] = false;
          } else {
            length[i+1][j+1] = length[i+1][j];
            prev[i+1][j+1] = true;
          }
        }
      }
    }

    // construct the sequence in reverse order
    int i = len1;
    int j = len2;
    while (i > 0 && j > 0) {
      if (seq1[i - 1] == seq2[j - 1]) {
        out->push_back(seq1[i - 1]);
      } else {
        if (prev[i][j]) {
          j--;
        } else {
          i--;
        }
      }
    }
    // length[] should be consistent with output itself
    assert(out->size() == length[len1][len2]);

    std::reverse(out->begin(), out->end());
    return out->size();
  }
};

} // namespace lcs
} // namespace strings
#endif
