#ifndef H_LCS_LCS_HIRSCHBERG
#define H_LCS_LCS_HIRSCHBERG
/**
 * @file Contains Longest Common Subsequence computation
 * in <i> O(nm) </i>  time and <i>O(n+m)</i> memory.
 *
 * Code based on pseudocode from
 * A linear space algorithm for computing maximal common subsequences
 * by D. S. Hirschberg
 *
 */
#include <vector>
#include "utils/preconditions/preconditions.h"
#include "strings/utils/sequence_helper.h"

namespace strings {
namespace lcs {

/**
 * LCS computation in <i>O(nm)</i> time and <i>O(n+m)<i/> space
 * based on paper by Hirschberg.
 */
template <typename T>
class LCSHirschberg {
 private:
  static void saveBest(utils::SequenceHelper<const T> seq1,
      utils::SequenceHelper<const T> seq2,
      std::vector<int>* out)
  {
    std::vector<int> current(seq2.size() + 1, 0);
    std::vector<int> next(seq2.size() + 1);

    for (int i = 0; i < seq1.size(); i++) {
      // we are going to process best results for
      // LCS(seq1, seq2[0,j) into next[j] )
      for (int j = 0; j < seq2.size(); j++) {
        next[0] = 0;
        if (seq1[i] == seq2[j]) { // match
          next[j + 1] = current[j] + 1;
        } else { // take the best of previous values
          next[j + 1] = std::max(current[j + 1], next[j]);
        }
      }
      current.swap(next); // advance by 1 line
    }

    out->swap(current);
  };
  
  static void recurse(utils::SequenceHelper<const T> seq1,
      utils::SequenceHelper<const T> seq2,
      std::vector<T>* out)
  {
    assert(seq2.size() != 0);

    if (seq1.size() == 1) {
      // if there is match in character, output it
      for (int j = 0; j < seq2.size(); j++) {
        if (seq1[0] == seq2[j]) {
          out->push_back(seq1[0]);
          break;
        }
      }
      return;
    };

    // and now the hard work
    int middle = seq1.size() / 2;
    vector<int> best1;
    vector<int> best2;
    saveBest(seq1.subsequence(0, middle), seq2, &best1);
    saveBest(seq1.subsequence(middle, seq1.size()).reversed(),
             seq2.reversed(), &best2);

    // Find j such that L(i,j) + L*(i,j) = L(m,n)
    int j = -1;
    int best = -1;
    for (int i = 0; i <= seq2.size(); i++) {
      int t = best1[i] + best2[seq2.size() - i];
      if (t > best) {
        best = t;
        j = i;
      }
    }

    if (j > 0) {
      recurse(seq1.subsequence(0, middle),
              seq2.subsequence(0, j),
              out);
    }
    if (j < seq2.size()) {
      recurse(seq1.subsequence(middle, seq1.size()),
              seq2.subsequence(j, seq2.size()),
              out);
    }
  }

 public:
  static int subsequence(const T* seq1, int len1, const T* seq2, int len2,
      std::vector<T>* out) {
    Preconditions::check(len1 >= 0);
    Preconditions::check(len2 >= 0);
    out->clear();
    if (len1 == 0 || len2 == 0) {
      return 0;
    }
    
    recurse(utils::SequenceHelper<const T>(seq1, len1),
            utils::SequenceHelper<const T>(seq2, len2), out);
    return out->size();
  }


};

} // namespace lcs
} // namespace strings

#endif
