#ifndef H_STRINGS_SEARCH_KMP_KMP
#define H_STRINGS_SEARCH_KMP_KMP

#pragma GCC diagnostic warning "-Wsign-compare"

#include "utils/preconditions/preconditions.h"
#include <vector>
#include <iterator>
#include <algorithm>
#include "strings/search_callback/search_callback.h"
#include "utils/macros/unused.h"
#include <stdio.h>
/**
 * Implementation of the suffix array search
 *
 * Time: O(p log n) worst case where p is length of the pattern and
 * n is length of the suffix array
 */
namespace strings {
namespace search {

class KMP {
 private:
  template <typename _PatternIterator, typename IndexType>
  static void prepare(_PatternIterator pattern_first, _PatternIterator pattern_last,
      std::vector<IndexType > *out_)
  {
    // typedefs
    typedef typename std::iterator_traits<_PatternIterator>::difference_type LengthType;
    typedef typename std::vector<IndexType>::size_type VectorSizeType;
    // preconditions
    LengthType pattern_length = pattern_last - pattern_first;
    Preconditions::check(pattern_length > 0, "wrong order of [first, last) ?");
    Preconditions::check(std::numeric_limits<VectorSizeType>::max() >= pattern_length);
    // for unsigned types, we want to reserve max() as "-1"
    Preconditions::check(std::numeric_limits<IndexType>::max() > pattern_length);
    Preconditions::checkNotNull(out_);
    // algorithm
    std::vector<IndexType> &out = *out_;
    out.clear();
    out.resize(pattern_length);
    const IndexType MINUS_ONE = (IndexType) -1;
    out[0] = MINUS_ONE;
    if (pattern_length == 1) {
      return; // special case!
    }

    IndexType pos = MINUS_ONE;
    for (LengthType i = 1; i < pattern_length; i++) {
      _PatternIterator it_current = pattern_first + i;
      while ((pos != MINUS_ONE) && 
             (*(pattern_first + (pos + 1)) != *it_current)) {
        pos = out[pos]; // go with backlink
      }
      if (*(pattern_first + (pos +1) ) == *it_current) {
        pos++;
      }
      out[i] = pos;
    }
  }

  template <typename _Iterator, typename _PatternIterator, typename IndexType>
  static void search(
      _Iterator first,
      _Iterator last, 
      _PatternIterator pattern_first,
      _PatternIterator pattern_last,
      std::vector<IndexType>& data,
      strings::search_callback::SearchCallback<_Iterator>* callback
      ) {
    // typedefs
    typedef typename std::iterator_traits<_PatternIterator>::difference_type PatternLengthType;
    typedef typename std::iterator_traits<_Iterator>::difference_type LengthType;
    typedef typename std::vector<IndexType>::size_type VectorSizeType;
    // lengths
    PatternLengthType pattern_length = pattern_last - pattern_first;
    LengthType length = last - first;
    // preconditions
    Preconditions::check(pattern_length > 0, "wrong order of [first, last) ?");
    Preconditions::check(length > 0, "wrong order of [first, last) ?");
    Preconditions::check(std::numeric_limits<VectorSizeType>::max() >= pattern_length);
    Preconditions::check(std::numeric_limits<IndexType>::max() > pattern_length);
    // algorithm
    const IndexType MINUS_ONE = (IndexType) -1;

    IndexType pos = MINUS_ONE;
    for (LengthType i = 0; i < length; i++) {
      _Iterator it_current = first + i;
      while ((pos != MINUS_ONE) && 
             (*(pattern_first + (pos + 1)) != *it_current)) {
        pos = data[pos]; // go with backlink
      }
      if (*(pattern_first + (pos + 1) ) == *it_current) {
        pos++;
      }
      if (pos == pattern_length - 1) {
        callback->foundMatch(first + (i - pattern_length + 1));
        pos = data[pos]; // go with backlink
      }
    }

  }

 public:
  template <typename _Iterator, typename _PatternIterator>
  static void search(
      _Iterator first,
      _Iterator last, 
      _PatternIterator pattern_first,
      _PatternIterator pattern_last,
      strings::search_callback::SearchCallback<_Iterator>* callback 
      ) {
    Preconditions::check(last - first >= 0);
    Preconditions::check(pattern_last - pattern_first >= 0);
    // nothing to do
    // note that both sides are positive and thus there is no problem
    // if automatic cast will occur
    if (last - first < pattern_last - pattern_first) {
      return;
    }
    vector<int> data;
    prepare(pattern_first, pattern_last, &data);
    search(first, last, pattern_first, pattern_last, data, callback);
  }
};

} // namespace suffix_array
} // namespace strings
#endif
