#ifndef H_STRINGS_SEARCH_RABIN_KARP_RABIN_KARP
#define H_STRINGS_SEARCH_RABIN_KARP_RABIN_KARP
/**
 * This file holds Rabin-Karp string search algorithm
 */
#include "utils/preconditions/preconditions.h"
#include <vector>
#include <iterator>
#include <algorithm>
#include "strings/suffix_array_naive/sort_helper.h"
#include "strings/search_callback/search_callback.h"
#include "utils/macros/unused.h"
#include <stdio.h>
#include "rolling_hash.h"
/**
 * Implementation of the Rabin-Karp algorithm from article
 * Efficient randomized pattern-matching algorithms
 * by Karp, Richard M. and Rabin, Michael O.
 *
 *
 */
namespace strings {
namespace search {

class RabinKarp {
 private:
   template<typename _Iterator, typename _PatternIterator>
   static bool checkMatch(_Iterator start, _PatternIterator pattern_first, _PatternIterator
       pattern_last)
   {
     _Iterator it_text = start;
     _PatternIterator it_pat = pattern_first;
     while (it_pat != pattern_last) {
       if (*it_text != *it_pat) {
         return false;
       }
       ++it_text;
       ++it_pat;
     }
    return true;
   }

 public:
  template <typename _Iterator, typename _PatternIterator>
  static void search(
      _Iterator first,
      _Iterator last, 
      _PatternIterator pattern_first,
      _PatternIterator pattern_last,
      bool checkFalsePositives,
      strings::search_callback::SearchCallback<_Iterator>* callback 
      ) {
    /*
    STATIC_ASSERT(std::iterator_traits<_Iterator>::value_type ==
                  std::iterator_traits<_PatternIterator>::value_type,
                  "text and pattern should be of same type");
    STATIC_ASSERT(std::iterator_traits<_Iterator>::difference_type ==
                  std::iterator_traits<_PatternIterator>::value_type,
      "Iterator difference types are different!");
    */
    typedef typename std::iterator_traits<_PatternIterator>::value_type ValueType;
    typedef typename std::iterator_traits<_PatternIterator>::difference_type SizeType;
    SizeType pattern_length = pattern_last - pattern_first;
    SizeType text_length = last - first;
    Preconditions::check(pattern_length > 0, "wrong order of [first, last) ?");
    Preconditions::check(text_length > 0, "wrong order of [first, last) ?");
    if (text_length < pattern_length) {
      return; // nothing to do
    }

    typedef long long int HashValue;
    // Empirically determined good values causing no collisions on real testdata
    const HashValue GOOD_PRIME = 2038074743;
    const HashValue GOOD_C = 101; // is a group generator
    RollingHash<HashValue> hashFunction = RollingHash<HashValue>(pattern_length, GOOD_PRIME, GOOD_C);
    
    // compute hash value of the pattern
    _PatternIterator pattern_it = pattern_first;
    while (pattern_it != pattern_last) {
      hashFunction.roll(*pattern_it, (ValueType) 0);
      ++pattern_it;
    }

    const HashValue patternHash = hashFunction.getHash();

    hashFunction = RollingHash<HashValue>(pattern_length, GOOD_PRIME, GOOD_C);
    _Iterator it_new = first;
    for (SizeType i = 0; i < pattern_length; i++) {
      hashFunction.roll(*it_new, (ValueType) 0);
      ++it_new;
    }

    _Iterator it_old = first;
    while (it_new != last) {
      if (patternHash == hashFunction.getHash()) {
        if (checkFalsePositives && !checkMatch(it_old, pattern_first,
              pattern_last)) {
          // report an mismatch
        } else {
          callback->foundMatch(it_old);
        }
      }

      hashFunction.roll(*it_new, *it_old);
      ++it_new;
      ++it_old;
    }

  }
};

} // namespace suffix_array
} // namespace strings
#endif
