#ifndef H_STRINGS_SEARCH_RABIN_KARP_ROLLING_HASH
#define H_STRINGS_SEARCH_RABIN_KARP_ROLLING_HASH
/**
 * This file holds and rolling-hash function implementation
 * used by Rabin-Karp string matching algorithm.
 */
#include "utils/preconditions/preconditions.h"
namespace strings {
namespace search {

/**
 * Rolling hash implementation -- can efficiently compute
 * hash of sequence of fixed length.
 *
 */
template <typename BaseType>
class RollingHash
{
  typedef size_t SizeType;
public:
  /**
   * Create rolling hash and initialize it to zero sequence.
   *
   * The rolling hash value is defined as
   * hash = sum_{i=0}^{length-1} seq[i] * c_^(length-1-i) mod modulus_
   *
   * @param length length of the sequence
   * @param modulus_ modulus is best to be a prime
   * @param c_ 
   */
  RollingHash(SizeType length, BaseType modulus_, BaseType c_) {
    Preconditions::check(length > 0);
    Preconditions::check(modulus_ > 0);
    Preconditions::checkRange(c_, modulus_);
    // TODO: check for overflows!
    modulus = modulus_;
    c = c_;
    hash = 0;
    c_len = 1;
    // 
    for (SizeType i = 1; i < length; i++) {
      c_len = (c_len * c) % modulus;
    }
  }
  
  /**
   * Moves the hashing function window by one character
   *
   * Note: This implementation is very conservative in considering
   * overflows. If you know about your data limitations,
   * it can be probably optimized to two multiplications and one modular
   * division.
   *
   * @precondition discarded elements should be consistent with the #new_elements
   *  when rolling length times
   *
   * @param new_element new element of the sequence
   * @param discarded_element element of the sequence which "fell out"
   *
   * @returns new hash value
   */
  template <typename ValueType>
  inline
  BaseType roll(ValueType new_element, ValueType discarded_element) {
    hash = (modulus + hash - (discarded_element * c_len) % modulus) % modulus;
    hash = (hash * c) % modulus;
    hash = (hash + new_element) % modulus;
    return hash;
  }

  /**
   * @returns current hash value
   */
  BaseType getHash() const {
    return hash;
  }

private:
  /** modulus of the hash function */
  BaseType modulus;
  
  /** TODO */
  BaseType c;

  /** current value of the hash */
  BaseType hash;

  /** value \f$ c^{length - 1} \f$ */
  BaseType c_len;

  /** TODO: potrebne?*/
  SizeType length;
};

} // namespace suffix_array
} // namespace strings
#endif
