#ifndef H_STRINGS_SUFFIX_ARRAY_BINSEARCH_BINSEARCH
#define H_STRINGS_SUFFIX_ARRAY_BINSEARCH_BINSEARCH
/**
 * This file contains basic string-search algorithm
 * using suffix arrays
 */

#include "utils/preconditions/preconditions.h"
#include <vector>
#include <iterator>
#include <algorithm>
#include "strings/suffix_array_naive/sort_helper.h"
#include "strings/search_callback/search_callback.h"
#include "utils/macros/unused.h"
#include <stdio.h>

namespace strings {
namespace suffix_array {

/**
 * Helper for searching
 * Can compare two suffixes
 */
template <typename _Iterator>
class SearchHelper {
 public:
  /**
   * @param input whole sequence
   * @param length length of the sequence
   */
  SearchHelper(const _Iterator& first, const _Iterator& last_) {
    base = first;
    last = last_;
  }

  /**
   * @returns -1 if pattern < suffix
   *           0 if pattern = suffix
   *           1 if pattern > suffix
   */
  template <typename _PatternIterator>
  int compare(const std::pair<_PatternIterator, _PatternIterator>& pattern, const int& a) {
    // FIXME(ppershing): type
    const int l = last - base;
    Preconditions::checkRange(a, l);
    _PatternIterator itp = pattern.first;
    _Iterator ita = base + a;

    while ((ita != last) && (itp != pattern.second) && (*ita == *itp)) {
      ++ita;
      ++itp;
    }
    bool seq_last = (ita == last);
    bool pat_last = (itp == pattern.second);
    if (UNLIKELY(seq_last || pat_last)) {
      if (pat_last) return 0;
      return 1;
    }

    if (*itp < *ita) {
      return -1;
    } else {
      return 1;
    }
  }
  /**
   * Returns which of the suffixes is lexicographically smaller.
   * Note that end of sequence is less than any of the characters,
   * i.e. "x" < "xa"
   *
   * @returns true iff pattern < suffix[a]
   */
  template <typename _PatternIterator>
  bool operator()(const std::pair<_PatternIterator, _PatternIterator>& pattern, const int& a) {
    return compare(pattern, a) < 0;
  }

  template <typename _PatternIterator>
  bool operator()(const int& a, const std::pair<_PatternIterator, _PatternIterator>& pattern) {
    return compare(pattern, a) > 0;
  }

  private:
   _Iterator base;
   _Iterator last;
};


/**
 * Implementation of the suffix array search using bin-search algorithm
 *
 * Time: <it>O(p log n)</it> worst case where \a p is the length of the pattern and
 * \a n is the length of the suffix array
 *
 */
class Binsearch {
 public:
  /**
   * Search suffix array and report any matches via #callback
   *
   * @precondition <it> [first, last) </it> is valid range
   * @precondition <it> [pattern_first, pattern_last) </it> is valid range
   * @precondition \a #array is suffix array of sequence <it> [first, last) </it>
   *
   * @param first points to the start of the sequence
   * @param last points the the element after the end of the sequence
   * @param array suffix array corresponding to [first, last)
   * @param pattern_first points to the start of the pattern
   * @param pattern_last points to the element after the end of the pattern
   * @param callback to be called for each match
   */
  template <typename _Iterator, typename _PatternIterator>
  static void searchSuffixArray(
      _Iterator first,
      _Iterator last, 
      const std::vector<int>& array,
      _PatternIterator pattern_first,
      _PatternIterator pattern_last,
      strings::search_callback::SearchCallback<_Iterator>* callback 
      ) {
    typedef typename std::iterator_traits<_Iterator>::difference_type DifferenceType;
    DifferenceType length = last - first;
    // TODO(ppershing): fix unsigned comparison
    Preconditions::check(length == (DifferenceType) array.size());

    SearchHelper<_Iterator> helper(first, last);

    typedef std::vector<int>::const_iterator SAIterator;
    SAIterator match_first = std::lower_bound(
              array.begin(), array.end(),
              std::make_pair(pattern_first, pattern_last),
              helper);
    SAIterator match_last = std::upper_bound(
              array.begin(), array.end(),
              std::make_pair(pattern_first, pattern_last),
              helper);
    while (match_first != match_last) {
      int suffix = *match_first;
      callback->foundMatch(first + suffix);
      ++match_first;
    }
  }
};

} // namespace suffix_array
} // namespace strings
#endif
