#ifndef H_STRINGS_SUFFIX_ARRAY_CHECK
#define H_STRINGS_SUFFIX_ARRAY_CHECK
/**
 * This file holds a checker of suffix array consistency.
 *
 * The check is based on paper
 * Fast Lightweight Suffix Array Construction and Checking
 * by Stefan Burkhardt and Juha Kärkkäinen
 *
 */
#include "utils/preconditions/preconditions.h"
#include <vector>
#include <map>

namespace strings {
namespace suffix_array {

/*
Quick intro:

Theorem 2 from the paper
An array SA[0,n) of integers is the suffixx array of a string s[0,n)
if and only if the following conditions are satisfied:
1. For all i \in [0,n), SA[i] \in [0,n).
2. For all i \in [1,n), s[SA[i - 1]] <= s[SA[i]].
3. For all i \in [1,n) such that s[SA[i - 1]] = s[SA[i]] and
  SA[i - 1] = n - 1, there exists j, k \in [0,n) such that
  SA[j] = SA[i - 1]+1, SA[k] = SA[i]+1 and j<k.

Authors of the papers suggest this pseudocode for (rephrased) 3rd condition
  for i=n-1 downto 0:
    A[s[SA[i]]] = i
  A[s[n-1]]++ // skip SA[i]=n-1
  for i=0 to n-1:
    if SA[i]>1:
      c = s[SA[i] - 1]
      check SA[A[c]] + 1 == SA[i]
      A[c]++

Where A is map from alphabet to integers.
*/


/**
 * Class that can check validity of suffix array conditions.
 */
template <typename T>
class SuffixArrayChecker {
 FRIEND_TEST(SuffixArrayCheck, condition1);
 protected:
  /**
   * Checks that:
   * For all i \in [0,length), SA[i] \in [0,length).
   *
   * @returns true if condition holds
   */
  static bool checkCondition1Holds(const int SA[], int length) {
    for (int i = 0; i < length; i++) {
      if (SA[i] < 0 || SA[i] >= length) {
        return false;
      }
    }
    return true;
  }

  /**
   * Checks that:
   * For all i \in [1,length), s[SA[i - 1]] <= s[SA[i]].
   * Warning: there is no out-of bounds checking in this function,
   * so be sure to call checkCondition1Holds() first!
   *
   * @returns true if condition holds
   */
  static bool checkCondition2Holds(const T s[], const int SA[], int length) {
    for (int i = 1; i < length; i++) {
      if (s[SA[i - 1]] > s[SA[i]]) {
        return false;
      }
    }
    return true;
  }
  
  /**
   * Checks following reformulation of 3. condition:
   * For all characters c \in alphabet:
   * If SA[a, b] contains the suffixes starting with the character c,
   * then SA[a]+1, SA[a+1]+1,... , SA[b]+1 occur in SA in
   * this order (but not consecutively in general),
   * except that the first entry SA[a] + 1 is missing when c = s[n - 1].
   *
   * Warning: this function does not check for out-of-the bounds,
   * so you should check condition 1 first!
   *
   * Note that memory O(sigma) where sigma is size of the alphabet.
   * If your alphabet is positive-indexed and not sparse, you
   * may change std::map to array in this implementation to gain
   * speed and less memory storage.
   *
   * Note: reformulation of 3rd condition is equal to the 3rd condition
   * itself only when the first two conditions hold. The results of
   * checkCondition3HoldsKarkkainen may be therefore different from
   * checkCondition3HoldsInverses unless you assert the condition 1,2.
   *
   * Based on pseudocode from the paper.
   *
   * @returns true if condition holds
   */
  static bool checkCondition3HoldsKarkkainen(
      const T s[], const int SA[], int length) {
    std::map<T, int> A;
    for (int i = length - 1; i >= 0; i--) {
      A[s[SA[i]]] = i; 
    }
    A[s[length - 1]]++; // skip SA[i]=n-1
    for (int i = 0; i < length; i++) {
      // !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
      // ! BIG WARNING: In original pseudocode, there is SA[i] > 1 !
      // ! hovewer correct version is SA[i] >= 1                   !
      // !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
      if (SA[i] >= 1) {
        T c = s[SA[i] - 1];
        if (SA[A[c]] + 1 != SA[i]) {
          return false;
        }
        A[c]++;
      }
    }
    return true;
  }

  /**
   * Check that: 
   * For all i \in [1,length) such that s[SA[i - 1]] = s[SA[i]] and
   * SA[i - 1] != n - 1, there exists j, k \in [0,n) such that
   * SA[j] = SA[i - 1]+1, SA[k] = SA[i]+1 and j<k.
   *
   * This algorithm requires linear memory becouse it needs
   * inverse array to suffix array.
   *
   * Warning: this function does not check for out-of-bounds,
   * check condition 1 first.
   *
   * @returns true if condition holds
   */
  static bool checkCondition3HoldsInverses(
      const T s[], const int SA[], int length) {
    // Note, we need length+1, see below
    std::vector<int> ISA(length + 1, -1); // inverse array
    for (int i = 0; i < length; i++) ISA[SA[i]] = i;

    for (int i = 1; i < length; i++) {
      if ((s[SA[i - 1]] == s[SA[i]]) && (SA[i-1] != length - 1)) {
        // No need to worry about bounds, we assume 0 < SA[] < length
        // and ISA's size is length+1 so we do not need to check for
        // overflow in ISA[SA[]+1]
        int j = ISA[SA[i-1] + 1];
        int k = ISA[SA[i] + 1];
        if (j == -1 || k == -1 || j >= k) {
          return false;
        }
      }
    }
    return true;
  }

 public:
  /**
   * Liner-time check of suffix-array invariant.
   * Uses O(alphabet) memory.
   *
   * @param s original sequence
   * @param sa suffix array to be checked
   * @param length length of the sequence (and suffix array)
   *
   * @returns true if the suffix array is correct for the sequence
   */
  static bool isValidSuffixArray(const T s[], const int SA[], int
      length) {
    // Note: we need to first check out-of-bounds
    if (!checkCondition1Holds(SA, length)) {
      return false;
    }
    return checkCondition2Holds(s, SA, length) && 
           checkCondition3HoldsKarkkainen(s, SA, length);
  }

  /**
   * Liner-time check of suffix-array invariant.
   * Uses O(n) memory.
   *
   * @param s original sequence
   * @param sa suffix array to be checked
   * @param length length of the sequence (and suffix array)
   *
   * @returns true if the suffix array is correct for the sequence
   */
  static bool isValidSuffixArrayInverses(const T s[], const int SA[], int
      length) {
    // Note: we need to first check out-of-bounds
    if (!checkCondition1Holds(SA, length)) {
      return false;
    }
    return checkCondition2Holds(s, SA, length) && 
           checkCondition3HoldsInverses(s, SA, length);
  }

};

} // namespace suffix_array
} // namespace strings

#endif
