#include "gtest/gtest.h"
#include "suffix_array_check.h"
#include "debug/ppdebug.h"
#include <cstdio>
#include "utils/macros/array_size.h"
#include <algorithm>

namespace strings {
namespace suffix_array {

/**
 * Test condition 1
 */
TEST(SuffixArrayCheck, condition1) {
  typedef SuffixArrayChecker<char> Checker;

  char str[] = "abracadabra";
  int suffix[] = {10, 7, 0, 3, 5, 8, 1, 4, 6, 9, 2};
  int badrange1[] = {-1, 7, 0, 3, 5, 8, 1, 4, 6, 9, 2};
  int badrange2[] = {10, 7, 0, 3, 5, 8, 1, 4, 6, 11, 2};
  int okrange[] = {1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1};

  // check test input itself
  ASSERT_EQ(ARRAY_SIZE(suffix), strlen(str));
  ASSERT_EQ(ARRAY_SIZE(badrange1), strlen(str));
  ASSERT_EQ(ARRAY_SIZE(badrange2), strlen(str));
  ASSERT_EQ(ARRAY_SIZE(okrange), strlen(str));

  EXPECT_TRUE(
      Checker::checkCondition1Holds(suffix, strlen(str)));
  EXPECT_FALSE(
      Checker::checkCondition1Holds(badrange1, strlen(str)));
  EXPECT_FALSE(
      Checker::checkCondition1Holds(badrange2, strlen(str)));
  EXPECT_TRUE(
      Checker::checkCondition1Holds(okrange, strlen(str)));
}

/**
 * Test that different implementations are consistent.
 */
TEST(SuffixArrayCheck, KarkkainenInversesConsistency) {
  typedef SuffixArrayChecker<char> Checker;

  char str[] = "abracadabra";
  int perm[] = {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10};
  const int N = ARRAY_SIZE(perm);
  do {
    EXPECT_EQ(
        Checker::isValidSuffixArrayInverses(str, perm, N),
        Checker::isValidSuffixArray(str, perm, N));
  } while (std::next_permutation(perm, perm + N));
}

/**
 * Test that check passes only for the one permutation
 * which is the suffix array itself
 */
TEST(SuffixArrayCheck, permutations) {
  typedef SuffixArrayChecker<char> Checker;

  char str[] = "abracadabra";
  int suffix[] = {10, 7, 0, 3, 5, 8, 1, 4, 6, 9, 2};
  int perm[] = {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10};

  ASSERT_EQ(ARRAY_SIZE(perm), ARRAY_SIZE(suffix));
  ASSERT_EQ(ARRAY_SIZE(perm), strlen(str));

  const int N = ARRAY_SIZE(perm);

  do {
    bool ok = true;
    for (int i = 0; i < N; i++) {
      ok &= (suffix[i] == perm[i]);
    };
    EXPECT_EQ(ok, Checker::isValidSuffixArray(str, perm, N));

  } while (std::next_permutation(perm, perm + N));
}

} // namespace suffix_arrays
} // namespace strings
