#ifndef H_STRINGS_SUFFIX_ARRAY_LCP
#define H_STRINGS_SUFFIX_ARRAY_LCP
#include "utils/preconditions/preconditions.h"

namespace strings {
namespace suffix_array {


class LCPKasai {
 public:
  /**
   * Calculate "Height array", which is defined as follows
   * height[i] = longest_common_prefix(
   *  sequence[pos[k-1]..n), sequence[pos[k]..n))
   * i.e. it is longest common prefix of two adjacent suffixes
   * (in sorted order).
   */
  template<typename _Iterator, typename _SAIterator>
  static void getHeightArray(_Iterator seq_first, _Iterator seq_last,
      _SAIterator sa_first, _SAIterator sa_last,
      std::vector<int> * out) {
    Preconditions::checkNotNull(out);
    Preconditions::check(seq_last - seq_first > 0,
        "sequence must contain at least one character");
    Preconditions::check(seq_last - seq_first == sa_last - sa_first,
        "sequence and it's corresponding suffix array must have "
        "same length");

    unsigned int n = seq_last - seq_first;
    /**
     * Rank is inverse array to pos, i.e. rank[pos[i]]=i.
     * This actually means, that if we sort suffixes,
     * the i-th position in the sorted array will belong to
     * suffix starting at rank[i]
     */
    std::vector<int> rank(n);
    out->clear();
    out->resize(n);


    for (unsigned int i = 0; i < n; i++) {
      rank[*(sa_first + i)] = i;
    }

    const int undefined = -1;
    (*out)[0] = undefined;

    int h = 0;
    // Basic idea of the algorithm:
    // let j-1 = sa[rank[i]]
    //
    // Theorem 1 from paper:
    // Let Height[Rank[i-1]] = lcp(A_{j-1}, A_{i-1}) > 1. Then
    // Height[Rank[i]] = lcp(A_{k}, A_{i}) >= Height[Rank[i-1]] - 1
    // where k = sa[Rank[i] - 1].
    // 
    // Therefore, we compute Height[Rank[i]] iterating over i.
    // in each step, new Height is at least old minus one.

    for (unsigned int i = 0; i < n; i++) {
      assert(rank[i] >= 0);
      if (rank[i] == 0) {
        // we set the pos[0] to undefined already, nothing to do
        continue;
      }

      // !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
      // ! Warning: original pseudocode is k:=Pos[Rank[i]-1] !
      // ! but the correct variable is not k but j           !
      // ! @see Two Space Saving Tricks                      !
      // ! for Linear Time LCP Array Computation             !
      // ! by Manzini, Giovanni                              !
      // !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
      int j = *(sa_first + (rank[i] - 1));
      // we know, that the height is at least h,
      // compute the height by finding first non-match
      _Iterator ita = seq_first + (i + h);
      _Iterator itb = seq_first + (j + h);
      while ((ita != seq_last) && (itb != seq_last) &&
             (*ita == *itb)) {
        // Note: this will be called at most 2n
        // times because maximum value of h is n and it
        // can decrease by one at most n times
        ++ita;
        ++itb;
        ++h;
      }
      (*out)[rank[i]] = h;
      // decrease h for next iteration
      if (h > 0) {
        h--;
      }
    }
  }
};

} // namespace suffix_array
} // namespace strings
#endif
