#ifndef H_STRINGS_SUFFIX_ARRAY_LCP_MANZINI
#define H_STRINGS_SUFFIX_ARRAY_LCP_MANZINI
/**
 * Implementation of suffix array LCP computation
 * in linear time from article [ltLCP]
 *  Two space saving tricks for linear time LCP computation
 *  by Giovanni Manzini
 */

#include "utils/branch_predict/branch_predict.h"
#include "utils/preconditions/preconditions.h"
#include "utils/macros/evil_constructors.h"
#include "debug/ppdebug.h"

namespace strings {
namespace suffix_array {

class LCPManzini {

template<typename _Iterator>
static void compute_counts(_Iterator first, _Iterator last, int alphabet_size,
    std::vector<int>* out)
{
  // Check preconditions
  Preconditions::check(alphabet_size > 0);
  _Iterator it = first;
  while (it != last) {
    Preconditions::checkRange((int)*it, alphabet_size);
    ++it;
  }
  // Do the real work
  out->clear();
  out->resize(alphabet_size + 1); // Note: +1 is mandatory!
  it = first;
  while (it != last) {
    (*out)[1 + (*it)]++;
    ++it;
  }

  for (int i = 1; i <= alphabet_size; i++) {
    (*out)[i] += (*out)[i - 1];
  }
  out->pop_back(); // remove last element, which is not needed
  // and we stored it only for convenience

}

/**
 * Compute the "rankNext" map as defined in [ltLCP]
 * RankNext is defined as following
 * RankNext[i] = Rank[sa[i] + 1] for i=[0..n), i != Rank[n]
 * for i = Rank[i] is RankNext[i] undefined as it is the last
 * element;
 *
 *
 * @returns
 *   value i which is not amongs values of RankNext array
 *   (one value is missing)
 */

// Pseudocode taken from [ltLCP]
// Procedure Sa2RankNext(rank next)
// 1.  j = count[t[n]]++;
// 2.  rank_next[j]=0;
// 3.  for(i=1;i<=n;i++) {
// 4.    if(sa[i]==1)
// 5.      eos_pos = i;
// 6.    else {
// 7.      c = t[sa[i]-1];
// 8.      j = count[c]++;
// 9.      rank_next[j]=i;
// 10.   }
// 11. }
// 12. return eos_pos;
template<typename _Iterator, typename _SAIterator>
static int compute_rank_next(_Iterator seq_first, _Iterator seq_last,
    _SAIterator sa_first, _SAIterator sa_last, int alphabet_size,
    std::vector<int> *rank_next)
{
  Preconditions::check((seq_last - seq_first) == (sa_last - sa_first),
      "Sequence and suffix array lengths does not match");
  Preconditions::check(seq_last - seq_first > 0,
      "Can't compute rank_next on empty sequence");

  std::vector<int> count;
  // Warning: this step is not in the pseudocode, but
  // if you read paper carefully, you will know that
  // count should be initialized
  compute_counts(seq_first, seq_last, alphabet_size, &count);
  int n = seq_last - seq_first;
  int j = count[*(seq_last - 1)]++;
  int eos_pos = - 1; // undefined
  rank_next->clear();
  rank_next->resize(seq_last - seq_first);

  (*rank_next)[j] = -1; // marker
  for (int i = 0; i < n; i++) {
    if (UNLIKELY(*(sa_first + i) == 0)) {
      eos_pos = i;
    } else {
      int tmp = *(sa_first + i) - 1;
      j = count[*(seq_first + tmp)]++;
      (*rank_next)[j] = i;
    }
  }
  assert(eos_pos != -1);
  return eos_pos;
}

// Procedure Lcp9
// 1.  k = Sa2RankNext(lcp);
// 2.  h=0;
// 3.  for(i=1;i<=n;i++) {
// 4.    nextk = lcp[k];
// 5.    if(k==1) lcp[k]=-1;
// 6.    else {
// 7.      j = sa[k-1];
// 8.      while(i+h<=n && j+h<=n && t[i+h]==t[j+h])
// 9.        h++;
// 10.     lcp[k] = h;
// 11.   }
// 12.   if(h>0) h--;
// 13.   k=nextk;
// 14. }
public:
template<typename _Iterator, typename _SAIterator>
static void getHeightArray(_Iterator seq_first, _Iterator seq_last,
    _SAIterator sa_first, _SAIterator sa_last, int alphabet_size,
    std::vector<int> *out)
{
  Preconditions::check((seq_last - seq_first) == (sa_last - sa_first));

  std::vector<int>& lcp(*out); // shorthand
  int current = compute_rank_next(seq_first, seq_last,
      sa_first, sa_last, alphabet_size, out);
  int h = 0;
  int n = seq_last - seq_first;
  for (int i = 0; i < n; i++) {
    assert(current >= 0); // current is not a marker
    assert(i == *(sa_first + current)); // current == rank[i]
    int next = lcp[current];
    // skip the rank[i]==0
    if (current != 0) {
      int j = *(sa_first + (current - 1));
      _Iterator ita = seq_first + (i + h);
      _Iterator itb = seq_first + (j + h);
      while ((ita != seq_last) && (itb != seq_last) &&
             (*ita == *itb)) {
        ++ita;
        ++itb;
        ++h;
      }
      lcp[current] = h;
      if (h > 0) {
        h--;
      }
    }
    // rank[i+1] = rankNext[rank[i]]
    current = next;
  }
  assert(current == -1); // marker
}

private:
  DISALLOW_EVIL_CONSTRUCTORS(LCPManzini);
};

} // namespace suffix_array
} // namespace strings

#endif
