#ifndef H_STRINGS_SUFFIX_ARRAY_LCP_NAIVE
#define H_STRINGS_SUFFIX_ARRAY_LCP_NAIVE
#include "utils/preconditions/preconditions.h"
namespace strings {
namespace suffix_array {

/**
 * Calculates longest common prefixes of suffix arrays
 * Note: This is trivial naive implementation used mainly for testing.
 * It may have serious performance problems when used with long lcp-s.
 * Use LCPKasai or LCPManzini for real data.
 *
 */

class LCPNaive {
private:
  template<typename _Iterator>
  static inline int lcp(_Iterator pos1, _Iterator pos2, _Iterator last)
  {
    int i = 0;
    while (pos1 != last && pos2 != last && *pos1 == *pos2) {
      ++i;
      ++pos1;
      ++pos2;
    }
    return i;
  }

public:
  /**
   * Calculate "Height array", which is defined as follows
   * height[i] = longest_common_prefix(
   *  sequence[pos[k-1]..n), sequence[pos[k]..n))
   * i.e. it is longest common prefix of two adjacent suffixes
   * (in sorted order).
   */
  template<typename _Iterator, typename _SAIterator>
  static void getHeightArray(_Iterator seq_first, _Iterator seq_last,
      _SAIterator sa_first, _SAIterator sa_last, 
      std::vector<int> * out)
  {
    Preconditions::checkNotNull(out);

    Preconditions::check(seq_last - seq_first == sa_last - sa_first);

    unsigned int n = sa_last - sa_first;

    out->clear();
    out->resize(n);

    const int undefined = 0;
    (*out)[0] = undefined;

    for (unsigned int i = 1; i < n; i++) {
      _Iterator pos1 = seq_first + *(sa_first + (i - 1));
      _Iterator pos2 = seq_first + *(sa_first + i);
      (*out)[i] = lcp(pos1, pos2, seq_last);
    }
  }
};

} // namespace suffix_array
} // namespace strings
#endif
