#ifndef H_SUFFIX_ARRAY_MAMBER_MEYERS_OPT
#define H_SUFFIX_ARRAY_MAMBER_MEYERS_OPT
/**
 * This file holds implementation to O(n log n)
 * suffix array generation from article
 * Suffix arrays: A new method for on-line string searches
 * by Udi Manber, Gene Myers
 */

#include "utils/preconditions/preconditions.h"
#include <vector>
#include <algorithm>
#include "debug/ppdebug.h"
#include "utils/static_assert/static_assert.h"
#include <limits>
#include "debug/ppdebug.h"

namespace strings {
namespace suffix_array {

/**
 * Constructs suffix array in O(n (log n)^2) time
 * with Mamber-Myers algorithm, but using std::sort
 * instead of countsorts.
 *
 * The implementation uses 4N ints of memory.
 * The code is based on original from Michal Forisek.
 */
template <typename IndexType>
class ManberMyersLog2_ {
 private:

  /**
   * Suffix is helper for sorting suffixes.
   *
   * Warning: this struct does not verify any arguments,
   * it expects that caller uses it correctly!
   */
  struct Suffix {
    IndexType index;
    IndexType pos_n;
    IndexType pos_2n;

    bool operator<(const Suffix& other) const {
      return (pos_n < other.pos_n) ||
             ((pos_n == other.pos_n) && (pos_2n < other.pos_2n));
    }
  };

 public:

   /**
    * Construct suffix array from input sequence.
    *
    * Note that end-of-string character $ is less than any other
    * character. If you need $ to be greater than all characters
    * in the sorting, add one maximal characted at the end of the
    * sequence.
    *
    * Note: the implementation needs (4*sizeof(int))*N
    * bytes. This means, that processing sequence of 100M elements
    * consumes about 1.6G of memory!
    *
    * @param sequence
    * @param sigma size of alphabet
    * @param out resulting suffix array, will be rewritten.
    *   out[i] will be holding position of suffix i in the sorted
    *   array.
    */

  template<typename _Iterator>
   static void buildSuffixArray(_Iterator first,
       _Iterator last, std::vector<IndexType>* out)
   {
     typedef typename iterator_traits<_Iterator>::difference_type DifferenceType;
     typedef typename iterator_traits<_Iterator>::value_type ValueType;

     STATIC_ASSERT(std::numeric_limits<ValueType>::is_integer,
        "Value type of iterator should be integer");

     STATIC_ASSERT(std::numeric_limits<IndexType>::digits >=
                   std::numeric_limits<ValueType>::digits,
                   "IndexType is unable to hold iterator value type!");

     // This is very special case -- check that if IndexType does not have
     // additional bits over ValueType, we can't use maxvalue
     if (std::numeric_limits<IndexType>::digits ==
         std::numeric_limits<ValueType>::digits) {
       for (_Iterator it = first; it != last; ++it) {
         if ((*it) == std::numeric_limits<ValueType>::max()) {
           throw std::invalid_argument("Input sequence can't contain maxvalue."
               "Try to reduce size of the alphabet.");
         }
       }
     }

     /** @var length of the whole sequence*/
     DifferenceType n = last - first;

     if (n < 0 || n >= std::numeric_limits<IndexType>::max() ) {
        throw std::overflow_error("The length of the sequence is too big "
            "to process. Have you passed the [first,last) in correct order?");
     }


     /** @var shorthand, use result[i] instead of (*out)[i] */
     std::vector<IndexType>& result = *out;
     result.resize(n);

     /** @var holds the information about suffix positions and order*/
     std::vector<Suffix> suffixes(n);

     _Iterator it = first;
     for (IndexType i = 0; i < n; i++) {
        suffixes[i].index = i;
        // Move the range [ValueType::min, ValueType::max] to
        // [IndexType::min+1, IndexType::min+ValueType::size]
        suffixes[i].pos_n = (std::numeric_limits<IndexType>::min() + 1) + 
                ((*it) - std::numeric_limits<ValueType>::min());
        // note that suffixes[i].pos_2n does not need to be initialized here
        ++it;
     }
     assert(it == last);

     IndexType current_size = 1; // the processed size
     do {
       for (IndexType i = 0; i < n; i++) {
          result[suffixes[i].index] = suffixes[i].pos_n;
       }

       // Compute the pos_2n
       for (IndexType i = 0; i < n; i++) {
         IndexType tail_start = suffixes[i].index + current_size;
         // short suffixes (already sorted at this stage) goes before
         // other, we reserved value 0 for them
         suffixes[i].pos_2n = (tail_start >= n) ? 
           std::numeric_limits<IndexType>::min() : 1 + result[tail_start];
       }

       sort(suffixes.begin(), suffixes.end());
       result[0] = 0;
       for (IndexType i = 1; i < n; i++) {
         // check for equality using two operators less
         if (!(suffixes[i-1] < suffixes[i]) &&
             !(suffixes[i] < suffixes[i-1])) {
            result[i] = result[i-1];
         } else {
            result[i] = i;
         }
       }

       for (IndexType i = 0; i < n; i++) {
          suffixes[i].pos_n = result[i];
       }

       current_size *= 2;
     } while (current_size < n);

     for (IndexType i = 0; i < n; i++) {
        result[i] = suffixes[i].index;
     }
   }
};

typedef ManberMyersLog2_<int> ManberMyersLog2;

} // namespace suffix_array
} // namespace strings

#endif
