#ifndef H_SUFFIX_ARRAY_MAMBER_MEYERS_OPT
#define H_SUFFIX_ARRAY_MAMBER_MEYERS_OPT
/**
 * This file holds implementation to O(n log n)
 * suffix array generation from article
 * Suffix arrays: A new method for on-line string searches
 * by Udi Manber, Gene Myers
 */

#include "utils/preconditions/preconditions.h"
#include <vector>
#include <algorithm>
#include "debug/ppdebug.h"

namespace strings {
namespace suffix_array {

/**
 * Constructs suffix array in O(n log n) time
 * with Mamber-Myers algorithm.
 *
 * The implementation is optimized to use as little memory as
 * possible without great increase in code complexity,
 * in this case it is 3N ints and 2N bools of storage.
 * Note that authors of the paper claims it can be done in 2N ints,
 * but I haven't been able to fully comprehend and write bug-free
 * code for that version.
 * Note: another implementation and some discussion can be found at
 * http://forums.topcoder.com/?module=RevisionHistory&messageID=1171511
 */
class ManberMyers {
 private:
   /**
    * Sort suffixes by first character.
    * Note: This implementation is count-sort, because
    * for reasonable small alphabets quicksort is awfully slow.
    */
  template<typename _Iterator>
  static void sortByFirstCharacter(
      _Iterator first, _Iterator last, int sigma, std::vector<int>* out) {
    for (_Iterator it = first; it != last; ++it) {
      Preconditions::checkRange((int)(*it), sigma);
    }

    // Note: we NEED count sort here, std::sort is awfully slow
    // when sorting reasonable small alphabet
    // TODO(ppershing): benchmark it.
    vector<int> csort(sigma, 0);

    for (_Iterator it = first; it != last; ++it) {
      csort[*it]++;
    }

    for (int i = 1; i < sigma; i++) {
      csort[i] += csort[i - 1]; // make prefix sums
    }

    int n = last - first;
    out->resize(n);
    int i = 0;
    for (_Iterator it = first; it != last; ++it) {
      (*out)[--csort[*it]] = i++;
    }
    assert(i == n);
  }

 public:

   /**
    * Construct suffix array from input sequence.
    * Note that sequence must consist of non-negative
    * numbers from range [0, alphabet size)
    *
    * Note that end-of-string character $ is less than any other
    * character. If you need $ to be greater than all characters
    * in the sorting, add one maximal characted at the end of the
    * sequence.
    *
    * Note: the implementation needs (3*sizeof(int) + 2/8)*N
    * bytes. This means, that processing sequence of 100M elements
    * consumes about 1.2G of memory!
    *
    * @param sequence
    * @param sigma size of alphabet
    * @param out resulting suffix array, will be rewritten.
    *   out[i] will be holding position of suffix i in the sorted
    *   array.
    */

  template<typename _Iterator>
   static void buildSuffixArray(_Iterator first,
       _Iterator last, int sigma,
       std::vector<int>* out) {
     /** @var shorthand, use pos[i] instead of (*out)[i] */
     std::vector<int>& pos = *out;
     /** @var length of the whole sequence*/
     int n = last - first;

     /**
      * @var identifies start of the H-bucket.
      * During the computation, the following invariant holds:
      * bucket_h[i] == true iff Pos[i] contains leftmost suffix of
      * H-bucket, i.e. sequence[pos[i]] !=H sequence[pos[i-1]]
      */
     std::vector<bool> bucket_h(n);
     /** @var identifies start of the 2H-bucket */
     std::vector<bool> bucket_2h(n, false);

     /**
      * @var inverse array to the pos, i.e. rank[pos[i]] == i.
      * This actually means, that if we sort suffixes,
      * the i-th position in the sorted array will belong to
      * suffix starting at rank[i]
      */
     std::vector<int> rank(n);
     std::vector<int> count(n);


     // first stage
     sortByFirstCharacter(first, last, sigma, out);

     // find the initial buckets
     bucket_h[0] = false;
     for (int i = 1; i < n; i++) {
       bucket_h[i] = (*(first + pos[i]) != *(first + pos[i - 1]));
     }

     // second stage, sort by 2, 4, 8, 16, ... characters
     for (int h = 1; h < n; h*=2) {
       // Now we have sorted suffixes according to first h characters
       // let's do it for first 2h

       int bucket_start = 0;
       int bucket_count = 0;

       // for each =H bucket [l,r]
       // clear representative count values
       // and reset ranks to the left of the bucket
       // (we will use left of the bucket as representative
       // for counting)
       count[0] = 0;
       for (int i = 0; i < n; i++) {
         if (bucket_h[i]) {
            bucket_start = i;
            count[i] = 0;
            bucket_count ++;
         }
         rank[pos[i]] = bucket_start;
       }

       if (bucket_count == n) {
          return; // we are lucky and don't need another iteration
       }

       // indexing is ok, h < n
       // TODO(ppershing): why these?
       count[rank[n - h]]++;
       bucket_2h[rank[n - h]] = true;

       bucket_start = 0;
       // Go through all buckets, one at the time
       while (bucket_start < n) {
         // find the end of the bucket
         int bucket_end = bucket_start + 1;
         while (bucket_end < n && !bucket_h[bucket_end]) {
           bucket_end++;
         }

         for (int c = bucket_start; c < bucket_end; c++) {
           // select (pos[c] - h) \union [0, n)
           int d = pos[c] - h;
           if (d >= 0) {
             assert(d < n);
             int head = rank[d];
             rank[d] += count[head];
             count[head]++;
             bucket_2h[rank[d]] = true;
           }
         }
         for (int c = bucket_start; c < bucket_end; c++) {
           // select (pos[c] - h) \union [0, n)
           int d = pos[c] - h;
           if (d >= 0 && bucket_2h[rank[d]]) {
             for (int f = rank[d] + 1; !bucket_h[f] && bucket_2h[f]; f++) {
              bucket_2h[f] = false;
             }
           }
         }  

         bucket_start = bucket_end;
       }

       for (int i = 0; i < n; ++i) {
         pos[rank[i]] = i;
         if (bucket_2h[i] && !bucket_h[i]) {
           // set(i, h + min_height(rank[pos[i-1]+h], rank[pos[i]+h]))
           bucket_h[i] = true;
         }
       }
     }
   }
};

} // namespace suffix_array
} // namespace strings

#endif
