#ifndef H_SEQUENCE_HELPER
#define H_SEQUENCE_HELPER
/**
 * @file Contains helper class which helps with creating sequences/subsequences.
 */

#include "utils/preconditions/preconditions.h"

namespace strings {
namespace utils {

/**
 * This is a simple helper class.
 * It takes an pointer to the sequence array and can
 * be used to access this array.
 * Moreover, there are some handy functions like
 * Creating subsequences from the original sequence
 * or reversing the direction.
 *
 * Note that SequenceHelper holds only pointers to the
 * data, not the actual data itself. This means that
 * <li> you shouldn't deallocate the array itself if you have
 *  active instance of SequenceHelper
 * <li> you may edit the data and the changes will be reflected in
 *  SequenceHelper
 * <li> SequenceHelper is memory-cheap (no big arrays copying, can be on
 *  the stack
 */
template<typename T>
class SequenceHelper {
  private:
    T* base;
    int start;
    int length;

    SequenceHelper(T* base_, int start_, int length_) {
      base = base_;
      start = start_;
      length = length_;
    }
  public:

    /**
     * Construct the helper.
     *
     * @param base_ pointer to the start of the sequence (index 0)
     * @param length_ length of the sequence
     */
    SequenceHelper(T* base_, int length_) {
      base = base_;
      start = 0;
      length = length_;
    }

    /**
     * Reverse SequenceHelper, i.e. the resulting object will satisfy
     * new[0] = old[size - 1], ... , new[size - 1] = old[0]
     *
     * @returns reversed SequenceHelper
     */
    SequenceHelper reversed() const {
      if (length >= 0) {
        return SequenceHelper(base, start + length - 1, -length);
      } else {
        return SequenceHelper(base, start + length + 1, -length);
      }
    }

    /**
     * Create subsequence of this sequence, i.e. the resulting object
     * will be reindexed like this
     * SequenceHelper([0, 1, 2, 3, 4, 5]).subsequence(2, 5)
     * is equal to SequenceHelper([2, 3, 4])
     *
     * The resulting subsequence will be representing elements
     * [start_, end_) of the original sequence
     *
     * @param start_ starting index of the subsequence (will become
     *   zero in result)
     * @param end_ index after the last position of the subsequence
     * 
     * @returns subsequence SequenceHelper
     */
    SequenceHelper subsequence(int start_, int end_) const {
      Preconditions::check(start_ <= end_);
      Preconditions::checkRange(start_, size());
      Preconditions::checkRange(end_, size() + 1);
      int len = end_ - start_;
      if (length >= 0) {
        return SequenceHelper(base, start + start_, len);
      } else {
        return SequenceHelper(base, start - start_, -len);
      }
    }

    /**
     * Returns the size of the sequence
     */
    int size() const {
      return abs(length);
    }

    /**
     * [] operator like in arrays.
     *
     * @param pos index of the element we want to retrieve.
     */
    const T& operator[](int pos) const {
      Preconditions::checkRange(pos, size());
      if (length >= 0 ) {
        return base[start + pos];
      } else {
        return base[start - pos];
      }
    }
    
    /**
     * [] operator like in arrays.
     *
     * @param pos index of the element we want to retrieve.
     */
    T& operator[](int pos) {
      Preconditions::checkRange(pos, size());
      if (length >= 0 ) {
        return base[start + pos];
      } else {
        return base[start - pos];
      }
    }
};

} // namespace sequence_helper
} // namespace strings
#endif
