#ifndef H_STRINGS_UTILS_SEQUNENCE_LOADER
#define H_STRINGS_UTILS_SEQUNENCE_LOADER
/**
 * @file Contains helper class that can read sequences of characters
 * from files.
 */
#include <stdio.h>
#include <string>
#include <map>
#include <exception>
#include <fstream>
#include <stdexcept>
#include "utils/preconditions/preconditions.h"

namespace strings {
namespace utils {
/**
 * Implementation of file reader which can re-map
 * input characters according to same map.
 *
 * This class is useful for testing string manipulations,
 * as you can re-map 256 character alphabet to anything you want.
 */
class SequenceLoader{
 public:
  /**
   * Load sequence of re-mapped characters from input file
   *
   * @param filename
   * @param map which characters match to which value, 
   *    if the character read on input is not included in the map,
   *    it won't appear in the resulting sequence.
   *
   * @returns re-mapped sequence read from file
   */
  static std::vector<unsigned char> loadSequence(
      const char* filename,
      const std::map<unsigned char, unsigned char>& conversion_map) {
    Preconditions::check(filename != NULL);
    std::ifstream f(filename, std::ios::binary);
    if (!f) {
      throw std::runtime_error(std::string("Couldn't open file ") + filename);
    }

    std::vector<unsigned char> result;

    int cnt = 0;
    while (f.good()) {
      // Warning: statement "f >> c;"
      // is inherently bad, because default is to strip whitespace
      char tmp;
      f.get(tmp);
      unsigned char c = tmp;
      if (!f.good()) {
        if (f.bad()) {
          throw new std::runtime_error("Problem reading file.");
        }
        break;
      }
      cnt++;
      if (conversion_map.find(c) != conversion_map.end()) {
        // can't use conversion_map[c] as map is passed as const
        // and that call may modify it
        result.push_back(conversion_map.find(c)->second);
      }
    }
    f.close();
    return result;
  }
};

} // namespace utils
} // namespace strings
#endif
