#ifndef H_UTILS_ASSERT_INTEGER_OVERFLOW
#define H_UTILS_ASSERT_INTEGER_OVERFLOW

#include "utils/static_assert/static_assert.h"
#include <limits>

/**
 * This template is used to ensure (at compile type) that user can't
 * pass integral value of bigger type than specified.
 * This is very useful to remind users about possible overflows.
 * Note that this template is also checking unsigned->signed
 *
 * Note: this template is quite aggresive, you may want to consider
 * using run-time checks (especially signed constants versus unsigned expected)
 *
 * Basic usage is:
 *
 * template <typename T>
 * userFunction(T arg1) {
 *  STATIC_ASSERT_CHECK_INTEGER_OVERFLOW<T, short>();
 *  short tmp = (short) arg1;
 * }
 *
 * and this code won't compile for example for T=32bit int (and short = 16bit int)
 */
template <typename SuppliedType, typename RequiredType>
inline void STATIC_ASSERT_CHECK_INTEGER_OVERFLOW() {
  STATIC_ASSERT(std::numeric_limits<SuppliedType>::is_integer,
      "Invalid SuppliedType - should be integer");
  STATIC_ASSERT(std::numeric_limits<RequiredType>::is_integer,
      "Invalid RequiredType - should be integer");

  // disallow signed->unsigned 
  STATIC_ASSERT(std::numeric_limits<RequiredType>::is_signed
      || !(std::numeric_limits<SuppliedType>::is_signed),
      "Signed to unsigned overflow");

  // digits is enough to check (works with size of signed vs unsigned also).
  STATIC_ASSERT(std::numeric_limits<SuppliedType>::digits <=
                std::numeric_limits<RequiredType>::digits,
      "Possible overflow!");
}

#endif
