#ifndef H_BENCHMARK
#define H_BENCHMARK

/**
 * @file contains BENCHMARK and AUTO_BENCHMARK macros
 */

#include <limits>
#include "color.h"
#include "utils/timer/timer.h"
#include "utils/static_assert/static_assert.h"

namespace utils {
namespace benchmark {

/**
 * Minimum time (in seconds) for benchmark
 * to have usedul results.
 *
 * Note: current implementation of benchmark is using
 * system timer to determine running time.
 * This have the best resolution of 16ms and may
 * wildly change, the present constant is
 * conservative for reliable benchmarking
 */
const double MIN_BENCHMARK_TIME = 1.5;

/**
 * @param times How many times the test was run
 * @param run_time_sec Run time of the test in seconds
 * @param function_str string containing the name
 *        of the function and arguments
 */
void printBenchmarkResults(
    long long int times,
    double run_time_sec,
    const char* function_str)
{
  bool isLongEnough = run_time_sec > MIN_BENCHMARK_TIME;

  color::Color color = isLongEnough ? color::CYAN : color::PINK;
  double avg_time_sec = run_time_sec / times;
  double times_per_sec = 1 / avg_time_sec;

  color::colorPrintf(color,
    "BENCHMARK: time %6.2lfs avg.time %10.5lfms "
    "%12.5lf times/ms, %10lldx %s\n",
    run_time_sec, avg_time_sec*1000, times_per_sec/1000, times, function_str);
}


/**
 * Macro for benchmarking code.
 *
 * The benchmark runs specified \a times the code 
 * and prints nicely formatted report to stdout.
 * Usage:
 * <code>
 *   BENCHMARK(100, my_function());
 * </code>
 * calls 100 times function my_function.
 *
 * Note: "function" does not need to be the function,
 * in fact, it may be sequence of statements of event the {} block.
 *
 * @warning Sometimes compiler optimizes out parts of you code!
 *
 * @param times number of times the \a code should be run
 * @param code code that should be run
 *
 */
#define BENCHMARK(times, code) { \
    STATIC_ASSERT(std::numeric_limits<typeof(times)>::is_integer, "") \
    ::utils::timer::Timer timer = ::utils::timer::Timer(); \
    for (typeof(times) q__= 0; q__ < times; q__++) { \
        code;\
    } \
    ::utils::benchmark::printBenchmarkResults(\
        times, timer.elapsed_time_sec(), #code); \
  }

/**
 * Macro for benchmarking the code
 *.
 * The benchmark runs the code several times depenging
 * on the speed of execution and prints nicely formatted report to stdout.
 * Usage:
 * <pre>
 *   AUTO_BENCHMARK(my_function());
 * </pre>
 * calls the function my_function for approximately
 * \a MIN_BENCHMARK_TIME seconds.
 *
 * Note: This macro uses exponential decay to guess the correct
 * number of iterations, and so that instrumentation is not big.
 *
 * @warning Sometimes compiler optimizes out parts of you code!
 *
 * @param code code that should be run
 */
#define AUTO_BENCHMARK(code) { \
    ::utils::timer::Timer timer = ::utils::timer::Timer(); \
    long long int times = 0; \
    int i = 0; \
    while ((timer.elapsed_time_sec() < ::utils::benchmark::MIN_BENCHMARK_TIME) &&  \
        (i < std::numeric_limits<long long int>::digits)) { \
      long long int t = 1LL << i; \
        for (long long q__ = 0; q__ < t; q__++) { \
            code; \
        } \
      times += t; \
      i++; \
    } \
    ::utils::benchmark::printBenchmarkResults(\
        times, timer.elapsed_time_sec(), #code); \
  }

} // namespace benchmark
} // namespace utils
#endif
