#ifndef H_PRECONDITIONS
#define H_PRECONDITIONS

#include <stdexcept>
#include <stdio.h>
#include "utils/branch_predict/branch_predict.h"

// Note: Preconditions is in global namespace!
/**
 * Preconditions is a helper class containing static methods that
 * can check the validity of function arguments (function pre-conditions).
 * In case there is a failure, Preconditions will throw an
 * std::invalid_argument error.
 *
 * Note: Do not overuse the throwing/catching exceptions. Failed preconditions
 * are running much slower because of branch prediction and slowness of
 * exceptions.
 */
class Preconditions {
 public:
  /**
   * Checks whether the expression is true
   *
   * @param expression value that is expected to be true
   *
   * @throws invalid_argument if expression is false
   */
  static void check(bool expression) {
    Preconditions::check(expression, "Precondition failed");
  }

  /**
   * Checks whether the expression is true
   *
   * @param expression value that is expected to be true
   * @param message of the thrown expection if case of failure
   *
   * @throws invalid_argument if expression is false
   */
  static void check(bool expression, const char* message) {
    if (UNLIKELY(message == NULL)) { // self-check
      throw std::invalid_argument("message shoudn't be NULL");
    }

    if (UNLIKELY(!expression)) {
      throw std::invalid_argument(message);
    }
  }

  /**
   * Check that index is in range [0, size)
   *
   * Warning: never ever override this template for using
   * two different types - it may end up with
   * nasty results because of casting like
   * checkRange(unsigned int, int) and check will be done in ints!
   *
   * @throws invalid_argument in case of an error
   */
  template <typename T>
  static void checkRange(T index, T size) {
    // Note: conversion of zero to T is required!
    Preconditions::checkRange(index, (T)0, size);
  }

  /**
   * Check that index is in range [low, high)
   *
   * @see warnings for checkRange(index, size) implementation
   *
   * @throws invalid_argument in case of an error
   */
  template<typename T>
  static void checkRange(T index, T low, T high) {
    Preconditions::check(low < high, "Bad range!");
    Preconditions::check(index >= low, "Index out of range");
    Preconditions::check(index < high, "Index out of range");
  }

  /**
   * Checks that pointer is not NULL.
   *
   * @throws invalid_argument in case of an error
   */
  template<typename T>
  static void checkNotNull(const T* ptr) {
    Preconditions::check(ptr != NULL, "Variable can't be null pointer!");
  }
};

#endif
