#ifndef H_UTILS_TIMER_TIMER
#define H_UTILS_TIMER_TIMER

/**
 * @file Contains Timer class providing basic measuring of time intervals
 */
#include <time.h>
#include <stdexcept>

namespace utils {
namespace timer {

/**
 * Basic measurement of time intervals
 *
 * Usage:
 * <pre>
 *  Timer t;
 *  hard_work();
 *  cout << t.get_elapsed_time();
 *  t.reset();
 *  hard_work();
 *  cout << t.get_elapsed_time();
 * </pre>
 *
 * @warning On some systems long times may be wrapped!
 *
 */
class Timer {
 public:
  /**
   * Construct the Timer, the time is measured from this moment
   * @throws std::runtime_error on failure
   */
  Timer() {
    reset();
  }

  /**
   * Reset the timer. Measure time from this instant.
   * @throws std::runtime_error on failure
   */
  void reset() {
    clock_t new_time = clock();
    if (new_time == -1) {
      throw std::runtime_error("Error reading clock!");
    }
    start_time = new_time;
  }

  /**
   * Returns elapsed time since last reset()
   *
   * @warning on sime systems long times may get wrapped
   *
   * @returns elapsed time in seconds
   * @throws std::runtime_error on failure
   */
  double elapsed_time_sec() {
    clock_t current = clock();
    if (current == -1) {
      throw std::runtime_error("Error reading clock!");
    }
    clock_t elapsed = current - start_time;
    return (double) elapsed / (double) CLOCKS_PER_SEC;
  }

 private:
  /** Last time of the reset */
  clock_t start_time;
};

} // namespace timer
} // namespace utils
#endif
